/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "articulation_controller.h"

#include <cassert>
#include <iostream>

#include "articulation_loop_controller.h"
#include "controller_exception.h"
#include "math_util.h"
#include "parameters_logger.h"
#include "program_configuration.h"
#include "special_transition_processor.h"
#include "transition_processor.h"
#include "wavetable.h"



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
ArticulationController::ArticulationController(const std::string& configDirPath,
								const std::string& inputFilePath,
								const std::string& outputFilePath,
								const ProgramConfiguration& configuration,
								bool verbose)
		: tubeModel_(
#ifdef HAVE_LIBSAMPLERATE
			new TubeModel<ArticulationLoopController, Wavetable, ECLSampleRateConverter>(configuration, verbose)
#else
			new TubeModel<ArticulationLoopController, Wavetable, SampleRateConverter>(configuration, verbose)
#endif
		)
		, phase_(PHASE_STEADY_STATE)
		, rule_(0)
		, time_(0)
		, configuration_(configuration)
		, controlPeriod_(MathUtil::roundToInt<int>(tubeModel_->sampleRate() / configuration_.controlRate))
		, controlRate_(tubeModel_->sampleRate() / controlPeriod_)
		, currentPosture_(0)
		, pitchProcessor_(controlRate_)
		, pitchFilter_(controlRate_, 0.020 /* period in seconds */)
		, fricPosFilter_(controlRate_, 0.004 /* period in seconds */)
//		, fricCFFilter_(controlRate_, 0.020 /* period in seconds */)
//		, fricBWFilter_(controlRate_, 0.020 /* period in seconds */)
		, inFile_(inputFilePath)
		, model_(configDirPath)
{
	model_.load();
	model_.filterInput(inFile_);

	if (Main::verbose()) {
		printInputFileInfo();
		//model_.printInfo();
	}

	// Prepare transition list.
	for (TransitionProcessorList::size_type size = transProcessorList_.size(), i = 0; i < size; ++i) {
		transProcessorList_[i].reset(
				new TransitionProcessor(model_, static_cast<Parameter::Code>(i)));
	}

	// Prepare special transition list.
	for (SpecialTransitionProcessorList::size_type size = specialTransProcessorList_.size(), i = 0; i < size; ++i) {
		specialTransProcessorList_[i].reset(
				new SpecialTransitionProcessor(model_, static_cast<Parameter::Code>(i)));
	}

	nextRule();

	if (Main::logControlParameters()) {
		std::string loggerFile;
		std::string::size_type pos = outputFilePath.rfind('.');
		if (pos != std::string::npos) {
			loggerFile = outputFilePath.substr(0, pos + 1) + "txt";
		} else {
			loggerFile = outputFilePath + ".txt";
		}
		paramLogger_.reset(new ParametersLogger(loggerFile.c_str()));
	}

	paramBuf_.reserve(PARAMETER_BUFFER_MAX_SIZE);
}

/*******************************************************************************
 * Destructor.
 */
ArticulationController::~ArticulationController()
{
}

/*******************************************************************************
 *
 */
void
ArticulationController::nextRule()
{
	// *** Order is important! ***

	time_ = 0;
	phase_ = PHASE_STEADY_STATE;

	postureList_.clear();
	currentPosture_ = 0;
	durationList_.clear();
	model_.clearFormulaSymbolList();

	const InputFile::InputCommandList& cmdList = inFile_.cmdList();

	// Get the first 4 postures.
	int i = 0;
	for (InputFile::InputCommandList::const_iterator itCmd = cmdList.begin(); itCmd != cmdList.end() && i < 4; ++itCmd, ++i) {
		durationList_.push_back(itCmd->duration);
		const Posture* pp = model_.findPosture(itCmd->phoneme);
		if (!pp) {
			throw ControllerException(ControllerError("Posture not found: ") << itCmd->phoneme);
		}
		postureList_.push_back(pp);
	}

	// Find the first matching rule.
	rule_ = model_.findFirstMatchingRule(postureList_);
	if (rule_ == 0) {
		std::cerr << "cmdlist.size=" << inFile_.cmdList().size() << std::endl;
		throw ControllerException("No rule found.");
	}

	// Consider only the postures needed by the rule.
	postureList_.resize(rule_->getNumberOfPostures());
	durationList_.resize(postureList_.size());

	// Use the rule's minimum durations, if necessary.
	for (DurationList::size_type size = durationList_.size(), i = 0; i < size; ++i) {
		if (durationList_[i] > 0.0f) {
			const float adjustedDuration = durationList_[i] / configuration_.globalTempo;
			switch (i) {
			case 0:
				durationList_[i] = std::max(adjustedDuration, static_cast<float>(rule_->minDuration12()));
				break;
			case 1:
				durationList_[i] = std::max(adjustedDuration, static_cast<float>(rule_->minDuration23()));
				break;
			case 2:
				durationList_[i] = std::max(adjustedDuration, static_cast<float>(rule_->minDuration34()));
				break;
			case 3:
				durationList_[i] = adjustedDuration;
				break;
			}
		}
	}

	if (Main::verbose()) {
		std::cout << "Rule=" << rule_->number() << " [";
		for (PostureTmpList::size_type size = postureList_.size(), i = 0; i < size; ++i) {
			std::cout << " " << postureList_[i]->name() << '(' << durationList_[i] << ')';
		}
		std::cout << " ]" << std::endl;
	}

	fillFormulaSymbolList();

	preparePitchProcessor();

	prepareTransitionProcessor();

	// Remove consumed postures, except the last.
	inFile_.removeFirstCmd(postureList_.size() - 1);
}

/*******************************************************************************
 *
 */
void
ArticulationController::fillFormulaSymbolList()
{
	// *** Order is important! ***

	// Get Postures' symbols.
	switch (rule_->getNumberOfPostures()) {
	case 4:
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TRANSITION4, postureList_[3]->symbols().transition);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSA4      , postureList_[3]->symbols().qssa);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSB4      , postureList_[3]->symbols().qssb);
		// fall through
	case 3:
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TRANSITION3, postureList_[2]->symbols().transition);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSA3      , postureList_[2]->symbols().qssa);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSB3      , postureList_[2]->symbols().qssb);
		// fall through
	case 2:
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TRANSITION2, postureList_[1]->symbols().transition);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSA2      , postureList_[1]->symbols().qssa);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSB2      , postureList_[1]->symbols().qssb);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TRANSITION1, postureList_[0]->symbols().transition);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSA1      , postureList_[0]->symbols().qssa);
		model_.setFormulaSymbolValue(FormulaSymbol::SYMB_QSSB1      , postureList_[0]->symbols().qssb);
		break;
	}

	// Calculate symbols with tempo = 1.0.
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12, 1.0f);
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO23, 1.0f);
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO34, 1.0f);
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_RD   , model_.evalEquationFormula(rule_->exprSymbolEquations().ruleDuration));
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_MARK1, model_.evalEquationFormula(rule_->exprSymbolEquations().mark1));
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_MARK2, rule_->getNumberOfPostures() >= 3 ?
								model_.evalEquationFormula(rule_->exprSymbolEquations().mark2)
								: 0.0f);

	// Calculate adjusted tempos.
	switch (rule_->getNumberOfPostures()) {
	case 2:
		if (durationList_[0] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12,
				(postureList_[0]->symbols().duration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD)) / durationList_[0]);
		}
		break;
	case 3:
		if (durationList_[0] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12,
				(postureList_[0]->symbols().duration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1)) / durationList_[0]);
		}
		if (durationList_[1] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO23,
				(model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1)) / durationList_[1]);
		}
		break;
	case 4:
		if (durationList_[0] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12,
				(postureList_[0]->symbols().duration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1)) / durationList_[0]);
		}
		if (durationList_[1] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO23,
				(model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1)) / durationList_[1]);
		}
		if (durationList_[2] > 0.0f) {
			model_.setFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO34,
				(model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2)) / durationList_[2]);
		}
		break;
	}

	// Calculate symbols with adjusted tempos.
	// This adjustment may not be exact, because there are some equations that use values that are not divided by the tempo.
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_RD   , model_.evalEquationFormula(rule_->exprSymbolEquations().ruleDuration));
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_MARK1, model_.evalEquationFormula(rule_->exprSymbolEquations().mark1));
	model_.setFormulaSymbolValue(FormulaSymbol::SYMB_MARK2, rule_->getNumberOfPostures() >= 3 ?
								model_.evalEquationFormula(rule_->exprSymbolEquations().mark2)
								: 0.0f);
}

/*******************************************************************************
 *
 */
void
ArticulationController::preparePitchProcessor()
{
	pitchProcessor_.reset();

	const float steadyStateDuration = postureList_[0]->symbols().duration / model_.getFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12);
	const float totalDuration = steadyStateDuration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD);

	const InputFile::InputCommandList& cmdList = inFile_.cmdList();
	InputFile::InputCommandList::const_iterator itCmd = cmdList.begin();
	for (int i = 0, numCmd = rule_->getNumberOfPostures() - 1;
					itCmd != cmdList.end() && i < numCmd;
					++i, ++itCmd) {
		// Get time offset and interval.
		float timeOffset = 0.0, timeInterval = 0.0;
		assert(i >= 0 && i <= 2);
		switch (i) {
		case 0: // first command
			assert(numCmd >= 1 && numCmd <= 3);
			switch (numCmd) {
			case 1: // diphone
				timeInterval = totalDuration;
				break;
			case 2: // triphone
				// fall-through
			case 3: // tetraphone
				timeInterval = steadyStateDuration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
				break;
			}
			break;
		case 1: // second command
			timeOffset = steadyStateDuration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
			assert(numCmd == 2 || numCmd == 3);
			switch (numCmd) {
			case 2: // triphone
				timeInterval = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
				break;
			case 3: // tetraphone
				timeInterval = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
				break;
			}
			break;
		case 2: // third command
			assert(numCmd == 3);
			// tetraphone
			timeOffset = steadyStateDuration + model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2);
			timeInterval = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD) - model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2);
			break;
		}
		for (InputFile::InputCommand::FrequencyTargetList::const_iterator itFreq = itCmd->frequencyTargetList.begin();
				itFreq != itCmd->frequencyTargetList.end();
				++itFreq) {
			pitchProcessor_.addTarget((itFreq->position / 100.0f) * timeInterval + timeOffset, Util::pitch(itFreq->frequency));
		}
	}
	// Add the first frequency target from the next command, if its position is zero.
	if (itCmd != cmdList.end()
			&& !itCmd->frequencyTargetList.empty()
			&& itCmd->frequencyTargetList[0].position == 0.0f) {
		pitchProcessor_.addFinalTarget(totalDuration, Util::pitch(itCmd->frequencyTargetList[0].frequency));
	}
}

/*******************************************************************************
 *
 */
void
ArticulationController::prepareTransitionProcessor()
{
	// Transitions.
	for (TransitionProcessorList::size_type size = transProcessorList_.size(), i = 0; i < size; ++i) {
		const Transition* tp = model_.findTransition(rule_->getParamProfileTransition(static_cast<Parameter::Code>(i)));
		if (!tp) {
			throw ControllerException(ControllerError("Transition not found: ") << rule_->getParamProfileTransition(static_cast<Parameter::Code>(i)));
		}
		transProcessorList_[i]->prepare(*tp, postureList_);
	}

	// Special transitions.
	for (SpecialTransitionProcessorList::size_type size = specialTransProcessorList_.size(), i = 0; i < size; ++i) {
		const std::string& transitionName = rule_->getSpecialProfileTransition(static_cast<Parameter::Code>(i));
		if (!transitionName.empty()) {
			const Transition* stp = model_.findSpecialTransition(transitionName);
			if (!stp) {
				throw ControllerException(ControllerError("Special transition not found: ") << transitionName);
			}
			specialTransProcessorList_[i]->prepare(*stp);
		} else {
			specialTransProcessorList_[i]->clear();
		}
	}
}

/*******************************************************************************
 * Returns true if there is nothing more to do.
 */
bool
ArticulationController::isFinished()
{
	return rule_ == 0 && inFile_.cmdList().size() <= 1;
}

/*******************************************************************************
 *
 */
float
ArticulationController::getTransitionValue(Parameter::Code param, float time)
{
	float value = transProcessorList_[param]->getValue(currentPosture_, time); // base value

	if (!specialTransProcessorList_[param]->empty()) {
		value += specialTransProcessorList_[param]->getValue(time); // special value
	}

	return value;
}

/*******************************************************************************
 * Gets the next parameter values using the current Rule, and returns false if
 * and only if there is no more data.
 */
bool
ArticulationController::nextControlParametersUsingCurrentRule(ControlParameters& param)
{
	const float epsilon = 1.0e-6f;

	if (paramLogger_.get() && time_ == 0) {
		paramLogger_->logMarker(0, postureList_[0]->name());
	}

	++time_;
	const float timeMS = (time_ / controlRate_) * 1.0e3f;
	const float steadyStateDuration = postureList_[0]->symbols().duration / model_.getFormulaSymbolValue(FormulaSymbol::SYMB_TEMPO12);

	if (timeMS < steadyStateDuration + epsilon) {
		// For the duration of the first posture.

		const Posture& postureIni = *postureList_[0];

		param.glotPitch = postureIni.getParameterValue(Parameter::PARAMETER_GLOT_PITCH);
		param.glotVol   = Util::amplitude(postureIni.getParameterValue(Parameter::PARAMETER_GLOT_VOL));
		param.aspVol    = Util::amplitude(postureIni.getParameterValue(Parameter::PARAMETER_ASP_VOL));
		param.fricVol   = Util::amplitude(postureIni.getParameterValue(Parameter::PARAMETER_FRIC_VOL));
		param.fricPos   = postureIni.getParameterValue(Parameter::PARAMETER_FRIC_POS);
		param.fricCF    = postureIni.getParameterValue(Parameter::PARAMETER_FRIC_CF);
		param.fricBW    = postureIni.getParameterValue(Parameter::PARAMETER_FRIC_BW);
		param.radius[0] = postureIni.getParameterValue(Parameter::PARAMETER_R1);
		param.radius[1] = postureIni.getParameterValue(Parameter::PARAMETER_R2);
		param.radius[2] = postureIni.getParameterValue(Parameter::PARAMETER_R3);
		param.radius[3] = postureIni.getParameterValue(Parameter::PARAMETER_R4);
		param.radius[4] = postureIni.getParameterValue(Parameter::PARAMETER_R5);
		param.radius[5] = postureIni.getParameterValue(Parameter::PARAMETER_R6);
		param.radius[6] = postureIni.getParameterValue(Parameter::PARAMETER_R7);
		param.radius[7] = postureIni.getParameterValue(Parameter::PARAMETER_R8);
		param.velum     = postureIni.getParameterValue(Parameter::PARAMETER_VELUM);

		if (paramLogger_.get()) {
			paramLogger_->logParam(param);
		}

		modifyControlParameters(param);

		return true;
	}

	int marker = -1;

	if (phase_ == PHASE_STEADY_STATE) {
		phase_ = PHASE_TRANSITION_1;
		marker = 1;
	}

	// Transition time;
	float transTimeMS = timeMS - steadyStateDuration;

	if (transTimeMS > model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD)) { // time > rule duration
		//if (Main::verbose()) cout << "End of transition." << endl;
		rule_ = 0;
		return false;
	}

	switch (rule_->getNumberOfPostures()) {
	case 2:
		// Ok.
		break;
	case 3:
		if (currentPosture_ == 0 && transTimeMS > model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1) + epsilon) {
			phase_ = PHASE_TRANSITION_2;
			marker = 2;
			++currentPosture_;
		}
		break;
	case 4:
		if (currentPosture_ == 1 && transTimeMS > model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2) + epsilon) {
			phase_ = PHASE_TRANSITION_3;
			marker = 3;
			++currentPosture_;
		} else if (currentPosture_ == 0 && transTimeMS > model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1) + epsilon) {
			phase_ = PHASE_TRANSITION_2;
			marker = 2;
			++currentPosture_;
		}
		break;
	}

	if (paramLogger_.get() && marker > 0) {
		paramLogger_->logMarker(marker, postureList_[currentPosture_]->name());
	}

	param.glotPitch = getTransitionValue(Parameter::PARAMETER_GLOT_PITCH, transTimeMS);
	param.glotVol   = getTransitionValue(Parameter::PARAMETER_GLOT_VOL  , transTimeMS);
	param.aspVol    = getTransitionValue(Parameter::PARAMETER_ASP_VOL   , transTimeMS);
	param.fricVol   = getTransitionValue(Parameter::PARAMETER_FRIC_VOL  , transTimeMS);
	param.fricPos   = getTransitionValue(Parameter::PARAMETER_FRIC_POS  , transTimeMS);
	param.fricCF    = getTransitionValue(Parameter::PARAMETER_FRIC_CF   , transTimeMS);
	param.fricBW    = getTransitionValue(Parameter::PARAMETER_FRIC_BW   , transTimeMS);
	param.radius[0] = getTransitionValue(Parameter::PARAMETER_R1        , transTimeMS);
	param.radius[1] = getTransitionValue(Parameter::PARAMETER_R2        , transTimeMS);
	param.radius[2] = getTransitionValue(Parameter::PARAMETER_R3        , transTimeMS);
	param.radius[3] = getTransitionValue(Parameter::PARAMETER_R4        , transTimeMS);
	param.radius[4] = getTransitionValue(Parameter::PARAMETER_R5        , transTimeMS);
	param.radius[5] = getTransitionValue(Parameter::PARAMETER_R6        , transTimeMS);
	param.radius[6] = getTransitionValue(Parameter::PARAMETER_R7        , transTimeMS);
	param.radius[7] = getTransitionValue(Parameter::PARAMETER_R8        , transTimeMS);
	param.velum     = getTransitionValue(Parameter::PARAMETER_VELUM     , transTimeMS);

	if (paramLogger_.get()) {
		paramLogger_->logParam(param);
	}

	modifyControlParameters(param);

	//model_.clipParameterValue(param.glotPitch, Parameter::PARAMETER_GLOT_PITCH);
	model_.clipParameterValue(param.glotVol  , Parameter::PARAMETER_GLOT_VOL);
	model_.clipParameterValue(param.aspVol   , Parameter::PARAMETER_ASP_VOL);
	model_.clipParameterValue(param.fricVol  , Parameter::PARAMETER_FRIC_VOL);
	model_.clipParameterValue(param.fricPos  , Parameter::PARAMETER_FRIC_POS);
	model_.clipParameterValue(param.fricCF   , Parameter::PARAMETER_FRIC_CF);
	model_.clipParameterValue(param.fricBW   , Parameter::PARAMETER_FRIC_BW);
	model_.clipParameterValue(param.radius[0], Parameter::PARAMETER_R1);
	model_.clipParameterValue(param.radius[1], Parameter::PARAMETER_R2);
	model_.clipParameterValue(param.radius[2], Parameter::PARAMETER_R3);
	model_.clipParameterValue(param.radius[3], Parameter::PARAMETER_R4);
	model_.clipParameterValue(param.radius[4], Parameter::PARAMETER_R5);
	model_.clipParameterValue(param.radius[5], Parameter::PARAMETER_R6);
	model_.clipParameterValue(param.radius[6], Parameter::PARAMETER_R7);
	model_.clipParameterValue(param.radius[7], Parameter::PARAMETER_R8);
	model_.clipParameterValue(param.velum    , Parameter::PARAMETER_VELUM);

	return true;
}

/*******************************************************************************
 *
 */
void
ArticulationController::modifyControlParameters(ControlParameters& param) {

	// Modify the glotPitch.
	param.glotPitch = pitchFilter_.filter(
			(param.glotPitch + pitchProcessor_.getNext()) * configuration_.pitchFactor
			 + configuration_.pitchOffset);

	// Modify radiuses.
	param.radius[0] *= configuration_.radius1Factor;
	param.radius[1] *= configuration_.radius2Factor;
	param.radius[2] *= configuration_.radius3Factor;
	param.radius[3] *= configuration_.radius4Factor;
	param.radius[4] *= configuration_.radius5Factor;
	param.radius[5] *= configuration_.radius6Factor;
	param.radius[6] *= configuration_.radius7Factor;
	param.radius[7] *= configuration_.radius8Factor;
	for (int i = 0; i < 8; ++i) {
		if (param.radius[i] < configuration_.minRadius) {
			param.radius[i] = configuration_.minRadius;
		}
	}

	// Modify the aspiration.
	param.aspVol *= configuration_.aspirationFactor;

	// Modify the frication.
	param.fricVol *= configuration_.fricationFactor;
	param.fricCF *= configuration_.fricationFrequencyFactor;
	param.fricBW *= configuration_.fricationBandwidthFactor;
	param.fricPos = fricPosFilter_.filter(param.fricPos);
	//param.fricCF = fricCFFilter_.filter(param.fricCF) * configuration_.fricationFrequencyFactor;
	//param.fricBW = fricBWFilter_.filter(param.fricBW) * configuration_.fricationBandwidthFactor;
}

/*******************************************************************************
 *
 */
void
ArticulationController::printInputFileInfo()
{
	//---------------------------------------------------------
	// Input file.
	typedef InputFile::InputCommandList CommandList;
	typedef InputFile::InputCommand::FrequencyTargetList FrequencyList;
	std::cout << std::string(40, '-') << "\nInput file (after rewriting):\n\n";
	const CommandList& cmdList = inFile_.cmdList();
	for (CommandList::const_iterator iter = cmdList.begin(); iter != cmdList.end(); ++iter) {
		std::cout << "[" << iter->phoneme << "] dur:" << iter->duration << '\n';
		const FrequencyList& freqList = iter->frequencyTargetList;
		for (FrequencyList::const_iterator itFreq = freqList.begin(); itFreq != freqList.end(); ++itFreq) {
			std::cout << "  freq pos:" << itFreq->position << " target:" << itFreq->frequency << '\n';
		}
	}
}

/*******************************************************************************
 * Gets the next parameter values, and returns false if and only if there is
 * no more data.
 */
bool
ArticulationController::nextControlParameters(ControlParameters& param)
{
	if (isFinished()) return false;

	if (nextControlParametersUsingCurrentRule(param)) return true;

	if (isFinished()) return false;

	nextRule();

	// Try again.
	if (nextControlParametersUsingCurrentRule(param)) return true;

	return false;
}

/*******************************************************************************
 *
 */
void
ArticulationController::process()
{
	while (fillParameterBuffer()) {
		for (size_t i = 1, size = paramBuf_.size(); i < size; ++i) {
			tubeModel_->loopController().reset(controlPeriod_, paramBuf_[i - 1], paramBuf_[i]);
			tubeModel_->synthesize();
		}
	}

	// Be sure to flush the SRC buffer.
	tubeModel_->flush();
}

/*******************************************************************************
 *
 */
bool
ArticulationController::fillParameterBuffer()
{
	assert(paramBuf_.size() != 1);

	if (paramBuf_.size() > 0) {
		// Copy the last value.
		paramBuf_[0] = paramBuf_[paramBuf_.size() - 1];
		paramBuf_.resize(1);
	} else {
		paramBuf_.resize(1);
		if (!nextControlParameters(paramBuf_[0])) {
			throw ControllerException("ArticulationController::fillParameterBuffer(): Error in the first control period.");
		}
	}

	size_t size = paramBuf_.size();
	bool mayContinue = true;
	while (mayContinue && size < PARAMETER_BUFFER_MAX_SIZE) {
		paramBuf_.resize(++size);

		mayContinue = nextControlParameters(paramBuf_[size - 1]);
	}
	if (!mayContinue) {
		// Couldn't get the last item: discard the element.
		paramBuf_.resize(--size);
	}

	return paramBuf_.size() > 1;
}

} /* namespace Cgstmad */
