/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef ARTICULATION_CONTROLLER_H_
#define ARTICULATION_CONTROLLER_H_

#include "main.h"

#include <memory>
#include <string>
#include <tr1/array>
#include <tr1/memory>
#include <vector>

#include "articulation_loop_controller.h"
#include "average_filter.h"
#include "input_file.h"
#include "model.h"
#include "parameter.h"
#include "pitch_processor.h"
#include "tube_model.h"
#ifdef HAVE_LIBSAMPLERATE
# include "ecl_sample_rate_converter.h"
#else
# include "sample_rate_converter.h"
#endif



namespace Cgstmad {

class ParametersLogger;
class ProgramConfiguration;
class SpecialTransitionProcessor;
class TransitionProcessor;
class Wavetable;



/*******************************************************************************
 *
 */
class ArticulationController {
public:
	ArticulationController(const std::string& configDirPath,
				const std::string& inputFilePath,
				const std::string& outputFilePath,
				const ProgramConfiguration& configuration,
				bool verbose);
	~ArticulationController();

	void process();
	const std::vector<float>& outputBuffer() const {
		return tubeModel_->outputBuffer();
	}
	float maximumOutputSampleValue() const {
		return tubeModel_->maximumOutputSampleValue();
	}
private:
	enum Phase {
		PHASE_STEADY_STATE,
		PHASE_TRANSITION_1,
		PHASE_TRANSITION_2,
		PHASE_TRANSITION_3
	};
	enum {
		PARAMETER_BUFFER_MAX_SIZE = 128
	};

	typedef std::vector<float> DurationList;
	typedef std::tr1::shared_ptr<TransitionProcessor> TransitionProcessor_ptr;
	typedef std::tr1::array<TransitionProcessor_ptr, Parameter::NUM_PARAMETERS> TransitionProcessorList;
	typedef std::tr1::shared_ptr<SpecialTransitionProcessor> SpecialTransitionProcessor_ptr;
	typedef std::tr1::array<SpecialTransitionProcessor_ptr, Parameter::NUM_PARAMETERS> SpecialTransitionProcessorList;
	typedef std::vector<ControlParameters> ParameterBuffer;

	ArticulationController(const ArticulationController&);
	ArticulationController& operator=(const ArticulationController&);

	void fillFormulaSymbolList();
	void preparePitchProcessor();
	void prepareTransitionProcessor();
	float getTransitionValue(Parameter::Code param, float time);
	void modifyControlParameters(ControlParameters& param);
	void printInputFileInfo();
	bool isFinished();
	void nextRule();
	bool nextControlParameters(ControlParameters& param);
	bool nextControlParametersUsingCurrentRule(ControlParameters& param);
	bool fillParameterBuffer();
#ifdef HAVE_LIBSAMPLERATE
	std::auto_ptr<TubeModel<ArticulationLoopController, Wavetable, ECLSampleRateConverter> > tubeModel_;
#else
	std::auto_ptr<TubeModel<ArticulationLoopController, Wavetable, SampleRateConverter> > tubeModel_;
#endif
	Phase phase_;
	const Rule* rule_;
	int time_;
	const ProgramConfiguration& configuration_;
	std::auto_ptr<ParametersLogger> paramLogger_;
	int controlPeriod_; // samples
	float controlRate_; // samples / second
	PostureTmpList::size_type currentPosture_;
	PostureTmpList postureList_;
	DurationList durationList_;
	PitchProcessor pitchProcessor_;
	AverageFilter<float> pitchFilter_;
	AverageFilter<float> fricPosFilter_;
//	AverageFilter<float> fricCFFilter_;
//	AverageFilter<float> fricBWFilter_;
	TransitionProcessorList transProcessorList_;
	SpecialTransitionProcessorList specialTransProcessorList_;
	InputFile inFile_;
	Model model_;
	ParameterBuffer paramBuf_;
};

} /* namespace Cgstmad */

#endif /*ARTICULATION_CONTROLLER_H_*/
