/***************************************************************************
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef ARTICULATION_LOOP_CONTROLLER_H_
#define ARTICULATION_LOOP_CONTROLLER_H_

#include "main.h"

#include "control_parameters.h"
#include "tube_model_constants.h"
#include "tube_parameters.h"
#include "util.h"



namespace Cgstmad {

/*******************************************************************************
 *
 */
class ArticulationLoopController {
public:
	ArticulationLoopController(float /*sampleRate*/, TubeParameters& tubeParameters)
			: tubeParameters_(tubeParameters)
			, remainingLoops_(0)
			, deltaGlotFreq_(0.0)
			, deltaGlotVol_(0.0)
			, deltaAspVol_(0.0)
			, deltaFricVol_(0.0)
			, deltaFricPos_(0.0)
			, deltaFricCF_(0.0)
			, deltaFricBW_(0.0)
			, deltaVelum_(0.0)
	{
		for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) deltaRadius_[i] = 0.0;
	}

	void reset(int period, const ControlParameters& base, const ControlParameters& target);
	bool isFinished();
	void next();
private:
	TubeParameters& tubeParameters_;
	int remainingLoops_;
	float deltaGlotFreq_;
	float deltaGlotVol_;
	float deltaAspVol_;
	float deltaFricVol_;
	float deltaFricPos_;
	float deltaFricCF_;
	float deltaFricBW_;
	float deltaRadius_[TRM_TOTAL_REGIONS];
	float deltaVelum_;
};



/*******************************************************************************
 *
 */
inline
void
ArticulationLoopController::reset(int period, const ControlParameters& base, const ControlParameters& target)
{
	const float invPeriod = 1.0f / period;

	const float baseGlotFreq   = Util::frequency(base.glotPitch);
	const float targetGlotFreq = Util::frequency(target.glotPitch);

	// Set current parameters.
	tubeParameters_.glotFreq = baseGlotFreq;
	tubeParameters_.glotVol  = base.glotVol;
	tubeParameters_.aspVol   = base.aspVol;
	tubeParameters_.fricVol  = base.fricVol;
	tubeParameters_.fricPos  = base.fricPos;
	tubeParameters_.fricCF   = base.fricCF;
	tubeParameters_.fricBW   = base.fricBW;
	for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) {
		tubeParameters_.radius[i] = base.radius[i];
	}
	tubeParameters_.velum    = base.velum;

	// Set deltas.
	deltaGlotFreq_ = (targetGlotFreq - baseGlotFreq) * invPeriod;
	deltaGlotVol_  = (target.glotVol - base.glotVol) * invPeriod;
	deltaAspVol_   = (target.aspVol  - base.aspVol ) * invPeriod;
	deltaFricVol_  = (target.fricVol - base.fricVol) * invPeriod;
	deltaFricPos_  = (target.fricPos - base.fricPos) * invPeriod;
	deltaFricCF_   = (target.fricCF  - base.fricCF ) * invPeriod;
	deltaFricBW_   = (target.fricBW  - base.fricBW ) * invPeriod;
	for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) {
		deltaRadius_[i] = (target.radius[i] - base.radius[i]) * invPeriod;
	}
	deltaVelum_    = (target.velum   - base.velum  ) * invPeriod;

	remainingLoops_ = period;
}

/*******************************************************************************
 *
 */
inline
bool
ArticulationLoopController::isFinished()
{
	return remainingLoops_ == 0;
}

/*******************************************************************************
 *
 */
inline
void
ArticulationLoopController::next()
{
	if (--remainingLoops_ == 0) return;

	tubeParameters_.glotFreq += deltaGlotFreq_;
	tubeParameters_.glotVol  += deltaGlotVol_;
	tubeParameters_.aspVol   += deltaAspVol_;
	tubeParameters_.fricVol  += deltaFricVol_;
	tubeParameters_.fricPos  += deltaFricPos_;
	tubeParameters_.fricCF   += deltaFricCF_;
	tubeParameters_.fricBW   += deltaFricBW_;
	for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) {
		tubeParameters_.radius[i] += deltaRadius_[i];
	}
	tubeParameters_.velum    += deltaVelum_;
}

} // namespace Cgstmad

#endif /* ARTICULATION_LOOP_CONTROLLER_H_ */
