/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "config_file.h"

#include <iostream>

#include "category.h"
#include "model.h"
#include "parameter.h"
#include "posture.h"
#include "text.h"



namespace Cgstmad {

using std::cerr;
using std::cout;
using std::endl;

#ifdef USE_XERCES

/*******************************************************************************
 * Constructor.
 */
ConfigFile::ConfigFile(Model& model, const std::string& filePath)
			: model_(model)
			, document_(0)
			, root_(0)
{
	try {
		parser_.reset(new XercesDOMParser());

		parser_->setCreateCommentNodes(false);
		parser_->setCreateEntityReferenceNodes(false);
		//parser_->setDoValidation(true); // default: false
		parser_->setIncludeIgnorableWhitespace(false);
		parser_->setLoadExternalDTD(false);

		errHandler_.reset(new LocalXmlErrorHandler());
		parser_->setErrorHandler(errHandler_.get());

		parser_->parse(filePath.c_str());

		document_ = parser_->getDocument();
		root_ = document_->getDocumentElement();

	} catch (const XMLException& e) {
		std::string msg;
		util_.convertFromXmlString(e.getMessage(), msg);
		throw ControllerException(ControllerError("Error in ConfigFile::ConfigFile(Model&, const std::string&): ") << msg);
	}

	util_.convertToXmlString("boolean-expression"  , booleanExpressionTagName_);
	util_.convertToXmlString("boolean-expressions" , booleanExpressionsTagName_);
	util_.convertToXmlString("categories"          , categoriesTagName_);
	util_.convertToXmlString("category"            , categoryTagName_);
	util_.convertToXmlString("category-ref"        , categoryRefTagName_);
	util_.convertToXmlString("equation"            , equationTagName_);
	util_.convertToXmlString("equation-group"      , equationGroupTagName_);
	util_.convertToXmlString("equations"           , equationsTagName_);
	util_.convertToXmlString("expression-symbols"  , expressionSymbolsTagName_);
	util_.convertToXmlString("minimum-durations"   , minimumDurationsTagName_);
	util_.convertToXmlString("parameter"           , parameterTagName_);
	util_.convertToXmlString("parameter-profiles"  , parameterProfilesTagName_);
	util_.convertToXmlString("parameters"          , parametersTagName_);
	util_.convertToXmlString("parameter-targets"   , parameterTargetsTagName_);
	util_.convertToXmlString("parameter-transition", parameterTransitionTagName_);
	util_.convertToXmlString("posture"             , postureTagName_);
	util_.convertToXmlString("point"               , pointTagName_);
	util_.convertToXmlString("points"              , pointsTagName_);
	util_.convertToXmlString("posture-categories"  , postureCategoriesTagName_);
	util_.convertToXmlString("postures"            , posturesTagName_);
	util_.convertToXmlString("rule"                , ruleTagName_);
	util_.convertToXmlString("rules"               , rulesTagName_);
	util_.convertToXmlString("special-profiles"    , specialProfilesTagName_);
	util_.convertToXmlString("special-transitions" , specialTransitionsTagName_);
	util_.convertToXmlString("symbol-equation"     , symbolEquationTagName_);
	util_.convertToXmlString("symbol-targets"      , symbolTargetsTagName_);
	util_.convertToXmlString("target"              , targetTagName_);
	util_.convertToXmlString("transition"          , transitionTagName_);
	util_.convertToXmlString("transition-group"    , transitionGroupTagName_);
	util_.convertToXmlString("transitions"         , transitionsTagName_);
}

/*******************************************************************************
 * Destructor.
 */
ConfigFile::~ConfigFile()
{
}

void
ConfigFile::parseCategories(const DOMNode* base)
{
	typedef CategoryCodeMap::iterator MI;
	typedef CategoryCodeMap::value_type VT;

	int n = 0;

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, categoryTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::pair<MI,bool> res = model_.categoryCodeMap_.insert(VT(name, ++n));
			if (!res.second) {
				throw ControllerException(ControllerError("Duplicate category: ") << name);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseParameters(const DOMNode* base)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, parameterTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::string minimum;
			util_.getAttributeValue(node, "minimum", minimum);
			std::string maximum;
			util_.getAttributeValue(node, "maximum", maximum);
			Parameter::setInfo(model_.parameterInfoArray_, model_.parameterCodeMap_,
					name, Text::parseString<float>(minimum), Text::parseString<float>(maximum));
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parsePostureSymbols(const DOMNode* base, Posture_ptr posture)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, targetTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::string value;
			util_.getAttributeValue(node, "value", value);
			if (name == "duration") {
				posture->symbols_.duration = Text::parseString<float>(value);
			} else if (name == "transition") {
				posture->symbols_.transition = Text::parseString<float>(value);
			} else if (name == "qssa") {
				posture->symbols_.qssa = Text::parseString<float>(value);
			} else if (name == "qssb") {
				posture->symbols_.qssb = Text::parseString<float>(value);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parsePostureCategories(const DOMNode* base, Posture_ptr posture)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, categoryRefTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			posture->categoryList_.push_back(Category(name, 0));
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parsePostureParameters(const DOMNode* base, Posture_ptr posture)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, targetTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::string value;
			util_.getAttributeValue(node, "value", value);
			Parameter::setValue(model_.parameterCodeMap_, posture->parameters_, name, Text::parseString<float>(value));
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parsePosture(const DOMNode* base)
{
	std::string symbol;
	util_.getAttributeValue(base, "symbol", symbol);
	Posture_ptr p(new Posture(symbol));
	insert(model_.postureMap_, p->name(), p);

	//if (Main::verbose()) cout << "=== Posture Symbol=" << p->name << endl;

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isElement(node)) {
			if (XmlUtil::isNamed(node, parameterTargetsTagName_.c_str())) {
				parsePostureParameters(node, p);
			} else if (XmlUtil::isNamed(node, postureCategoriesTagName_.c_str())) {
				parsePostureCategories(node, p);
			} else if (XmlUtil::isNamed(node, symbolTargetsTagName_.c_str())) {
				parsePostureSymbols(node, p);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parsePostures(const DOMNode* base)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, postureTagName_.c_str())) {
			parsePosture(node);
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseEquationsGroup(const DOMNode* base)
{
	typedef EquationMap::iterator MI;
	typedef EquationMap::value_type VT;

	std::string groupName;
	util_.getAttributeValue(base, "name", groupName);

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, equationTagName_.c_str())) {
			Equation_ptr p(new Equation());
			p->groupName = groupName;
			util_.getAttributeValue(node, "name", p->name);
			util_.getAttributeValue(node, "formula", p->formula);
			std::pair<MI,bool> res = model_.equationMap_.insert(VT(p->name, p));
			if (!res.second) {
				throw ControllerException(ControllerError("Duplicate equation: ") << p->name);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseEquations(const DOMNode* base)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, equationGroupTagName_.c_str())) {
			parseEquationsGroup(node);
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseTransitionPoint(const DOMNode* base, Transition_ptr transition, Transition::Point::Type type)
{
	Transition::Point_ptr p(new Transition::Point());
	p->type = type;
	std::string value;
	util_.getAttributeValue(base, "value", value);
	p->value = 1.0e-2f * Text::parseString<float>(value);
	util_.getAttributeValue(base, "time-expression", p->timeExpression);
	if (p->timeExpression.empty()) {
		std::string freeTime;
		util_.getAttributeValue(base, "free-time", freeTime);
		p->freeTime = Text::parseString<float>(freeTime);
	}
	transition->pointListList_[p->type == Transition::Point::TYPE_SPECIAL ? 0 : p->type].push_back(p);
}

void
ConfigFile::parseTransitionPoints(const DOMNode* base, Transition_ptr transition)
{
	std::string typeName;
	util_.getAttributeValue(base, "type", typeName);
	Transition::Point::Type type = Transition::Point::getTypeFromName(typeName);

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, pointTagName_.c_str())) {
			parseTransitionPoint(node, transition, type);
		}

		node = node->getNextSibling();
	}
}

void
ConfigFile::parseTransitionsGroup(const DOMNode* base, bool special)
{
	typedef TransitionMap::iterator MI;
	typedef TransitionMap::value_type VT;

	std::string groupName;
	util_.getAttributeValue(base, "name", groupName);

	TransitionMap& map = special ? model_.specialTransitionMap_ : model_.transitionMap_;

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, transitionTagName_.c_str())) {
			Transition_ptr p(new Transition(special));
			p->groupName_ = groupName;
			util_.getAttributeValue(node, "name", p->name_);
			std::string type;
			util_.getAttributeValue(node, "type", type);
			p->type_ = Transition::getTypeFromName(type);
			p->pointListList_.resize(special ? 1 : p->type_ + 1);

			std::pair<MI,bool> res = map.insert(VT(p->name_, p));
			if (!res.second) {
				cerr << "Duplicate transition: " << p->name_ << " (ignored)" << endl;
				//throw ControllerException(ControllerError("Duplicate transition: ") << p->name);
			} else {
				DOMNode* child = node->getFirstChild();
				while (child != 0) {
					if (XmlUtil::isElement(child)) {
						if (XmlUtil::isNamed(child, pointsTagName_.c_str())) {
							parseTransitionPoints(child, p);
						} else if (XmlUtil::isNamed(child, pointTagName_.c_str())) {
							// Special transition.
							parseTransitionPoint(child, p, Transition::Point::TYPE_SPECIAL);
						}
					}
					child = child->getNextSibling();
				}
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseTransitions(const DOMNode* base, bool special)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, transitionGroupTagName_.c_str())) {
			parseTransitionsGroup(node, special);
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRuleParameterProfiles(const DOMNode* base, Rule_ptr rule)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, parameterTransitionTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::string value;
			util_.getAttributeValue(node, "transition", value);
			Parameter::setValue(model_.parameterCodeMap_, rule->paramProfileTransitions_, name, value);
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRuleSpecialProfiles(const DOMNode* base, Rule_ptr rule)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, parameterTransitionTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			std::string value;
			util_.getAttributeValue(node, "transition", value);
			Parameter::setValue(model_.parameterCodeMap_, rule->specialProfileTransitions_, name, value);
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRuleExpressionSymbols(const DOMNode* base, Rule_ptr rule)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, symbolEquationTagName_.c_str())) {
			std::string name;
			util_.getAttributeValue(node, "name", name);
			if (name == "rd") {
				util_.getAttributeValue(node, "equation", rule->exprSymbolEquations_.ruleDuration);
			} else if (name == "mark1") {
				util_.getAttributeValue(node, "equation", rule->exprSymbolEquations_.mark1);
			} else if (name == "mark2") {
				util_.getAttributeValue(node, "equation", rule->exprSymbolEquations_.mark2);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRuleBooleanExpressions(const DOMNode* base, Rule_ptr rule)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, booleanExpressionTagName_.c_str())) {
			std::string text;
			util_.getText(node, text);
			rule->booleanExpressionList_.push_back(text);
			//if (Main::verbose()) cout << "[" << rule->booleanExpressionList.back() << "]" << endl;
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRuleMinDurations(const DOMNode* node, Rule_ptr rule)
{
	std::string duration;

	util_.getAttributeValue(node, "p12", duration);
	rule->minDuration12_ = Text::parseString<int>(duration);

	util_.getAttributeValue(node, "p23", duration);
	rule->minDuration23_ = duration.size() > 0 ? Text::parseString<int>(duration) : 0;

	util_.getAttributeValue(node, "p34", duration);
	rule->minDuration34_ = duration.size() > 0 ? Text::parseString<int>(duration) : 0;
}

void
ConfigFile::parseRule(const DOMNode* base)
{
	Rule_ptr p(new Rule());
	p->number_ = model_.ruleList_.size() + 1;
	model_.ruleList_.push_back(p);

	//if (Main::verbose()) cout << "=== Rule Number=" << p->number << endl;

	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isElement(node)) {
			if (XmlUtil::isNamed(node, parameterProfilesTagName_.c_str())) {
				parseRuleParameterProfiles(node, p);
			} else if (XmlUtil::isNamed(node, expressionSymbolsTagName_.c_str())) {
				parseRuleExpressionSymbols(node, p);
			} else if (XmlUtil::isNamed(node, specialProfilesTagName_.c_str())) {
				parseRuleSpecialProfiles(node, p);
			} else if (XmlUtil::isNamed(node, booleanExpressionsTagName_.c_str())) {
				parseRuleBooleanExpressions(node, p);
			} else if (XmlUtil::isNamed(node, minimumDurationsTagName_.c_str())) {
				parseRuleMinDurations(node, p);
			}
		}
		node = node->getNextSibling();
	}
}

void
ConfigFile::parseRules(const DOMNode* base)
{
	DOMNode* node = base->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isNamedElement(node, ruleTagName_.c_str())) {
			parseRule(node);
		}
		node = node->getNextSibling();
	}
}

/*******************************************************************************
 *
 */
void
ConfigFile::loadModel()
{
	DOMNode* node = root_->getFirstChild();
	while (node != 0) {
		if (XmlUtil::isElement(node)) {
			if (XmlUtil::isNamed(node, categoriesTagName_.c_str())) {
				//if (Main::verbose()) cout << "CATEGORIES" << endl;
				parseCategories(node);
			} else if (XmlUtil::isNamed(node, parametersTagName_.c_str())) {
				//if (Main::verbose()) cout << "PARAMETERS" << endl;
				parseParameters(node);
			} else if (XmlUtil::isNamed(node, posturesTagName_.c_str())) {
				//if (Main::verbose()) cout << "POSTURES" << endl;
				parsePostures(node);
			} else if (XmlUtil::isNamed(node, equationsTagName_.c_str())) {
				//if (Main::verbose()) cout << "EQUATIONS" << endl;
				parseEquations(node);
			} else if (XmlUtil::isNamed(node, transitionsTagName_.c_str())) {
				//if (Main::verbose()) cout << "TRANSITIONS" << endl;
				parseTransitions(node, false);
			} else if (XmlUtil::isNamed(node, specialTransitionsTagName_.c_str())) {
				//if (Main::verbose()) cout << "SPECIAL TRANS" << endl;
				parseTransitions(node, true);
			} else if (XmlUtil::isNamed(node, rulesTagName_.c_str())) {
				//if (Main::verbose()) cout << "RULES" << endl;
				parseRules(node);
			}
		}
		node = node->getNextSibling();
	}
}

#else /* USE_XERCES */

void
ConfigFile::parseCategories()
{
	typedef CategoryCodeMap::iterator MI;
	typedef CategoryCodeMap::value_type VT;

	int n = 0;
	for (const std::string* category = parser_.getFirstChild(categoryTagName_);
				category;
				category = parser_.getNextSibling(categoryTagName_)) {
		const std::string& name = parser_.getAttribute(nameAttrName_);
		std::pair<MI,bool> res = model_.categoryCodeMap_.insert(VT(name, ++n));
		if (!res.second) {
			throw ControllerException(ControllerError("Duplicate category: ") << name);
		}

		if (parser_.getFirstChild()) { // skip comment
			parser_.getNextSibling();
		}
	}
}

void
ConfigFile::parseParameters()
{
	for (const std::string* parameter = parser_.getFirstChild(parameterTagName_);
				parameter;
				parameter = parser_.getNextSibling(parameterTagName_)) {
		Parameter::setInfo(
			model_.parameterInfoArray_,
			model_.parameterCodeMap_,
			parser_.getAttribute(nameAttrName_),
			Text::parseString<float>(parser_.getAttribute(minimumAttrName_)),
			Text::parseString<float>(parser_.getAttribute(maximumAttrName_)));
	}
}

void
ConfigFile::parseSymbols()
{
	for (const std::string* symbol = parser_.getFirstChild(symbolTagName_);
				symbol;
				symbol = parser_.getNextSibling(symbolTagName_)) {
		// Ignore the attributes.
	}
}

void
ConfigFile::parsePostureSymbols(Posture_ptr posture)
{
	for (const std::string* target = parser_.getFirstChild(targetTagName_);
				target;
				target = parser_.getNextSibling(targetTagName_)) {
		const std::string& name = parser_.getAttribute(nameAttrName_);
		const std::string& value = parser_.getAttribute(valueAttrName_);
		if (name == durationSymbolName_) {
			posture->symbols_.duration = Text::parseString<float>(value);
		} else if (name == transitionSymbolName_) {
			posture->symbols_.transition = Text::parseString<float>(value);
		} else if (name == qssaSymbolName_) {
			posture->symbols_.qssa = Text::parseString<float>(value);
		} else if (name == qssbSymbolName_) {
			posture->symbols_.qssb = Text::parseString<float>(value);
		}
	}
}

void
ConfigFile::parsePostureCategories(Posture_ptr posture)
{
	for (const std::string* catRef = parser_.getFirstChild(categoryRefTagName_);
				catRef;
				catRef = parser_.getNextSibling(categoryRefTagName_)) {
		posture->categoryList_.push_back(Category(parser_.getAttribute(nameAttrName_), 0));
	}
}

void
ConfigFile::parsePostureParameters(Posture_ptr posture)
{
	for (const std::string* target = parser_.getFirstChild(targetTagName_);
				target;
				target = parser_.getNextSibling(targetTagName_)) {
		Parameter::setValue(
			model_.parameterCodeMap_,
			posture->parameters_,
			parser_.getAttribute(nameAttrName_),
			Text::parseString<float>(parser_.getAttribute(valueAttrName_)));
	}
}

void
ConfigFile::parsePosture()
{
	Posture_ptr p(new Posture(parser_.getAttribute(symbolAttrName_)));
	insert(model_.postureMap_, p->name_, p);

	//if (Main::verbose()) cout << "=== Posture Symbol=" << p->name << endl;

	for (const std::string* child = parser_.getFirstChild();
				child;
				child = parser_.getNextSibling()) {
		if (*child == postureCategoriesTagName_) {
			parsePostureCategories(p);
		} else if (*child == parameterTargetsTagName_) {
			parsePostureParameters(p);
		} else if (*child == symbolTargetsTagName_) {
			parsePostureSymbols(p);
		}
	}
}

void
ConfigFile::parsePostures()
{
	for (const std::string* posture = parser_.getFirstChild(postureTagName_);
				posture;
				posture = parser_.getNextSibling(postureTagName_)) {
		parsePosture();
	}
}

void
ConfigFile::parseEquationsGroup()
{
	typedef EquationMap::iterator MI;
	typedef EquationMap::value_type VT;

	const std::string groupName = parser_.getAttribute(nameAttrName_);
	// Can't be a reference because it is used many times in the loop.

	for (const std::string* equation = parser_.getFirstChild(equationTagName_);
				equation;
				equation = parser_.getNextSibling(equationTagName_)) {
		Equation_ptr p(new Equation());
		p->groupName = groupName;
		p->name = parser_.getAttribute(nameAttrName_);
		p->formula = parser_.getAttribute(formulaAttrName_);
		std::pair<MI,bool> res = model_.equationMap_.insert(VT(p->name, p));
		if (!res.second) {
			throw ControllerException(ControllerError("Duplicate equation: ") << p->name);
		}

		if (parser_.getFirstChild()) {
			parser_.getNextSibling(); // skip comment
		}
	}
}

void
ConfigFile::parseEquations()
{
	for (const std::string* group = parser_.getFirstChild(equationGroupTagName_);
				group;
				group = parser_.getNextSibling(equationGroupTagName_)) {
		parseEquationsGroup();
	}
}

void
ConfigFile::parseTransitionPoint(Transition_ptr transition, Transition::Point::Type pointType)
{
	Transition::Point_ptr p(new Transition::Point());
	p->type = pointType;
	p->value = 1.0e-2f * Text::parseString<float>(parser_.getAttribute(valueAttrName_));
	p->timeExpression = parser_.getAttribute(timeExpressionAttrName_);
	if (p->timeExpression.empty()) {
		p->freeTime = Text::parseString<float>(parser_.getAttribute(freeTimeAttrName_));
	}
	transition->pointListList_[p->type == Transition::Point::TYPE_SPECIAL ? 0 : p->type].push_back(p);
}

void
ConfigFile::parseTransitionPoints(Transition_ptr transition)
{
	Transition::Point::Type pointType = Transition::Point::getTypeFromName(parser_.getAttribute(typeAttrName_));

	for (const std::string* child = parser_.getFirstChild(pointTagName_);
				child;
				child = parser_.getNextSibling(pointTagName_)) {
		parseTransitionPoint(transition, pointType);
	}
}

void
ConfigFile::parseTransitionsGroup(bool special)
{
	typedef TransitionMap::iterator MI;
	typedef TransitionMap::value_type VT;

	const std::string groupName = parser_.getAttribute(nameAttrName_);
	// Can't be a reference because it is used many times in the loop.

	TransitionMap& map = special ? model_.specialTransitionMap_ : model_.transitionMap_;

	for (const std::string* transition = parser_.getFirstChild(transitionTagName_);
				transition;
				transition = parser_.getNextSibling(transitionTagName_)) {
		Transition_ptr p(new Transition(special));
		p->groupName_ = groupName;
		p->name_ = parser_.getAttribute(nameAttrName_);
		p->type_ = Transition::getTypeFromName(parser_.getAttribute(typeAttrName_));
		p->pointListList_.resize(special ? 1 : p->type_ + 1);

 		for (const std::string* child = parser_.getFirstChild();
 					child;
 					child = parser_.getNextSibling()) {
			if (*child == pointsTagName_) {
				parseTransitionPoints(p);
			} else if (*child == pointTagName_) {
				// Special transition.
				parseTransitionPoint(p, Transition::Point::TYPE_SPECIAL);
			}
		}

		std::pair<MI,bool> res = map.insert(VT(p->name_, p));
		if (!res.second) {
			cerr << "Duplicate transition: " << p->name_ << " (ignored)" << endl;
			//throw ControllerException(ControllerError("Duplicate transition: ") << p->name);
		}
	}
}

void
ConfigFile::parseTransitions(bool special)
{
	for (const std::string* group = parser_.getFirstChild(transitionGroupTagName_);
				group;
				group = parser_.getNextSibling(transitionGroupTagName_)) {
		parseTransitionsGroup(special);
	}
}

void
ConfigFile::parseRuleParameterProfiles(Rule_ptr rule)
{
	for (const std::string* paramTrans = parser_.getFirstChild(parameterTransitionTagName_);
				paramTrans;
				paramTrans = parser_.getNextSibling(parameterTransitionTagName_)) {
		Parameter::setValue(
			model_.parameterCodeMap_,
			rule->paramProfileTransitions_,
			parser_.getAttribute(nameAttrName_),
			parser_.getAttribute(transitionAttrName_));
	}
}

void
ConfigFile::parseRuleSpecialProfiles(Rule_ptr rule)
{
	for (const std::string* paramTrans = parser_.getFirstChild(parameterTransitionTagName_);
				paramTrans;
				paramTrans = parser_.getNextSibling(parameterTransitionTagName_)) {
		Parameter::setValue(
			model_.parameterCodeMap_,
			rule->specialProfileTransitions_,
			parser_.getAttribute(nameAttrName_),
			parser_.getAttribute(transitionAttrName_));
	}
}

void
ConfigFile::parseRuleExpressionSymbols(Rule_ptr rule)
{
	for (const std::string* symbEqu = parser_.getFirstChild(symbolEquationTagName_);
				symbEqu;
				symbEqu = parser_.getNextSibling(symbolEquationTagName_)) {
		const std::string& name = parser_.getAttribute(nameAttrName_);
		const std::string& equation = parser_.getAttribute(equationAttrName_);
		if (name == rdSymbolName_) {
			rule->exprSymbolEquations_.ruleDuration = equation;
		} else if (name == mark1SymbolName_) {
			rule->exprSymbolEquations_.mark1 = equation;
		} else if (name == mark2SymbolName_) {
			rule->exprSymbolEquations_.mark2 = equation;
		}
	}
}

void
ConfigFile::parseRuleBooleanExpressions(Rule_ptr rule)
{
	for (const std::string* boolExpr = parser_.getFirstChild(booleanExpressionTagName_);
				boolExpr;
				boolExpr = parser_.getNextSibling(booleanExpressionTagName_)) {
		rule->booleanExpressionList_.push_back(parser_.getText());
//		if (Main::verbose()) cout << "[" << rule->booleanExpressionList.back() << "]" << endl;
	}
}

void
ConfigFile::parseRuleMinDurations(Rule_ptr rule)
{
	rule->minDuration12_ = Text::parseString<int>(parser_.getAttribute(p12AttrName_));

	const std::string& p23 = parser_.getAttribute(p23AttrName_);
	rule->minDuration23_ = p23.size() > 0 ? Text::parseString<int>(p23) : 0;

	const std::string& p34 = parser_.getAttribute(p34AttrName_);
	rule->minDuration34_ = p34.size() > 0 ? Text::parseString<int>(p34) : 0;
}

void
ConfigFile::parseRule()
{
	Rule_ptr p(new Rule());
	p->number_ = model_.ruleList_.size() + 1;
	model_.ruleList_.push_back(p);

//	if (Main::verbose()) cout << "=== Rule Number=" << p->number << endl;

	for (const std::string* child = parser_.getFirstChild();
				child;
				child = parser_.getNextSibling()) {
		if (*child == booleanExpressionsTagName_) {
			parseRuleBooleanExpressions(p);
		} else if (*child == parameterProfilesTagName_) {
			parseRuleParameterProfiles(p);
		} else if (*child == specialProfilesTagName_) {
			parseRuleSpecialProfiles(p);
		} else if (*child == expressionSymbolsTagName_) {
			parseRuleExpressionSymbols(p);
		} else if (*child == minimumDurationsTagName_) {
			parseRuleMinDurations(p);
		}
	}
}

void
ConfigFile::parseRules()
{
	for (const std::string* rule = parser_.getFirstChild(ruleTagName_);
				rule;
				rule = parser_.getNextSibling(ruleTagName_)) {
		parseRule();
	}
}



/*******************************************************************************
 * Constructor.
 */
ConfigFile::ConfigFile(Model& model, const std::string& filePath)
		: model_(model)

		, booleanExpressionTagName_  ("boolean-expression")
		, booleanExpressionsTagName_ ("boolean-expressions")
		, categoriesTagName_         ("categories")
		, categoryTagName_           ("category")
		, categoryRefTagName_        ("category-ref")
		, equationTagName_           ("equation")
		, equationGroupTagName_      ("equation-group")
		, equationsTagName_          ("equations")
		, expressionSymbolsTagName_  ("expression-symbols")
		, minimumDurationsTagName_   ("minimum-durations")
		, parameterTagName_          ("parameter")
		, parameterProfilesTagName_  ("parameter-profiles")
		, parametersTagName_         ("parameters")
		, parameterTargetsTagName_   ("parameter-targets")
		, parameterTransitionTagName_("parameter-transition")
		, postureTagName_            ("posture")
		, pointTagName_              ("point")
		, pointsTagName_             ("points")
		, postureCategoriesTagName_  ("posture-categories")
		, posturesTagName_           ("postures")
		, ruleTagName_               ("rule")
		, rulesTagName_              ("rules")
		, specialProfilesTagName_    ("special-profiles")
		, specialTransitionsTagName_ ("special-transitions")
		, symbolEquationTagName_     ("symbol-equation")
		, symbolsTagName_            ("symbols")
		, symbolTagName_             ("symbol")
		, symbolTargetsTagName_      ("symbol-targets")
		, targetTagName_             ("target")
		, transitionTagName_         ("transition")
		, transitionGroupTagName_    ("transition-group")
		, transitionsTagName_        ("transitions")

		, equationAttrName_      ("equation")
		, formulaAttrName_       ("formula")
		, freeTimeAttrName_      ("free-time")
		, maximumAttrName_       ("maximum")
		, minimumAttrName_       ("minimum")
		, nameAttrName_          ("name")
		, p12AttrName_           ("p12")
		, p23AttrName_           ("p23")
		, p34AttrName_           ("p34")
		, symbolAttrName_        ("symbol")
		, timeExpressionAttrName_("time-expression")
		, typeAttrName_          ("type")
		, transitionAttrName_    ("transition")
		, valueAttrName_         ("value")

		, durationSymbolName_  ("duration")
		, mark1SymbolName_     ("mark1")
		, mark2SymbolName_     ("mark2")
		, qssaSymbolName_      ("qssa")
		, qssbSymbolName_      ("qssb")
		, rdSymbolName_        ("rd")
		, transitionSymbolName_("transition")

		, parser_(filePath)
{
	const std::string* root = parser_.getFirstChild();
	if (root == 0) {
		throw ControllerException("ConfigFile::ConfigFile(Model&, const std::string&): Root element not found.");
	}
}

/*******************************************************************************
 * Destructor.
 */
ConfigFile::~ConfigFile()
{
}

/*******************************************************************************
 *
 */
void
ConfigFile::loadModel()
{
//	if (Main::verbose()) cout << "categories" << endl;
	if (parser_.getFirstChild(categoriesTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Categories element not found.");
	}
	parseCategories();

//	if (Main::verbose()) cout << "parameters" << endl;
	if (parser_.getNextSibling(parametersTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Parameters element not found.");
	}
	parseParameters();

//	if (Main::verbose()) cout << "symbols" << endl;
	if (parser_.getNextSibling(symbolsTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Categories element not found.");
	}
	parseSymbols();

//	if (Main::verbose()) cout << "postures" << endl;
	if (parser_.getNextSibling(posturesTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Postures element not found.");
	}
	parsePostures();

//	if (Main::verbose()) cout << "equations" << endl;
	if (parser_.getNextSibling(equationsTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Equations element not found.");
	}
	parseEquations();

//	if (Main::verbose()) cout << "transitions" << endl;
	if (parser_.getNextSibling(transitionsTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Transitions element not found.");
	}
	parseTransitions(false);

//	if (Main::verbose()) cout << "special-transitions" << endl;
	if (parser_.getNextSibling(specialTransitionsTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Special-transitions element not found.");
	}
	parseTransitions(true);

//	if (Main::verbose()) cout << "rules" << endl;
	if (parser_.getNextSibling(rulesTagName_) == 0) {
		throw ControllerException("ConfigFile::loadModel(): Rules element not found.");
	}
	parseRules();
}

#endif /* USE_XERCES */

} /* namespace Cgstmad */
