/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef CONFIG_FILE_H_
#define CONFIG_FILE_H_

#include "main.h"

#ifdef USE_XERCES
# include <xercesc/dom/DOM.hpp>
# include <xercesc/parsers/XercesDOMParser.hpp>
# include "xml_util.h"
#else
# include <string>
# include "simple_xml_parser.h"
#endif

#include "equation.h"
#include "posture.h"
#include "rule.h"
#include "transition.h"



namespace Cgstmad {

#ifdef USE_XERCES

/*******************************************************************************
 *
 */
class ConfigFile {
public:
	ConfigFile(Model& model, const std::string& filePath);
	~ConfigFile();

	void loadModel();
private:
	Model& model_;
	std::auto_ptr<XercesDOMParser> parser_;
	std::auto_ptr<LocalXmlErrorHandler> errHandler_;
	const DOMDocument* document_;
	const DOMElement* root_;
	XmlUtil util_;

	XmlString booleanExpressionTagName_;
	XmlString booleanExpressionsTagName_;
	XmlString categoriesTagName_;
	XmlString categoryTagName_;
	XmlString categoryRefTagName_;
	XmlString equationTagName_;
	XmlString equationGroupTagName_;
	XmlString equationsTagName_;
	XmlString expressionSymbolsTagName_;
	XmlString minimumDurationsTagName_;
	XmlString parameterTagName_;
	XmlString parameterProfilesTagName_;
	XmlString parametersTagName_;
	XmlString parameterTargetsTagName_;
	XmlString parameterTransitionTagName_;
	XmlString postureTagName_;
	XmlString pointTagName_;
	XmlString pointsTagName_;
	XmlString postureCategoriesTagName_;
	XmlString posturesTagName_;
	XmlString ruleTagName_;
	XmlString rulesTagName_;
	XmlString specialProfilesTagName_;
	XmlString specialTransitionsTagName_;
	XmlString symbolEquationTagName_;
	XmlString symbolTargetsTagName_;
	XmlString targetTagName_;
	XmlString transitionTagName_;
	XmlString transitionGroupTagName_;
	XmlString transitionsTagName_;

	ConfigFile(const ConfigFile&);
	ConfigFile& operator=(const ConfigFile&);

	void parseCategories(const DOMNode* base);

	void parseParameters(const DOMNode* base);

	void parsePostureSymbols(const DOMNode* base, Posture_ptr posture);
	void parsePostureCategories(const DOMNode* base, Posture_ptr posture);
	void parsePostureParameters(const DOMNode* base, Posture_ptr posture);
	void parsePosture(const DOMNode* base);
	void parsePostures(const DOMNode* base);

	void parseEquationsGroup(const DOMNode* base);
	void parseEquations(const DOMNode* base);

	void parseTransitionPoint(const DOMNode* base, Transition_ptr transition, Transition::Point::Type type);
	void parseTransitionPoints(const DOMNode* base, Transition_ptr transition);
	void parseTransitionsGroup(const DOMNode* base, bool special);
	void parseTransitions(const DOMNode* base, bool special);

	void parseRuleParameterProfiles(const DOMNode* base, Rule_ptr rule);
	void parseRuleSpecialProfiles(const DOMNode* base, Rule_ptr rule);
	void parseRuleExpressionSymbols(const DOMNode* base, Rule_ptr rule);
	void parseRuleBooleanExpressions(const DOMNode* base, Rule_ptr rule);
	void parseRuleMinDurations(const DOMNode* node, Rule_ptr rule);
	void parseRule(const DOMNode* base);
	void parseRules(const DOMNode* base);
};

#else /* USE_XERCES */

class Model;

/*******************************************************************************
 *
 */
class ConfigFile {
public:
	ConfigFile(Model& model, const std::string& filePath);
	~ConfigFile();

	void loadModel();
private:
	Model& model_;

	std::string booleanExpressionTagName_;
	std::string booleanExpressionsTagName_;
	std::string categoriesTagName_;
	std::string categoryTagName_;
	std::string categoryRefTagName_;
	std::string equationTagName_;
	std::string equationGroupTagName_;
	std::string equationsTagName_;
	std::string expressionSymbolsTagName_;
	std::string minimumDurationsTagName_;
	std::string parameterTagName_;
	std::string parameterProfilesTagName_;
	std::string parametersTagName_;
	std::string parameterTargetsTagName_;
	std::string parameterTransitionTagName_;
	std::string postureTagName_;
	std::string pointTagName_;
	std::string pointsTagName_;
	std::string postureCategoriesTagName_;
	std::string posturesTagName_;
	std::string ruleTagName_;
	std::string rulesTagName_;
	std::string specialProfilesTagName_;
	std::string specialTransitionsTagName_;
	std::string symbolEquationTagName_;
	std::string symbolsTagName_;
	std::string symbolTagName_;
	std::string symbolTargetsTagName_;
	std::string targetTagName_;
	std::string transitionTagName_;
	std::string transitionGroupTagName_;
	std::string transitionsTagName_;

	std::string equationAttrName_;
	std::string formulaAttrName_;
	std::string freeTimeAttrName_;
	std::string maximumAttrName_;
	std::string minimumAttrName_;
	std::string nameAttrName_;
	std::string p12AttrName_;
	std::string p23AttrName_;
	std::string p34AttrName_;
	std::string symbolAttrName_;
	std::string timeExpressionAttrName_;
	std::string typeAttrName_;
	std::string transitionAttrName_;
	std::string valueAttrName_;

	std::string durationSymbolName_;
        std::string mark1SymbolName_;
        std::string mark2SymbolName_;
        std::string qssaSymbolName_;
        std::string qssbSymbolName_;
        std::string rdSymbolName_;
	std::string transitionSymbolName_;

	SimpleXmlParser parser_;

	ConfigFile(const ConfigFile&);
	ConfigFile& operator=(const ConfigFile&);

	void parseCategories();

	void parseParameters();

	void parseSymbols();

	void parsePostureSymbols(Posture_ptr posture);
	void parsePostureCategories(Posture_ptr posture);
	void parsePostureParameters(Posture_ptr posture);
	void parsePosture();
	void parsePostures();

	void parseEquationsGroup();
	void parseEquations();

	void parseTransitionPoint(Transition_ptr transition, Transition::Point::Type pointType);
	void parseTransitionPoints(Transition_ptr transition);
	void parseTransitionsGroup(bool special);
	void parseTransitions(bool special);

	void parseRuleParameterProfiles(Rule_ptr list);
	void parseRuleSpecialProfiles(Rule_ptr list);
	void parseRuleExpressionSymbols(Rule_ptr list);
	void parseRuleBooleanExpressions(Rule_ptr rule);
	void parseRuleMinDurations(Rule_ptr rule);
	void parseRule();
	void parseRules();
};

#endif /* USE_XERCES */

} /* namespace Cgstmad */

#endif /*CONFIG_FILE_H_*/
