/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef EQUATION_H_
#define EQUATION_H_

#include "main.h"

#include <iostream>
#include <string>
#include <tr1/memory>
#include <tr1/unordered_map>

#include "formula_symbol.h"



namespace Cgstmad {

class FormulaNode {
public:
	virtual ~FormulaNode() {}

	virtual float eval(const FormulaSymbolList& symbolList) const = 0;
	virtual void print(std::ostream& out, int level = 0) const = 0;
};

typedef std::tr1::shared_ptr<FormulaNode> FormulaNode_ptr;

class FormulaMinusUnaryOp : public FormulaNode {
public:
	FormulaMinusUnaryOp(FormulaNode_ptr c)
			: FormulaNode(), child_(c) {}
	virtual ~FormulaMinusUnaryOp() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaNode_ptr child_;
};

class FormulaAddBinaryOp : public FormulaNode {
public:
	FormulaAddBinaryOp(FormulaNode_ptr c1, FormulaNode_ptr c2)
			: FormulaNode(), child1_(c1), child2_(c2) {}
	virtual ~FormulaAddBinaryOp() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaNode_ptr child1_;
	FormulaNode_ptr child2_;
};

class FormulaSubBinaryOp : public FormulaNode {
public:
	FormulaSubBinaryOp(FormulaNode_ptr c1, FormulaNode_ptr c2)
			: FormulaNode(), child1_(c1), child2_(c2) {}
	virtual ~FormulaSubBinaryOp() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaNode_ptr child1_;
	FormulaNode_ptr child2_;
};

class FormulaMultBinaryOp : public FormulaNode {
public:
	FormulaMultBinaryOp(FormulaNode_ptr c1, FormulaNode_ptr c2)
			: FormulaNode(), child1_(c1), child2_(c2) {}
	virtual ~FormulaMultBinaryOp() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaNode_ptr child1_;
	FormulaNode_ptr child2_;
};

class FormulaDivBinaryOp : public FormulaNode {
public:
	FormulaDivBinaryOp(FormulaNode_ptr c1, FormulaNode_ptr c2)
			: FormulaNode(), child1_(c1), child2_(c2) {}
	virtual ~FormulaDivBinaryOp() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaNode_ptr child1_;
	FormulaNode_ptr child2_;
};

class FormulaConst : public FormulaNode {
public:
	FormulaConst(float value)
			: FormulaNode(), value_(value) {}
	virtual ~FormulaConst() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	float value_;
};

class FormulaSymbolValue : public FormulaNode {
public:
	FormulaSymbolValue(FormulaSymbol::Code symbol)
			: FormulaNode(), symbol_(symbol) {}
	virtual ~FormulaSymbolValue() {}

	virtual float eval(const FormulaSymbolList& symbolList) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	FormulaSymbol::Code symbol_;

};

std::ostream&
operator<<(std::ostream& out, const FormulaNode& node);

struct Equation {
	std::string name;
	std::string formula;
	std::string groupName;
	FormulaNode_ptr formulaRoot;

	Equation() : name(), formula(), groupName(), formulaRoot() {}

	void parseFormula(const FormulaSymbol& formulaSymbol);
	float evalFormula(const FormulaSymbolList& symbolList) const;
};

typedef std::tr1::shared_ptr<Equation> Equation_ptr;
typedef std::tr1::unordered_map<std::string, Equation_ptr> EquationMap; // type of container that manages the Equation instances

} /* namespace Cgstmad */

#endif /*EQUATION_H_*/
