/***************************************************************************
 *  Copyright 2007 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "input_file.h"

#include <fstream>

#include "controller_exception.h"
#include "phoneme_rewriter.h"
#include "posture_rewriter.h"
#include "text.h"



namespace {

using namespace Cgstmad;

void
throwException(const std::string& filePath, int lineNumber, const char* message)
{
	throw ControllerException(ControllerError("Error in InputFile: ")
					<< filePath << " Line " << lineNumber << ": " << message);
}

}

//==============================================================================

namespace Cgstmad {

using std::cout;
using std::endl;



/*******************************************************************************
 * Constructor.
 */
InputFile::InputFile(const std::string& filePath) : cmdList_()
{
	if (filePath == "-") {
		if (Main::verbose()) cout << "[InputFile] Reading from stdin..." << endl;

		read(std::cin, "stdin");
	} else {
		if (Main::verbose()) cout << "Reading input file: " << filePath << endl;

		std::ifstream in(filePath.c_str());
		if (!in) throw ControllerException(ControllerError("Could not open input file: ") << filePath);
		read(in, filePath);
	}
}

/*******************************************************************************
 *
 */
void
InputFile::read(std::istream& in, const std::string& filePath)
{
	std::string line;
	int lineNum = 0;
	while (getline(in, line)) {
		lineNum++;

		// Empty line.
		if (line.length() == 0) continue;
		// Comment.
		if (line[0] == ';') continue;

		cmdList_.push_back(InputCommand(line, lineNum, filePath));
	}
}

/*******************************************************************************
 * Constructor.
 */
InputFile::InputCommand::InputCommand()
		: phoneme()
		, duration(0.0)
		, frequencyTargetList()
{
}

/*******************************************************************************
 * Constructor.
 */
InputFile::InputCommand::InputCommand(const std::string& line, int lineNum, const std::string& filePath)
		: phoneme()
		, duration(0.0)
		, frequencyTargetList()
{
	std::istringstream strIn(line);
	if (!(strIn >> phoneme)) {
		throwException(filePath, lineNum, "Phoneme not found.");
	}
	if (!(strIn >> duration)) {
		throwException(filePath, lineNum, "Duration not found.");
	}
	float freqPos;
	while (strIn >> freqPos) {
		float freqTarget;
		if (!(strIn >> freqTarget)) {
			throwException(filePath, lineNum, "Frequency target not found.");
		}
		frequencyTargetList.push_back(FrequencyTarget(freqPos, freqTarget));
	}
}

/*******************************************************************************
 *
 */
void
InputFile::rewritePhonemes(const PhonemeRewriter& rewriter)
{
	if (Main::verbose()) cout << "Rewriting phonemes..." << endl;

	rewriter.rewrite(cmdList_);
}

/*******************************************************************************
 *
 */
void
InputFile::rewritePostures(PostureRewriter& rewriter)
{
	if (Main::verbose()) cout << "Rewriting postures..." << endl;

	for (InputCommandList::iterator iter = cmdList_.begin(); iter != cmdList_.end(); ++iter) {
		rewriter.rewritePosture(cmdList_, iter);
	}
}

/*******************************************************************************
 * Removes first n elements from command list.
 */
void
InputFile::removeFirstCmd(int n)
{
	while (--n >= 0 && !cmdList_.empty()) {
		cmdList_.pop_front();
	}
}

} /* namespace Cgstmad */
