/***************************************************************************
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "model.h"

#include <iostream>

#include "config_file.h"
#include "input_file.h"



namespace {

const char* ARTIC_CONFIG_FILE_NAME = "/artic.xml";
const char* PHONEME_TRANSLATOR_CONFIG_FILE_NAME = "/phoneme_translation.config";
const char* POSTURE_REWRITER_CONFIG_FILE_NAME = "/posture_rewriting.config";

} /* namespace */

//==============================================================================

namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
Model::Model(const std::string& configDirPath)
		: configDirPath_(configDirPath)
		, phonemeRewriter_(configDirPath + PHONEME_TRANSLATOR_CONFIG_FILE_NAME)
		, postureRewriter_(configDirPath + POSTURE_REWRITER_CONFIG_FILE_NAME, postureMap_)
{
}

/*******************************************************************************
 * Destructor.
 */
Model::~Model()
{
}

/*******************************************************************************
 *
 */
void
Model::load()
{
	std::string filePath = configDirPath_ + ARTIC_CONFIG_FILE_NAME;

	// Load the configuration file.
	if (Main::verbose()) std::cout << "Loading xml configuration: " << filePath << std::endl;
	ConfigFile cfg(*this, filePath);
	cfg.loadModel();

	preparePostures();
	prepareEquations();
	prepareRules();
}

/*******************************************************************************
 *
 */
void
Model::preparePostures()
{
	if (Main::verbose()) std::cout << "Preparing postures..." << std::endl;

	// Fill the category code in the category list of each Posture.
	for (PostureMap::const_iterator itPost = postureMap_.begin(); itPost != postureMap_.end(); ++itPost) {
		const PostureMap::value_type& v = *itPost;
		for (CategoryList::iterator itCat = v.second->categoryList_.begin();
				itCat != v.second->categoryList_.end(); ++itCat) {

			CategoryCodeMap::iterator itMap = categoryCodeMap_.find(itCat->name);
			if (itMap != categoryCodeMap_.end()) {
				itCat->code = itMap->second;
			}
		}
	}
}

/*******************************************************************************
 *
 */
void
Model::prepareEquations()
{
	if (Main::verbose()) std::cout << "Preparing equations..." << std::endl;

	// Convert the formula expression to a tree.
	for (EquationMap::const_iterator iter = equationMap_.begin(); iter != equationMap_.end(); ++iter) {
		Equation& e = *iter->second;
		e.parseFormula(formulaSymbol_);
	}
}

/*******************************************************************************
 *
 */
void
Model::prepareRules()
{
	if (Main::verbose()) std::cout << "Preparing rules..." << std::endl;

	// Convert the boolean expression string of each Rule to a tree.
	for (RuleList::const_iterator iter = ruleList_.begin(); iter != ruleList_.end(); ++iter) {
		Rule& r = **iter;
		r.parseBooleanExpression(categoryCodeMap_);
	}
}

/*******************************************************************************
 *
 */
void
Model::filterInput(InputFile& inFile)
{
	inFile.rewritePhonemes(phonemeRewriter_);
	inFile.rewritePostures(postureRewriter_);
}

/*******************************************************************************
 *
 */
void
Model::printInfo() const
{
	//---------------------------------------------------------
	// Categories.
	std::cout << std::string(40, '-') << std::endl << "Categories:" << std::endl << std::endl;
	for (CategoryCodeMap::const_iterator iter = categoryCodeMap_.begin();
			iter != categoryCodeMap_.end(); ++iter) {
		std::cout << "category: " << iter->first << " code: " << iter->second << std::endl;
	}

	//---------------------------------------------------------
	// Postures.
	std::cout << std::string(40, '-') << std::endl << "Postures:" << std::endl << std::endl;
	for (PostureMap::const_iterator iter = postureMap_.begin(); iter != postureMap_.end(); ++iter) {
		const PostureMap::value_type& v = *iter;
		std::cout << "posture symbol: " << v.second->name() << " r8: " << v.second->getParameterValue(Parameter::PARAMETER_R8) << std::endl;

		for (CategoryList::const_iterator it2 = v.second->categoryList_.begin();
				it2 != v.second->categoryList_.end(); ++it2) {
			std::cout << "  categ: " << it2->name << "_" << it2->code << std::endl;
		}
		std::cout << "  symbols.duration: " << v.second->symbols().duration << std::endl;
		std::cout << "  symbols.transition: " << v.second->symbols().transition << std::endl;
		std::cout << "  symbols.qssa: " << v.second->symbols().qssa << std::endl;
		std::cout << "  symbols.qssb: " << v.second->symbols().qssb << std::endl;
	}

	//---------------------------------------------------------
	// Equations.
	std::cout << std::string(40, '-') << std::endl << "Equations:" << std::endl << std::endl;
	FormulaSymbolList symbolList;
	symbolList[FormulaSymbol::SYMB_TRANSITION1] = 100.1f;
	symbolList[FormulaSymbol::SYMB_TRANSITION2] = 100.2f;
	symbolList[FormulaSymbol::SYMB_TRANSITION3] = 100.3f;
	symbolList[FormulaSymbol::SYMB_TRANSITION4] = 100.4f;
	symbolList[FormulaSymbol::SYMB_QSSA1] = 20.5f;
	symbolList[FormulaSymbol::SYMB_QSSA2] = 20.6f;
	symbolList[FormulaSymbol::SYMB_QSSA3] = 20.7f;
	symbolList[FormulaSymbol::SYMB_QSSA4] = 20.8f;
	symbolList[FormulaSymbol::SYMB_QSSB1] = 20.9f;
	symbolList[FormulaSymbol::SYMB_QSSB2] = 20.0f;
	symbolList[FormulaSymbol::SYMB_QSSB3] = 20.1f;
	symbolList[FormulaSymbol::SYMB_QSSB4] = 20.2f;
	symbolList[FormulaSymbol::SYMB_TEMPO12] = 1.1f;
	symbolList[FormulaSymbol::SYMB_TEMPO23] = 1.2f;
	symbolList[FormulaSymbol::SYMB_TEMPO34] = 1.3f;
	symbolList[FormulaSymbol::SYMB_RD] = 150.4f;
	symbolList[FormulaSymbol::SYMB_MARK1] = 150.5f;
	symbolList[FormulaSymbol::SYMB_MARK2] = 150.6f;
	symbolList[FormulaSymbol::SYMB_NULL] = 1.0f;
	for (EquationMap::const_iterator iter = equationMap_.begin(); iter != equationMap_.end(); ++iter) {
		const EquationMap::value_type& v = *iter;
		std::cout << "=== Equation: [" << v.second->name << "] group: " << v.second->groupName  << std::endl;
		std::cout << "    [" << v.second->formula << "]" << std::endl;
		std::cout << *v.second->formulaRoot << std::endl;
		std::cout << "*** EVAL=" << v.second->formulaRoot->eval(symbolList) << std::endl;
	}

	//---------------------------------------------------------
	// Transitions.
	std::cout << std::string(40, '-') << std::endl << "Transitions:" << std::endl << std::endl;
	for (TransitionMap::const_iterator iter = transitionMap_.begin(); iter != transitionMap_.end(); ++iter) {
		const Transition& t = *iter->second;
		std::cout << "### Transition: [" << t.name() << "] group: " << t.groupName()  << std::endl;
		std::cout << "    type=" << t.type() << std::endl;

		for (size_t size = t.getNumberOfPointLists(), i = 0; i < size; ++i) {
			std::cout << "    point type=" << i << std::endl;
			for (size_t sizePoint = t.getNumberOfPointsInList(i), j = 0; j < sizePoint; ++j) {
				const Transition::Point& point = t.getPoint(i, j);
				std::cout << "       point: type=" << point.type << " value=" << point.value;
				std::cout << " freeTime=" << point.freeTime << " timeExpression=" << point.timeExpression << std::endl;
			}
		}
	}

	//---------------------------------------------------------
	// Rules.

	// Print boolean expression tree for each Rule.
	std::cout << std::string(40, '-') << std::endl << "Rules:" << std::endl << std::endl;
	for (RuleList::const_iterator iter = ruleList_.begin(); iter != ruleList_.end(); ++iter) {
		const Rule& r = **iter;
		std::cout << "--------------------------------------" << std::endl;
		std::cout << "Rule number: " << r.number() << std::endl << std::endl;
		r.printBooleanNodeTree();
	}
	std::cout << "--------------------------------------" << std::endl;
	PostureTmpList postList;
	typedef PostureMap::value_type VT;
	typedef PostureMap::const_iterator CI;
	CI iter = postureMap_.find("{");
	const Posture* pp = findPosture("{");
	pp = findPosture("m");
	if (pp) postList.push_back(pp);
	pp = findPosture("{");
	if (pp) postList.push_back(pp);
	pp = findPosture("s");
	if (pp) postList.push_back(pp);
	for (RuleList::const_iterator iter = ruleList_.begin(); iter != ruleList_.end(); ++iter) {
		const Rule& r = **iter;
		std::cout << "---" << std::endl;
		std::cout << "Rule number: " << r.number() << std::endl;
		std::cout << "bool=" << r.evalBooleanExpression(postList) << std::endl;
	}
	std::cout << "--------------------------------------" << std::endl;
	for (RuleList::const_iterator iter = ruleList_.begin(); iter != ruleList_.end(); ++iter) {
		const Rule& r = **iter;
		std::cout << "---" << std::endl;
		std::cout << "Rule number: " << r.number() << std::endl;
		std::cout << "Number of boolean expressions = " << r.getNumberOfPostures() <<
			" minDurations = " << r.minDuration12() <<
			' ' << r.minDuration23() <<
			' ' << r.minDuration34() << std::endl;
	}
}

} /* namespace Cgstmad */
