/***************************************************************************
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef MODEL_H_
#define MODEL_H_

#include "main.h"

#include <string>
#include <tr1/memory>
#include <tr1/unordered_map>
#include <vector>

#include "category.h"
#include "equation.h"
#include "formula_symbol.h"
#include "parameter.h"
#include "phoneme_rewriter.h"
#include "posture.h"
#include "posture_rewriter.h"
#include "rule.h"
#include "transition.h"



namespace Cgstmad {

class ConfigFile;
class InputFile;

class Model {
public:
	Model(const std::string& configDirPath);
	~Model();

	void load();
	void filterInput(InputFile& inFile);
	void printInfo() const;
	void clearFormulaSymbolList();
	void setFormulaSymbolValue(FormulaSymbol::Code symbol, float value);
	float getFormulaSymbolValue(FormulaSymbol::Code symbol) const;
	float evalEquationFormula(const std::string& equationName) const;
	float getParameterMinimum(Parameter::Code parameter) const;
	float getParameterSpan(Parameter::Code parameter) const;
	void clipParameterValue(float& value, Parameter::Code parameter) const;
	const Posture* findPosture(const std::string& name) const;
	const Transition* findTransition(const std::string& name) const;
	const Transition* findSpecialTransition(const std::string& name) const;
	const Rule* findFirstMatchingRule(const PostureTmpList& postureList) const;
private:
	typedef std::vector<Rule_ptr> RuleList;

	const std::string& configDirPath_;
	CategoryCodeMap categoryCodeMap_;
	Parameter::CodeMap parameterCodeMap_;
	Parameter::Info::Array parameterInfoArray_;
	PostureMap postureMap_;
	RuleList ruleList_;
	FormulaSymbol formulaSymbol_;
	EquationMap equationMap_;
	TransitionMap transitionMap_;
	TransitionMap specialTransitionMap_;
	PhonemeRewriter phonemeRewriter_;
	PostureRewriter postureRewriter_;
	FormulaSymbolList formulaSymbolList_;

	void preparePostures();
	void prepareEquations();
	void prepareRules();

	friend class ConfigFile;
};



/*******************************************************************************
 *
 */
inline
void
Model::clearFormulaSymbolList()
{
	formulaSymbolList_.assign(0.0);
}

/*******************************************************************************
 *
 */
inline
void
Model::setFormulaSymbolValue(FormulaSymbol::Code symbol, float value)
{
	formulaSymbolList_[symbol] = value;
}

/*******************************************************************************
 *
 */
inline
float
Model::getFormulaSymbolValue(FormulaSymbol::Code symbol) const
{
	return formulaSymbolList_[symbol];
}

/*******************************************************************************
 *
 */
inline
float
Model::evalEquationFormula(const std::string& equationName) const
{
	EquationMap::const_iterator iter = equationMap_.find(equationName);
	if (iter == equationMap_.end()) {
		throw ControllerException(ControllerError("Equation not found: ") << equationName);
	}

	return iter->second->evalFormula(formulaSymbolList_);
}

/*******************************************************************************
 *
 */
inline
float
Model::getParameterMinimum(Parameter::Code parameter) const
{
	return parameterInfoArray_[parameter].minimum;
}

/*******************************************************************************
 *
 */
inline
float
Model::getParameterSpan(Parameter::Code parameter) const
{
	return parameterInfoArray_[parameter].span;
}

/*******************************************************************************
 *
 */
inline
void
Model::clipParameterValue(float& value, Parameter::Code parameter) const
{
	if (Parameter::isVolumeParameter(parameter)) {
		// Volumes are not in dB.
		if (value < 0.0f) {
			value = 0.0f;
		} else if (value > 1.0f) {
			value = 1.0f;
		}
	} else {
		const Parameter::Info& info = parameterInfoArray_[parameter];
		if (value < info.minimum) {
			value = info.minimum;
		} else {
			const float max = info.minimum + info.span;
			if (value > max) {
				value = max;
			}
		}
	}
}

/*******************************************************************************
 * Find a Posture with the given name.
 *
 * Returns a pointer to the Posture, or 0 (zero) if a Posture
 * was not found.
 */
inline
const Posture*
Model::findPosture(const std::string& name) const
{
	PostureMap::const_iterator itPost = postureMap_.find(name);
	if (itPost == postureMap_.end()) {
		return 0;
	}
	return itPost->second.get();
}

/*******************************************************************************
 * Find a Transition with the given name.
 *
 * Returns a pointer to the Transition, or 0 (zero) if a Transition
 * was not found.
 */
inline
const Transition*
Model::findTransition(const std::string& name) const
{
	TransitionMap::const_iterator itTrans = transitionMap_.find(name);
	if (itTrans == transitionMap_.end()) {
		return 0;
	}
	return itTrans->second.get();
}

/*******************************************************************************
 * Find a Special Transition with the given name.
 *
 * Returns a pointer to the Special Transition, or 0 (zero) if a Special
 * Transition was not found.
 */
inline
const Transition*
Model::findSpecialTransition(const std::string& name) const
{
	TransitionMap::const_iterator itTrans = specialTransitionMap_.find(name);
	if (itTrans == specialTransitionMap_.end()) {
		return 0;
	}
	return itTrans->second.get();
}

/*******************************************************************************
 * Find the first Rule that matches the given sequence of Postures.
 *
 * Returns a pointer to the Rule, or 0 (zero) if a Rule was not found.
 */
inline
const Rule*
Model::findFirstMatchingRule(const PostureTmpList& postureList) const
{
	for (RuleList::const_iterator itRule = ruleList_.begin(); itRule != ruleList_.end(); ++itRule) {
		const Rule& r = **itRule;
		if (r.evalBooleanExpression(postureList)) {
			return &r;
		}
	}
	return 0;
}

} /* namespace Cgstmad */

#endif /*MODEL_H_*/
