/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "pitch_processor.h"



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
PitchProcessor::PitchProcessor(float controlRate)
		: firstPoint_(true)
		, position_(0)
		, timeStep_(1000.0f / controlRate)
		, previousPitch_(0.0)
		, currentPeriod_(0)
{
}

/*******************************************************************************
 *
 */
void
PitchProcessor::reset()
{
	position_ = 0;
	periodList_.clear();
	currentPeriod_ = 0;
}

/*******************************************************************************
 *
 */
void
PitchProcessor::addTarget(float targetTime, float targetPitch)
{
	const float epsilon = 1.0e-6f;

	PitchPeriod period;
	period.targetPitch = targetPitch;
	period.targetTime = targetTime;
	if (periodList_.empty()) { // first period
		if (firstPoint_) {
			period.basePitch = targetPitch; // there is no previous point
		} else {
			period.basePitch = previousPitch_;
		}
		period.baseTime = 0.0;
	} else {
		PitchPeriod& last = periodList_.back();
		// If the last period has the same target time:
		if ( (targetTime < last.targetTime + epsilon) && (targetTime > last.targetTime - epsilon) ) {
			// Ignore.
			return;
		}
		period.basePitch = last.targetPitch;
		period.baseTime = last.targetTime;
	}
	period.deltaPitch = period.targetPitch - period.basePitch;

	periodList_.push_back(period);
	firstPoint_ = false;
}

/*******************************************************************************
 *
 */
void
PitchProcessor::addFinalTarget(float targetTime, float targetPitch)
{
	if (!periodList_.empty()) {
		addTarget(targetTime, targetPitch);
	} else {
		PitchPeriod period;
		if (firstPoint_) {
			period.basePitch = targetPitch;
		} else {
			period.basePitch = previousPitch_;
		}
		period.baseTime = 0.0;
		period.targetPitch = targetPitch;
		period.targetTime = targetTime;
		period.deltaPitch = period.targetPitch - period.basePitch;
		periodList_.push_back(period);
	}
	firstPoint_ = false;
}

/*******************************************************************************
 *
 */
float
PitchProcessor::getNext()
{
	const float epsilon = 1.0e-6f;

	float time = timeStep_ * position_++;

	for (PitchPeriodList::size_type size = periodList_.size(), i = currentPeriod_; i < size; ++i) {
		const PitchPeriod& p = periodList_[i];
		if (time < p.targetTime + epsilon) {
			currentPeriod_ = i;

			const float dt = p.targetTime - p.baseTime;
			const float pitch = (dt > epsilon) ?
						p.basePitch + p.deltaPitch * (time - p.baseTime) / dt
						: p.basePitch;
			previousPitch_ = pitch;
			return pitch;
		}
	}

	return previousPitch_;
}

} /* namespace Cgstmad */
