/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2007         *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from ObjC to C++) and modified version     *
 *   of a GnuSpeech's source code.                                         *
 *                                                                         *
 ***************************************************************************/

#include "posture_rewriter.h"

#include <fstream>
#include <iostream>

#include "controller_exception.h"
#include "text.h"

//#define DEBUG_TRANSITION 1



namespace {

using namespace Cgstmad;

const char COMMENT_CHAR = '#';

void
throwException(const std::string& filePath, int lineNumber, const char* message)
{
	throw ControllerException(ControllerError("Error in PostureRewriter: File: ")
							<< filePath << " Line " << lineNumber << ": " << message);
}

template<typename T>
void
throwException(const std::string& filePath, int lineNumber, const char* message, const T& complement)
{
	throw ControllerException(ControllerError("Error in PostureRewriter: File: ")
							<< filePath << " Line " << lineNumber << ": " << message << complement);
}

} /* namespace */

//==============================================================================

namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
PostureRewriter::PostureRewriter(const std::string& filePath, const PostureMap& postureMap)
		: postureMap_(postureMap)
		, currentState_(0)
{
	actionNameTypeMap_["if-last_posture_has_prefix-insert" ] = ACTION_IF_LAST_POSTURE_HAS_PREFIX_INSERT;
	actionNameTypeMap_["if-posture_is_equal_to_last-insert"] = ACTION_IF_POSTURE_IS_EQUAL_TO_LAST_INSERT;
	actionNameTypeMap_["insert"                            ] = ACTION_INSERT;
	actionNameTypeMap_["replace_last"                      ] = ACTION_REPLACE_LAST;
	actionNameTypeMap_["remove_last"                       ] = ACTION_REMOVE_LAST;
	actionNameTypeMap_["set_duration"                      ] = ACTION_SET_DURATION;
	actionNameTypeMap_["join"                              ] = ACTION_JOIN;

	if (Main::verbose()) {
		std::cout << "Configuring posture rewriter..." << std::endl;
		std::cout << "Reading configuration file: " << filePath << std::endl;
	}

	std::ifstream in(filePath.c_str());
	if (!in) throw ControllerException(ControllerError("PostureRewriter::PostureRewriter: Could not open configuration file: ") << filePath);

	std::string line;
	int lineNum = 0;
	while (getline(in, line)) {
		lineNum++;

		// Empty line.
		if (line.length() == 0) continue;

		// Comment (must start at the beginning of the line).
		if (line[0] == COMMENT_CHAR) {
			continue;
		}

		parseLine(filePath, lineNum, line);
	}

	if (stateActionList_.size() == 0) {
		throw ControllerException("PostureRewriter::PostureRewriter: Didn't find any states.");
	}

#if 0
	for (StateActionList::size_type size = stateActionList_.size(), i = 0; i < size; ++i) {
		std::cout << "=== Actions state: " << i << std::endl;
		for (ActionList::size_type size = stateActionList_[i].size(), j = 0; j < size; ++j) {
			stateActionList_[i][j]->print();
		}
	}
	for (StateActionList::size_type size = stateTransitionList_.size(), i = 0; i < size; ++i) {
		std::cout << "=== Transitions state: " << i << std::endl;
		for (ActionList::size_type size = stateTransitionList_[i].size(), j = 0; j < size; ++j) {
			stateTransitionList_[i][j]->print();
		}
	}
	for (CategoryList::size_type size = categoryList_.size(), i = 0; i < size; ++i) {
		std::cout << "=== Category [" << i << "] [" << categoryList_[i] << "]" << std::endl;
	}
#endif
}

/*******************************************************************************
 * Destructor.
 */
PostureRewriter::~PostureRewriter()
{
}

/*******************************************************************************
 *
 */
CategoryList::size_type
PostureRewriter::addCategory(const std::string& name)
{
	for (CategoryList::size_type size = categoryList_.size(), i = 0; i < size; ++i) {
		if (categoryList_[i] == name) return i;
	}
	categoryList_.push_back(name);
	return categoryList_.size() - 1;
}

/*******************************************************************************
 *
 */
void
PostureRewriter::parseLine(const std::string& filePath, int lineNum, const std::string& line)
{
	std::istringstream strIn(line);
	std::string s;

	// Category.
	bool action = false;
	if (!(strIn >> s)) throwException(filePath, lineNum, "Category not found.");
	switch (s[0]) {
	case 'A':
		action = true;
		break;
	case 'T':
		// action = false;
		break;
	default:
		throwException(filePath, lineNum, "Invalid category: ", s[0]);
	}

	// State.
	if (!(strIn >> s)) throwException(filePath, lineNum, "State not found.");
	size_t state = Text::parseString<size_t>(s);
	if (stateActionList_.size() <= state) {
		stateActionList_.resize(state + 1);
	}
	if (stateTransitionList_.size() <= state) {
		stateTransitionList_.resize(state + 1);
	}

	if (action) {
		// =========================== Action.

		bool elseAction = false;

		// Action name.
		if (!(strIn >> s)) throwException(filePath, lineNum, "Action name not found.");
		if (s == "else") {
			elseAction = true;
			if (!(strIn >> s)) throwException(filePath, lineNum, "Action name not found after \"else\".");
		}

		// Action type.
		ActionNameTypeMap::iterator iterAction = actionNameTypeMap_.find(s);
		if (iterAction == actionNameTypeMap_.end()) throwException(filePath, lineNum, "Invalid action: ", s);
		switch (iterAction->second) {
		case ACTION_IF_LAST_POSTURE_HAS_PREFIX_INSERT:
			{
				std::string prefix;
				if (!(strIn >> prefix)) throwException(filePath, lineNum, "Prefix not found.");
				std::string newPosture;
				if (!(strIn >> newPosture)) throwException(filePath, lineNum, "New posture not found.");

				Action_ptr ptr(new IfLastPostureHasPrefixInsertAction(*this, elseAction, prefix, newPosture));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_IF_POSTURE_IS_EQUAL_TO_LAST_INSERT:
			{
				std::string newPosture;
				if (!(strIn >> newPosture)) throwException(filePath, lineNum, "New posture not found.");

				Action_ptr ptr(new IfPostureIsEqualToLastInsertAction(*this, elseAction, newPosture));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_INSERT:
			{
				std::string newPosture;
				if (!(strIn >> newPosture)) throwException(filePath, lineNum, "New posture not found.");

				Action_ptr ptr(new InsertAction(elseAction, newPosture));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_REPLACE_LAST:
			{
				std::string newPosture;
				if (!(strIn >> newPosture)) throwException(filePath, lineNum, "New posture not found.");

				Action_ptr ptr(new ReplaceLastAction(elseAction, newPosture));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_REMOVE_LAST:
			{
				Action_ptr ptr(new RemoveLastAction(elseAction));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_SET_DURATION:
			{
				float newDuration;
				if (!(strIn >> newDuration)) throwException(filePath, lineNum, "New duration not found or in wrong format.");

				Action_ptr ptr(new SetDurationAction(elseAction, newDuration));
				stateActionList_[state].push_back(ptr);
			}
			break;
		case ACTION_JOIN:
			{
				Action_ptr ptr(new JoinAction(elseAction));
				stateActionList_[state].push_back(ptr);
			}
			break;
		}
	} else {
		// =========================== Transition.

		std::string category;
		if (!(strIn >> category)) throwException(filePath, lineNum, "Category not found.");
		if (!(strIn >> s)) throwException(filePath, lineNum, "New state not found.");
		size_t newState = Text::parseString<size_t>(s);

		Transition_ptr newTrans(new StateTransition(*this, category, newState));

		// Search for duplicate transition.
		bool foundDuplicate = false;
		const PostureRewriter::TransitionList& trList = stateTransitionList_[state];
		for (TransitionList::size_type size = trList.size(), i = 0; i < size; ++i) {
			const StateTransition& tr = *trList[i];
			if (tr.categoryCode() == newTrans->categoryCode()) {
				if (tr.newState() != newTrans->newState()) {
					throwException(filePath, lineNum, "Inconsistent duplicate state transition. Category: ", category);
				}
				foundDuplicate = true;
				break;
			}
		}

		if (!foundDuplicate) {
			stateTransitionList_[state].push_back(newTrans);
		}
	}
}

/*******************************************************************************
 *
 */
bool
PostureRewriter::findAndExecuteTransition(int state, const Posture& posture)
{
	const TransitionList& trList = stateTransitionList_[state];
	const ActionList::size_type size = trList.size();
	for (ActionList::size_type i = 0; i < size; ++i) {
		// Use only the category that has a name equal to the name of the Posture.
		if (trList[i]->exec(posture, true)) {
			return true;
		}
	}
	for (ActionList::size_type i = 0; i < size; ++i) {
		if (trList[i]->exec(posture, false)) {
			return true;
		}
	}
	return false;
}

/*******************************************************************************
 *
 */
void
PostureRewriter::rewritePosture(InputFile::InputCommandList& cmdList,
				InputFile::InputCommandList::iterator& cmdListIter /*, bool followsWordMarker*/) //TODO: followsWordMarker?
{
	const std::string currentPosture = cmdListIter->phoneme;

	PostureMap::const_iterator postIter = postureMap_.find(currentPosture);
	if (postIter == postureMap_.end()) throw ControllerException(ControllerError("PostureRewriter::rewritePosture: Posture not found: ") << cmdListIter->phoneme);
	Posture& posture = *postIter->second;

	// State transition.
	// Search for a transition in the current state.
	if (!findAndExecuteTransition(currentState_, posture)) {
		// Search for a transition in the state 0 (zero).
		if (currentState_ == 0 || !findAndExecuteTransition(0, posture)) {
			currentState_ = 0;
			lastPosture_.clear();
#ifdef DEBUG_TRANSITION
			std::cout << "STATE TRANSITION newState=0 posture=" << currentPosture << std::endl;
#endif
			return;
		}
	}

	// Execute actions.
	bool lastResult = true;
	const ActionList& actList = stateActionList_[currentState_];
	for (ActionList::size_type size = actList.size(), i = 0; i < size; ++i) {
		if (lastResult == true && actList[i]->elseAction()) continue;

		lastResult = actList[i]->exec(cmdList, cmdListIter);
	}

	lastPosture_ = currentPosture;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::IfLastPostureHasPrefixInsertAction::exec(InputFile::InputCommandList& cmdList,
								InputFile::InputCommandList::iterator& cmdListIter) const
{
	Text::MatchAtBegin matchAtBegin;
	if (!matchAtBegin(rewriter_.lastPosture_, prefix_)) return false;

	InputFile::InputCommand cmd;
	cmd.phoneme = newPosture_;
	cmd.duration = 0.0; //cmdListIter->duration;
	cmdList.insert(cmdListIter, cmd);
	if (Main::verbose()) std::cout << "IfLastPostureHasPrefixInsertAction: Posture inserted: " << newPosture_ << std::endl;
	return true;
}

void
PostureRewriter::IfLastPostureHasPrefixInsertAction::print() const
{
	std::cout << "ACTION_IF_LAST_POSTURE_HAS_PREFIX_INSERT elseAction=" << elseAction_ << " prefix=" << prefix_ << " newPosture=" << newPosture_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::IfPostureIsEqualToLastInsertAction::exec(InputFile::InputCommandList& cmdList,
								InputFile::InputCommandList::iterator& cmdListIter) const
{
	if (rewriter_.lastPosture_ != cmdListIter->phoneme/* && followsWordMarker*/) return false; //TODO: followsWordMarker?

	InputFile::InputCommand cmd;
	cmd.phoneme = newPosture_;
	// TODO: Verify duration value.
	cmd.duration = 0.0; //cmdListIter->duration;
	cmdList.insert(cmdListIter, cmd);
	if (Main::verbose()) std::cout << "IfPostureIsEqualToLastInsertAction: Posture inserted: " << newPosture_ << std::endl;
	return true;
}

void
PostureRewriter::IfPostureIsEqualToLastInsertAction::print() const
{
	std::cout << "ACTION_IF_POSTURE_IS_EQUAL_TO_LAST_INSERT elseAction=" << elseAction_ << " newPosture=" << newPosture_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::InsertAction::exec(InputFile::InputCommandList& cmdList,
					InputFile::InputCommandList::iterator& cmdListIter) const
{
	InputFile::InputCommand cmd;
	cmd.phoneme = newPosture_;
	// TODO: Verify duration value.
	cmd.duration = 0.0; //cmdListIter->duration;
	cmdList.insert(cmdListIter, cmd);
	if (Main::verbose()) std::cout << "InsertAction: Posture inserted: " << newPosture_ << std::endl;
	return true;
}

void
PostureRewriter::InsertAction::print() const
{
	std::cout << "ACTION_INSERT elseAction=" << elseAction_ << " newPosture=" << newPosture_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::ReplaceLastAction::exec(InputFile::InputCommandList& cmdList,
						InputFile::InputCommandList::iterator& cmdListIter) const
{
	if (cmdListIter == cmdList.begin()) return false;

	InputFile::InputCommandList::iterator iter = cmdListIter;
	--iter;
	if (Main::verbose()) std::cout << "ReplaceLastAction: Posture " << iter->phoneme << " replaced by: " << newPosture_ << std::endl;
	iter->phoneme = newPosture_;
	return true;
}

void
PostureRewriter::ReplaceLastAction::print() const
{
	std::cout << "ACTION_REPLACE_LAST elseAction=" << elseAction_ << " newPosture=" << newPosture_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::RemoveLastAction::exec(InputFile::InputCommandList& cmdList,
						InputFile::InputCommandList::iterator& cmdListIter) const
{
	if (cmdListIter == cmdList.begin()) return false;

	InputFile::InputCommandList::iterator iter = cmdListIter;
	cmdList.erase(--iter);
	if (Main::verbose()) std::cout << "RemoveLastAction: Posture " << iter->phoneme << " removed." << std::endl;
	return true;
}

void
PostureRewriter::RemoveLastAction::print() const
{
	std::cout << "ACTION_REMOVE_LAST elseAction=" << elseAction_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::SetDurationAction::exec(InputFile::InputCommandList& /*cmdList*/,
						InputFile::InputCommandList::iterator& cmdListIter) const
{
	cmdListIter->duration = newDuration_;
	if (Main::verbose()) std::cout << "SetDurationAction: Duration of posture " << cmdListIter->phoneme << " set to " << newDuration_ << std::endl;
	return true;
}

void
PostureRewriter::SetDurationAction::print() const
{
	std::cout << "ACTION_SET_DURATION duration=" << newDuration_ << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the action was executed.
 */
bool
PostureRewriter::JoinAction::exec(InputFile::InputCommandList& cmdList,
						InputFile::InputCommandList::iterator& cmdListIter) const
{
	InputFile::InputCommand& currentCommand = *cmdListIter;
	InputFile::InputCommandList::iterator iter = cmdListIter;
	--iter;
	currentCommand.duration += iter->duration;
	cmdList.erase(iter);

	if (Main::verbose()) std::cout << "JoinAction: Two postures \"" << currentCommand.phoneme << "\" joined." << std::endl;
	return true;
}

void
PostureRewriter::JoinAction::print() const
{
	std::cout << "ACTION_JOIN" << std::endl;
}

/*******************************************************************************
 * Executes the command. Returns true if the transition was executed.
 */
bool
PostureRewriter::StateTransition::exec(const Posture& posture, bool postureNameOnly) const
{
	for (CategoryList::size_type size = rewriter_.categoryList_.size(), i = 0; i < size; ++i) {
		if (posture.isMemberOfCategory(rewriter_.categoryList_[i], postureNameOnly) && categoryCode_ == i) {
			rewriter_.currentState_ = newState_;
#ifdef DEBUG_TRANSITION
			print();
#endif
			return true;
		}
	}
	return false;
}

void
PostureRewriter::StateTransition::print() const
{
	std::cout << "STATE TRANSITION category=" << category_ << " categoryCode=" << categoryCode_ << " newState=" << newState_ << std::endl;
}

} /* namespace Cgstmad */
