/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2007         *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from ObjC to C++) and modified version     *
 *   of a GnuSpeech's source code.                                         *
 *                                                                         *
 ***************************************************************************/

#ifndef POSTURE_REWRITER_H_
#define POSTURE_REWRITER_H_

#include "main.h"

#include <string>
#include <vector>
#include <tr1/memory>
#include <tr1/unordered_map>

#include "input_file.h"
#include "posture.h"



namespace Cgstmad {

class PostureRewriter {
public:
	PostureRewriter(const std::string& filePath, const PostureMap& postureMap);
	~PostureRewriter();

	void rewritePosture(InputFile::InputCommandList& cmdList,
				InputFile::InputCommandList::iterator& cmdListIter /*, bool followsWordMarker*/);
private:
	enum ActionType {
		ACTION_IF_LAST_POSTURE_HAS_PREFIX_INSERT,
		ACTION_IF_POSTURE_IS_EQUAL_TO_LAST_INSERT,
		ACTION_INSERT,
		ACTION_REPLACE_LAST,
		ACTION_REMOVE_LAST,
		ACTION_SET_DURATION,
		ACTION_JOIN
	};

	class Action;
	class IfLastPostureHasPrefixInsertAction;
	class IfPostureIsEqualToLastInsertAction;
	class InsertAction;
	class ReplaceLastAction;
	class RemoveLastAction;
	class SetDurationAction;
	class JoinAction;
	class StateTransition;

	friend class IfLastPostureHasPrefixInsertAction;
	friend class IfPostureIsEqualToLastInsertAction;
	friend class StateTransition;

	typedef std::tr1::unordered_map<std::string, ActionType> ActionNameTypeMap;
	typedef std::vector<std::string> CategoryList;
	typedef std::tr1::shared_ptr<Action> Action_ptr;
	typedef std::vector<Action_ptr> ActionList;
	typedef std::vector<ActionList> StateActionList;
	typedef std::tr1::shared_ptr<StateTransition> Transition_ptr;
	typedef std::vector<Transition_ptr> TransitionList;
	typedef std::vector<TransitionList> StateTransitionList;

	ActionNameTypeMap actionNameTypeMap_;
	const PostureMap& postureMap_;
	CategoryList categoryList_;
	int currentState_;
	std::string lastPosture_;
	StateActionList stateActionList_;
	StateTransitionList stateTransitionList_;

	void parseLine(const std::string& filePath, int lineNum, const std::string& line);
	CategoryList::size_type addCategory(const std::string& name);
	bool findAndExecuteTransition(int state, const Posture& posture);
};

class PostureRewriter::Action {
public:
	Action() {}
	virtual ~Action() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const = 0;
	virtual void print() const = 0;

	virtual bool elseAction() const = 0;
};

class PostureRewriter::IfLastPostureHasPrefixInsertAction : public PostureRewriter::Action {
public:
	IfLastPostureHasPrefixInsertAction(const PostureRewriter& rewriter, bool elseAction, const std::string& prefix, const std::string& newPosture)
			: rewriter_(rewriter)
			, elseAction_(elseAction)
			, prefix_(prefix)
			, newPosture_(newPosture)
	{
	}
	virtual ~IfLastPostureHasPrefixInsertAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const PostureRewriter& rewriter_;
	const bool elseAction_;
	const std::string prefix_;
	const std::string newPosture_;
};

class PostureRewriter::IfPostureIsEqualToLastInsertAction : public PostureRewriter::Action {
public:
	IfPostureIsEqualToLastInsertAction(const PostureRewriter& rewriter, bool elseAction, const std::string& newPosture)
			: rewriter_(rewriter)
			, elseAction_(elseAction)
			, newPosture_(newPosture)
	{
	}
	virtual ~IfPostureIsEqualToLastInsertAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const PostureRewriter& rewriter_;
	const bool elseAction_;
	const std::string newPosture_;
};

class PostureRewriter::InsertAction : public PostureRewriter::Action {
public:
	InsertAction(bool elseAction, const std::string& newPosture)
			: elseAction_(elseAction)
			, newPosture_(newPosture)
	{
	}
	virtual ~InsertAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const bool elseAction_;
	const std::string newPosture_;
};

class PostureRewriter::ReplaceLastAction : public PostureRewriter::Action {
public:
	ReplaceLastAction(bool elseAction, const std::string& newPosture)
			: elseAction_(elseAction)
			, newPosture_(newPosture)
	{
	}
	virtual ~ReplaceLastAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const bool elseAction_;
	const std::string newPosture_;
};

class PostureRewriter::RemoveLastAction : public PostureRewriter::Action {
public:
	RemoveLastAction(bool elseAction)
			: elseAction_(elseAction)
	{
	}
	virtual ~RemoveLastAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const bool elseAction_;
};

class PostureRewriter::SetDurationAction : public PostureRewriter::Action {
public:
	SetDurationAction(bool elseAction, float newDuration)
			: elseAction_(elseAction)
			, newDuration_(newDuration)
	{
	}
	virtual ~SetDurationAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const bool elseAction_;
	const float newDuration_;
};

// Join the previous and the current postures, summing the durations.
// The previous and the current postures should have the same symbol.
class PostureRewriter::JoinAction : public PostureRewriter::Action {
public:
	JoinAction(bool elseAction)
			: elseAction_(elseAction)
	{
	}
	virtual ~JoinAction() {}

	virtual bool exec(InputFile::InputCommandList& cmdList, InputFile::InputCommandList::iterator& cmdListIter) const;
	virtual void print() const;

	virtual bool elseAction() const { return elseAction_; }
private:
	const bool elseAction_;
};

class PostureRewriter::StateTransition {
public:
	StateTransition(PostureRewriter& rewriter, const std::string& category, int newState)
			: rewriter_(rewriter)
			, category_(category)
			, categoryCode_(rewriter.addCategory(category))
			, newState_(newState)
	{
	}
	~StateTransition() {}

	bool exec(const Posture& posture, bool postureNameOnly) const;
	void print() const;

	int categoryCode() const { return categoryCode_; }
	int newState() const { return newState_; }
private:
	PostureRewriter& rewriter_;
	const std::string category_;
	const CategoryList::size_type categoryCode_;
	const int newState_;
};

} /* namespace Cgstmad */

#endif /*POSTURE_REWRITER_H_*/
