/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "rule.h"

#include <cassert>

#include "category.h"
#include "controller_exception.h"
#include "model.h"
#include "text.h"



namespace {

using std::cout;
using std::endl;

using namespace Cgstmad;

const char rightParenChar = ')';
const char leftParenChar  = '(';
const char anyStringChar  = '*';
const std::string  orOpSymb = "or";
const std::string notOpSymb = "not";
const std::string xorOpSymb = "xor";
const std::string andOpSymb = "and";

enum SymbolType {
	SYMBOL_TYPE_INVALID,
	SYMBOL_TYPE_OR_OP,
	SYMBOL_TYPE_NOT_OP,
	SYMBOL_TYPE_XOR_OP,
	SYMBOL_TYPE_AND_OP,
	SYMBOL_TYPE_RIGHT_PAREN,
	SYMBOL_TYPE_LEFT_PAREN,
	SYMBOL_TYPE_STRING
};

bool isSeparator(char c)
{
	switch (c) {
	case rightParenChar: return true;
	case leftParenChar:  return true;
	case ' ':            return true;
	default: return false;
	}
}

class Parser {
public:
	Parser(const std::string& s, const CategoryCodeMap& categoryCodeMap)
				: categoryCodeMap_(categoryCodeMap)
				, s_(s)
				, pos_(0)
				, symbolType_(SYMBOL_TYPE_INVALID) {
		if (s.empty()) {
			throw ControllerException("Boolean expression parser error: Empty string.");
		}
		nextSymbol();
	}

	void throwException(const char* errorDescription) const;
	template<typename T> void throwException(const char* errorDescription, const T& complement) const;
	bool finished() const {
		return pos_ >= s_.size();
	}
	void nextSymbol();
	RuleBooleanNode_ptr getBooleanNode();
private:
	const CategoryCodeMap& categoryCodeMap_;
	const std::string s_;
	std::string::size_type pos_;
	std::string symbol_;
	SymbolType symbolType_;

	void skipSpaces();
};

void
Parser::throwException(const char* errorDescription) const
{
	throw ControllerException(ControllerError("Boolean expression parser error: ")
					<< errorDescription << " at position " << pos_ << " of string [" << s_ << "].");
}

template<typename T>
void
Parser::throwException(const char* errorDescription, const T& complement) const
{
	throw ControllerException(ControllerError("Boolean expression parser error: ")
					<< errorDescription << complement << " at position " << pos_ << " of string [" << s_ << "].");
}

void
Parser::skipSpaces()
{
	while (!finished() && s_[pos_] == ' ') ++pos_;
}

void
Parser::nextSymbol()
{
	skipSpaces();

	symbol_.clear();

	if (finished()) {
		symbolType_ = SYMBOL_TYPE_INVALID;
		return;
	}

	char c = s_[pos_++];
	symbol_ += c;
	switch (c) {
	case rightParenChar:
		symbolType_ = SYMBOL_TYPE_RIGHT_PAREN;
		break;
	case leftParenChar:
		symbolType_ = SYMBOL_TYPE_LEFT_PAREN;
		break;
	default:
		while ( !finished() && !isSeparator(c = s_[pos_]) ) {
			symbol_ += c;
			++pos_;
		}
		if (symbol_ == orOpSymb) {
			symbolType_ = SYMBOL_TYPE_OR_OP;
		} else if (symbol_ == andOpSymb) {
			symbolType_ = SYMBOL_TYPE_AND_OP;
		} else if (symbol_ == notOpSymb) {
			symbolType_ = SYMBOL_TYPE_NOT_OP;
		} else if (symbol_ == xorOpSymb) {
			symbolType_ = SYMBOL_TYPE_XOR_OP;
		} else {
			symbolType_ = SYMBOL_TYPE_STRING;
		}
	}
}

RuleBooleanNode_ptr
Parser::getBooleanNode()
{
	switch (symbolType_) {
	case SYMBOL_TYPE_LEFT_PAREN:
	{
		RuleBooleanNode_ptr p;

		nextSymbol();
		if (symbolType_ == SYMBOL_TYPE_NOT_OP) {

			// Operand.
			nextSymbol();
			RuleBooleanNode_ptr op(getBooleanNode());

			p.reset(new RuleBooleanNotExpression(op));
		} else {
			// 1st operand.
			RuleBooleanNode_ptr op1(getBooleanNode());

			// Operator.
			nextSymbol();
			switch (symbolType_) {
			case SYMBOL_TYPE_OR_OP:
			{	// 2nd operand.
				nextSymbol();
				RuleBooleanNode_ptr op2(getBooleanNode());

				p.reset(new RuleBooleanOrExpression(op1, op2));
				break;
			}
			case SYMBOL_TYPE_AND_OP:
			{	// 2nd operand.
				nextSymbol();
				RuleBooleanNode_ptr op2(getBooleanNode());

				p.reset(new RuleBooleanAndExpression(op1, op2));
				break;
			}
			case SYMBOL_TYPE_XOR_OP:
			{	// 2nd operand.
				nextSymbol();
				RuleBooleanNode_ptr op2(getBooleanNode());

				p.reset(new RuleBooleanXorExpression(op1, op2));
				break;
			}
			default:
				throwException("Invalid operator: ", symbolType_);
			}
		}

		nextSymbol();
		if (symbolType_ != SYMBOL_TYPE_RIGHT_PAREN) {
			throwException("Right parenthesis not found");
		}
		return p;
	}
	case SYMBOL_TYPE_STRING:
	{
		if (symbolType_ == SYMBOL_TYPE_INVALID) {
			throwException("Could not find the category");
		}

		const std::string& symbolTmp = symbol_;
		bool matchPartial = false;
		if (symbolTmp[symbolTmp.size() - 1] == anyStringChar) {
			matchPartial = true;
		}

		int categoryCode = 0;
		if (!matchPartial) {
			// Try to find a code for the category.
			CategoryCodeMap::const_iterator itMap = categoryCodeMap_.find(symbolTmp);
			if (itMap != categoryCodeMap_.end()) {
				categoryCode = itMap->second;
			}
		}

		return RuleBooleanNode_ptr(new RuleBooleanTerminal(
						matchPartial ? symbolTmp.substr(0, symbolTmp.size() - 1) : symbolTmp,
						matchPartial,
						categoryCode));
	}
	case SYMBOL_TYPE_OR_OP:
		throwException("Unexpected OR op.");
	case SYMBOL_TYPE_NOT_OP:
		throwException("Unexpected NOT op.");
	case SYMBOL_TYPE_XOR_OP:
		throwException("Unexpected XOR op.");
	case SYMBOL_TYPE_AND_OP:
		throwException("Unexpected AND op.");
	case SYMBOL_TYPE_RIGHT_PAREN:
		throwException("Unexpected right parenthesis");
	default:
		throwException("Invalid symbol");
	}
	return RuleBooleanNode_ptr(); // unreachable
}

} /* namespace */

//==============================================================================

namespace Cgstmad {

/*******************************************************************************
 * Destructor.
 */
RuleBooleanNode::~RuleBooleanNode()
{
}

/*******************************************************************************
 * Destructor.
 */
RuleBooleanAndExpression::~RuleBooleanAndExpression()
{
}

bool
RuleBooleanAndExpression::eval(const Posture& posture) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	return child1_->eval(posture) && child2_->eval(posture);
}

void
RuleBooleanAndExpression::print(std::ostream& out, int level) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	std::string prefix(level * 8, ' ');
	out << prefix << andOpSymb << " [\n";

	child1_->print(out, level + 1);
	child2_->print(out, level + 1);

	out << prefix << "]" << endl;
}

/*******************************************************************************
 * Destructor.
 */
RuleBooleanOrExpression::~RuleBooleanOrExpression()
{
}

bool
RuleBooleanOrExpression::eval(const Posture& posture) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	return child1_->eval(posture) || child2_->eval(posture);
}

void
RuleBooleanOrExpression::print(std::ostream& out, int level) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	std::string prefix(level * 8, ' ');
	out << prefix << orOpSymb << " [\n";

	child1_->print(out, level + 1);
	child2_->print(out, level + 1);

	out << prefix << "]" << endl;
}

/*******************************************************************************
 * Destructor.
 */
RuleBooleanXorExpression::~RuleBooleanXorExpression()
{
}

bool
RuleBooleanXorExpression::eval(const Posture& posture) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	return child1_->eval(posture) != child2_->eval(posture);
}

void
RuleBooleanXorExpression::print(std::ostream& out, int level) const
{
	assert(child1_.get() != 0 && child2_.get() != 0);

	std::string prefix(level * 8, ' ');
	out << prefix << xorOpSymb << " [\n";

	child1_->print(out, level + 1);
	child2_->print(out, level + 1);

	out << prefix << "]" << endl;
}

/*******************************************************************************
 * Destructor.
 */
RuleBooleanNotExpression::~RuleBooleanNotExpression()
{
}

bool
RuleBooleanNotExpression::eval(const Posture& posture) const
{
	assert(child_.get() != 0);

	return !(child_->eval(posture));
}

void
RuleBooleanNotExpression::print(std::ostream& out, int level) const
{
	assert(child_.get() != 0);

	std::string prefix(level * 8, ' ');
	out << prefix << notOpSymb << " [\n";

	child_->print(out, level + 1);

	out << prefix << "]" << endl;
}

/*******************************************************************************
 * Destructor.
 */
RuleBooleanTerminal::~RuleBooleanTerminal()
{
}

bool
RuleBooleanTerminal::eval(const Posture& posture) const
{
	if (matchPartial_) {
		return posture.isMemberOfCategory(text_, Text::MatchWithOptionalCharSuffix('\''));
	} else {
		if (categoryCode_ != 0) {
			return posture.isMemberOfCategory(categoryCode_);
		} else {
			return posture.isMemberOfCategory(text_, true);
		}
	}
	return false;
}

void
RuleBooleanTerminal::print(std::ostream& out, int level) const
{
	std::string prefix(level * 8, ' ');

	out << prefix << "category [" << text_ << '_' << categoryCode_;
	if (matchPartial_) out << "[*]";
	out << "]" << endl;
}

/*******************************************************************************
 *
 */
void
Rule::printBooleanNodeTree() const
{
	for (RuleBooleanNodeList::size_type size = booleanNodeList_.size(), i = 0; i < size; ++i) {
		cout << "Posture: " << (i + 1) << endl;
		booleanNodeList_[i]->print(cout);
	}
}

/*******************************************************************************
 *
 */
bool
Rule::evalBooleanExpression(const PostureTmpList& postureList) const
{
	if (postureList.size() < booleanNodeList_.size()) return false;
	if (booleanNodeList_.empty()) return false;

	for (RuleBooleanNodeList::size_type size = booleanNodeList_.size(), i = 0; i < size; ++i) {
		if ( !(booleanNodeList_[i]->eval(*postureList[i])) ) {
			return false;
		}
	}
	return true;
}

/*******************************************************************************
 * Get the number of postures in the boolean expression.
 */
size_t
Rule::getNumberOfPostures() const
{
	return booleanNodeList_.size();
}

/*******************************************************************************
 *
 */
void
Rule::parseBooleanExpression(const CategoryCodeMap& categoryCodeMap)
{
	for (BooleanExpressionList::size_type size = booleanExpressionList_.size(), i = 0; i < size; ++i) {
		Parser p(booleanExpressionList_[i], categoryCodeMap);
		booleanNodeList_.push_back(p.getBooleanNode());
	}

	RuleBooleanNodeList::size_type size = booleanNodeList_.size();
	if (size < 2 || size > 4) {
		throw ControllerException(ControllerError("Invalid number of boolean expressions: ") << size);
	}
}

} /* namespace Cgstmad */
