/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef RULE_H_
#define RULE_H_

#include "main.h"

#include <iostream>
#include <string>
#include <tr1/array>
#include <tr1/memory>
#include <vector>

#include "posture.h"



namespace Cgstmad {

class Model;

class RuleBooleanNode {
public:
	virtual ~RuleBooleanNode();

	virtual bool eval(const Posture& posture) const = 0;
	virtual void print(std::ostream& out, int level = 0) const = 0;
};

typedef std::tr1::shared_ptr<RuleBooleanNode> RuleBooleanNode_ptr;
// Type of container that manages the RuleBooleanNode instances.
typedef std::vector<RuleBooleanNode_ptr> RuleBooleanNodeList;

class RuleBooleanAndExpression : public RuleBooleanNode {
public:
	RuleBooleanAndExpression(RuleBooleanNode_ptr c1, RuleBooleanNode_ptr c2)
			: RuleBooleanNode(), child1_(c1), child2_(c2) {}
	virtual ~RuleBooleanAndExpression();

	virtual bool eval(const Posture& posture) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	RuleBooleanNode_ptr child1_;
	RuleBooleanNode_ptr child2_;
};

class RuleBooleanOrExpression : public RuleBooleanNode {
public:
	RuleBooleanOrExpression(RuleBooleanNode_ptr c1, RuleBooleanNode_ptr c2)
			: RuleBooleanNode(), child1_(c1), child2_(c2) {}
	virtual ~RuleBooleanOrExpression();

	virtual bool eval(const Posture& posture) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	RuleBooleanNode_ptr child1_;
	RuleBooleanNode_ptr child2_;
};

class RuleBooleanXorExpression : public RuleBooleanNode {
public:
	RuleBooleanXorExpression(RuleBooleanNode_ptr c1, RuleBooleanNode_ptr c2)
			: RuleBooleanNode(), child1_(c1), child2_(c2) {}
	virtual ~RuleBooleanXorExpression();

	virtual bool eval(const Posture& posture) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	RuleBooleanNode_ptr child1_;
	RuleBooleanNode_ptr child2_;
};

class RuleBooleanNotExpression : public RuleBooleanNode {
public:
	RuleBooleanNotExpression(RuleBooleanNode_ptr c)
			: RuleBooleanNode(), child_(c) {}
	virtual ~RuleBooleanNotExpression();

	virtual bool eval(const Posture& posture) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	RuleBooleanNode_ptr child_;
};

class RuleBooleanTerminal : public RuleBooleanNode {
public:
	RuleBooleanTerminal(const std::string& text, bool matchPartial, int categoryCode)
			: RuleBooleanNode(), text_(text), matchPartial_(matchPartial), categoryCode_(categoryCode) {}
	virtual ~RuleBooleanTerminal();

	virtual bool eval(const Posture& posture) const;
	virtual void print(std::ostream& out, int level = 0) const;
private:
	std::string text_;
	bool matchPartial_;
	int categoryCode_;
};

class Rule;

typedef std::tr1::shared_ptr<Rule> Rule_ptr;
typedef std::vector<Rule_ptr>::size_type RuleNumber;

class Rule {
public:
	typedef std::tr1::array<std::string, Parameter::NUM_PARAMETERS> TransitionNameArray;

	struct ExpressionSymbolEquations {
		std::string ruleDuration;
		std::string mark1;
		std::string mark2;

		ExpressionSymbolEquations() : ruleDuration(), mark1(), mark2() {}
	};

	Rule()
		: number_(0)
		, minDuration12_(0)
		, minDuration23_(0)
		, minDuration34_(0)
	{
	}

	size_t getNumberOfPostures() const;
	bool evalBooleanExpression(const PostureTmpList& postureList) const;
	void printBooleanNodeTree() const;
	void parseBooleanExpression(const CategoryCodeMap& categoryCodeMap);

	RuleNumber number() const {
		return number_;
	}
	int minDuration12() const {
		return minDuration12_;
	}
	int minDuration23() const {
		return minDuration23_;
	}
	int minDuration34() const {
		return minDuration34_;
	}
	const ExpressionSymbolEquations& exprSymbolEquations() const {
		return exprSymbolEquations_;
	}
	const std::string& getParamProfileTransition(Parameter::Code parameter) const {
		return paramProfileTransitions_[parameter];
	}
	const std::string& getSpecialProfileTransition(Parameter::Code parameter) const {
		return specialProfileTransitions_[parameter];
	}
private:
	typedef std::vector<std::string> BooleanExpressionList;

	RuleNumber number_;
	int minDuration12_;
	int minDuration23_;
	int minDuration34_;
	BooleanExpressionList booleanExpressionList_;
	TransitionNameArray paramProfileTransitions_;
	TransitionNameArray specialProfileTransitions_;
	ExpressionSymbolEquations exprSymbolEquations_;
	RuleBooleanNodeList booleanNodeList_;

	friend class ConfigFile;
};

} /* namespace Cgstmad */

#endif /*RULE_H_*/
