/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "special_transition_processor.h"

#include <cassert>

#include "articulation_controller_util.h"
#include "model.h"
#include "transition.h"
#include "util.h"



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
SpecialTransitionProcessor::SpecialTransitionProcessor(const Model& model, Parameter::Code param)
		: model_(model)
		, parameter_(param)
		, pointIndex_(0)
{
}

/*******************************************************************************
 * Destructor.
 */
SpecialTransitionProcessor::~SpecialTransitionProcessor()
{
}

/*******************************************************************************
 *
 */
void
SpecialTransitionProcessor::clear()
{
	pointList_.resize(0);
	pointIndex_ = 0;
}

/*******************************************************************************
 *
 */
void
SpecialTransitionProcessor::prepare(const Transition& transition)
{
	clear();

	float lastTime = -1.0f;

	// For each point:
	for (size_t size = transition.getNumberOfPointsInList(0), j = 0; j < size; ++j) {
		const Transition::Point& point = transition.getPoint(0, j);
		float time = point.timeExpression.empty() ?
					point.freeTime :
					model_.evalEquationFormula(point.timeExpression);

		//--- Calculate the value.
		// The point value [0.0, 1.0] will be multiplied by the parameter span.
		float specialValue;
		if (Parameter::isVolumeParameter(parameter_)) {
			specialValue = point.value;
		} else {
			specialValue = point.value * model_.getParameterSpan(parameter_);
		}

		if (timeIsValid(time, lastTime)) {
			pointList_.push_back(Point(time, specialValue));
			lastTime = time;
		}
	}
}

/*******************************************************************************
 *
 */
bool
SpecialTransitionProcessor::timeIsValid(float time, float lastTime) const
{
	return ArticulationControllerUtil::validatePointTime(time, 0.0f, model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD), lastTime);
}

/*******************************************************************************
 * Get value of special transition.
 *
 * Obs.:
 * - Time is supposed to be positive.
 */
float
SpecialTransitionProcessor::getValue(float time)
{
	const float epsilon = 1.0e-6f;

	assert(!pointList_.empty());

	float t0, v0;
	if (pointIndex_ > 0) {
		const PointList::size_type prevIndex = pointIndex_ - 1;
		t0 = pointList_[prevIndex].time;
		v0 = pointList_[prevIndex].value;
	} else {
		t0 = 0.0;
		v0 = pointList_[0].value;
	}

	for (PointList::size_type size = pointList_.size(); pointIndex_ < size; ++pointIndex_) {
		const float t1 = pointList_[pointIndex_].time;
		const float v1 = pointList_[pointIndex_].value;
		if (time < t1 - epsilon) { // the time is before the target
			return v0 + (v1 - v0) * (time - t0) / (t1 - t0);
		}
		t0 = t1;
		v0 = v1;
	}

	// Last interval.
	return v0;
}

} /* namespace Cgstmad */
