/***************************************************************************
 *  Copyright 2007 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef TRANSITION_H_
#define TRANSITION_H_

#include "main.h"

#include <string>
#include <tr1/memory>
#include <tr1/unordered_map>
#include <vector>

#include "controller_exception.h"
#include "equation.h"



namespace Cgstmad {

class Transition {
public:
	enum Type {
		TYPE_DIPHONE,
		TYPE_TRIPHONE,
		TYPE_TETRAPHONE,
		TYPE_INVALID
	};
	struct Point {
		enum Type {
			TYPE_P12, // Posture 1 --> Posture 2
			TYPE_P23, // Posture 2 --> Posture 3
			TYPE_P34, // Posture 3 --> Posture 4
			TYPE_SPECIAL,
			TYPE_INVALID
		};
		Type type;
		float value;

		// If timeExpression is not empty, time = timeExpression, otherwise time = freeTime.
		std::string timeExpression;
		float freeTime; // milliseconds

		Point() : type(TYPE_INVALID), value(0.0), timeExpression(), freeTime(0.0) {}

		static Type getTypeFromName(const std::string& typeName) {
			if (typeName == "p12") {
				return Transition::Point::TYPE_P12;
			} else if (typeName == "p23") {
				return Transition::Point::TYPE_P23;
			} else if (typeName == "p34") {
				return Transition::Point::TYPE_P34;
			} else if (typeName == "special") {
				return Transition::Point::TYPE_SPECIAL;
			} else {
				throw ControllerException(ControllerError("Transition::Point::getTypeFromName: Invalid transition point type: ") << typeName);
			}
		}
	};

	Transition(bool special)
			: type_(TYPE_INVALID)
			, special_(special)
	{
	}

	const std::string& name() const {
		return name_;
	}
	const Type type() const {
		return type_;
	}
	const std::string& groupName() const {
		return groupName_;
	}
	bool special() const {
		return special_;
	}
	size_t getNumberOfPointLists() const {
		return pointListList_.size();
	}
	size_t getNumberOfPointsInList(size_t listNumber) const {
		return pointListList_[listNumber].size();
	}
	const Point& getPoint(size_t listNumber, size_t pointNumber) const {
		return *pointListList_[listNumber][pointNumber];
	}

	static Type getTypeFromName(const std::string& typeName) {
		if (typeName == "diphone") {
			return Transition::TYPE_DIPHONE;
		} else if (typeName == "triphone") {
			return Transition::TYPE_TRIPHONE;
		} else if (typeName == "tetraphone") {
			return Transition::TYPE_TETRAPHONE;
		} else {
			throw ControllerException(ControllerError("Transition::getTypeFromName: Invalid transition type: ") << typeName);
		}
	}
private:
	typedef std::tr1::shared_ptr<Point> Point_ptr;
	typedef std::vector<Point_ptr> PointList; // type of container that manages the Point instances
	typedef std::vector<PointList> PointListList; // type of container that manages the PointList instances

	std::string name_;
	Type type_;
	std::string groupName_;
	bool special_;
	PointListList pointListList_;

	friend class ConfigFile;
};

typedef std::tr1::shared_ptr<Transition> Transition_ptr;
typedef std::tr1::unordered_map<std::string, Transition_ptr> TransitionMap; // type of container that manages the Transition instances

} /* namespace Cgstmad */

#endif /*TRANSITION_H_*/
