/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "transition_processor.h"

#include <cassert>

#include "articulation_controller_util.h"
#include "controller_exception.h"
#include "model.h"
#include "util.h"



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
TransitionProcessor::TransitionProcessor(const Model& model, Parameter::Code param)
		: model_(model)
		, parameter_(param)
		, type_(Transition::TYPE_INVALID)
{
}

/*******************************************************************************
 * Destructor.
 */
TransitionProcessor::~TransitionProcessor()
{
}

/*******************************************************************************
 *
 */
void
TransitionProcessor::prepare(
		const Transition& transition,
		const PostureTmpList& postureList)
{
	type_ = transition.type();

	const size_t numBasePostures = transition.getNumberOfPointLists();
	pointListList_.resize(numBasePostures);
	for (PointListList::iterator it = pointListList_.begin(); it != pointListList_.end(); ++it) {
		it->resize(0);
	}
	pointIndexList_.resize(numBasePostures);
	for (PointIndexList::iterator it = pointIndexList_.begin(); it != pointIndexList_.end(); ++it) {
		*it = 0;
	}

	// For each base posture:
	for (size_t basePosture = 0; basePosture < numBasePostures; ++basePosture) {

		const size_t numPoints = transition.getNumberOfPointsInList(basePosture);
		if (numPoints < 2) {
			throw ControllerException(ControllerError("The following transition must have at least two points in each phase: ") << transition.name());
		}

		// Get the start value (v1) and the end value (v2) of the phase.
		float v1, v2;
		switch (basePosture) {
		case Transition::Point::TYPE_P12:
			v1 = postureList[0]->getParameterValue(parameter_);
			v2 = postureList[1]->getParameterValue(parameter_);
			break;
		case Transition::Point::TYPE_P23:
			v1 = postureList[1]->getParameterValue(parameter_);
			v2 = postureList[2]->getParameterValue(parameter_);
			break;
		case Transition::Point::TYPE_P34:
			v1 = postureList[2]->getParameterValue(parameter_);
			v2 = postureList[3]->getParameterValue(parameter_);
			break;
		default:
			// Should not happen.
			assert(false);
			v2 = v1 = 0.0;
		}
		if (Parameter::isVolumeParameter(parameter_)) {
			v1 = Util::amplitude(v1);
			v2 = Util::amplitude(v2);
		}
		const float deltaV = v2 - v1;

		float lastTime = -1.0f;

		// For each point from the base posture to the next:
		for (size_t j = 0; j < numPoints; ++j) {
			const Transition::Point& point = transition.getPoint(basePosture, j);
			float time = point.timeExpression.empty() ? point.freeTime : model_.evalEquationFormula(point.timeExpression);
			float value = v1 + deltaV * point.value;
			if (timeIsValid(basePosture, time, lastTime)) {
				pointListList_[basePosture].push_back(Point(time, value));
				lastTime = time;
			}
		}
	}
}

/*******************************************************************************
 *
 */
bool
TransitionProcessor::timeIsValid(size_t basePosture, float time, float lastTime) const
{
	float beginTime, endTime;
	switch (type_) {
	case Transition::TYPE_DIPHONE:
		        beginTime = 0.0;
		        endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD);
		break;
	case Transition::TYPE_TRIPHONE:
		if (basePosture == 0) {
			beginTime = 0.0;
			endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
		} else {
			beginTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
			endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD);
		}
		break;
	case Transition::TYPE_TETRAPHONE:
		switch (basePosture) {
		case 0:
			beginTime = 0.0;
			endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
			break;
		case 1:
			beginTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK1);
			endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2);
			break;
		default:
			beginTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_MARK2);
			endTime = model_.getFormulaSymbolValue(FormulaSymbol::SYMB_RD);
		}
		break;
	default:
		// Should not happen.
		assert(false);
		return false;
	}

	return ArticulationControllerUtil::validatePointTime(time, beginTime, endTime, lastTime);
}

/*******************************************************************************
 * Get value of normal transition.
 *
 * Obs.:
 * - Time is supposed to be positive.
 */
float
TransitionProcessor::getValue(size_t basePosture, float time)
{
	const float epsilon = 1.0e-6f;

	const PointList& pointList = pointListList_[basePosture];
	PointList::size_type& pointIndex = pointIndexList_[basePosture];
	float t0, v0;
	if (pointIndex > 0) {
		const PointList::size_type prevIndex = pointIndex - 1;
		t0 = pointList[prevIndex].time;
		v0 = pointList[prevIndex].value;
	} else {
		t0 = 0.0;
		v0 = pointList[0].value;
	}

	for (PointList::size_type size = pointList.size(); pointIndex < size; ++pointIndex) {
		const float t1 = pointList[pointIndex].time;
		const float v1 = pointList[pointIndex].value;
		if (time < t1 - epsilon) { // the time is before the target
			return v0 + (v1 - v0) * (time - t0) / (t1 - t0);
		}
		t0 = t1;
		v0 = v1;
	}

	// Last interval.
	return v0;
}

} /* namespace Cgstmad */
