/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#include "main.h"

#include <cstdlib>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <memory>

#include "articulation_controller.h"
#include "exception.h"
#include "output_file.h"
#include "program_configuration.h"
#include "simple_config_file.h"
#include "timer.h"
#include "tube_model.h"
#ifdef USE_XERCES
# include "xml_util.h"
#endif



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
Main::Main()
{
#ifdef USE_XERCES
	try {
		XMLPlatformUtils::Initialize(); // Xerces-C++
	} catch (const XMLException& e) {
		throw Exception("Main(): Error in XMLPlatformUtils::Initialize().");
	}
#endif
}

/*******************************************************************************
 * Destructor.
 */
Main::~Main()
{
#ifdef USE_XERCES
	XMLPlatformUtils::Terminate(); // Xerces-C++
#endif
}

/*******************************************************************************
 *
 */
void
Main::run(int argc, char* argv[])
{
	using std::cerr;
	using std::cout;
	using std::endl;

	std::string configDirPath;
	std::string inputFilePath;
	std::string outputFilePath;

	if (argc == 4) {
		configDirPath  = argv[1];
		inputFilePath  = argv[2];
		outputFilePath = argv[3];
	} else if ((argc == 5) && (strcmp("-v", argv[1]) == 0)) {
		verbose_ = true;
		configDirPath  = argv[2];
		inputFilePath  = argv[3];
		outputFilePath = argv[4];
	} else if ((argc == 5) && (strcmp("-l", argv[1]) == 0)) {
		verbose_ = true;
		logControlParameters_ = true;
		configDirPath  = argv[2];
		inputFilePath  = argv[3];
		outputFilePath = argv[4];
	} else {
		throw ProgramArgsException(Error(PACKAGE_STRING "\n\nUsage: ") << argv[0] <<
						" [ -v | -l ] config_dir input_file output_file\n\n"
						"Options:\n"
						" -v : verbose\n"
						" -l : log control parameters\n"
						"\nIf input_file = \"-\", read from stdin.\n");
	}

	if (verbose_) cout << "BEGIN " PACKAGE_STRING << endl;

	Timer initTimer;

	// Load the configuration.
	std::auto_ptr<ProgramConfiguration> configuration(new ProgramConfiguration(configDirPath));

	// Initialize the articulation controller.
	std::auto_ptr<ArticulationController> controller(new ArticulationController(configDirPath,
	                                                                            inputFilePath,
	                                                                            outputFilePath,
	                                                                            *configuration,
	                                                                            verbose_));

	const double initTime = initTimer.getTime();
	Timer synthTimer;

	// Synthesize the speech.
	controller->process();

	const double synthTime = synthTimer.getTime();
	Timer outputTimer;

	// Save the samples to the output file.
	OutputFile outFile(
			controller->outputBuffer(),
			controller->maximumOutputSampleValue(),
			*configuration,
			outputFilePath.c_str());
	outFile.write();

	const double outputTime = outputTimer.getTime();

	if (verbose_) {
		cout << std::setprecision(3);
		cout << "Wrote scaled samples to file: " << outputFilePath <<
			"\nTimes (seconds):"
			"\n  Initialization: " << initTime <<
			"\n  Synthesis:      " << synthTime <<
			"\n  Output file:    " << outputTime <<
			"\n  Total:          " << (initTime + synthTime + outputTime) <<
			"\nEND " PACKAGE_STRING << endl;
	}
}

bool Main::verbose_ = false;
bool Main::logControlParameters_ = false;

} /* namespace Cgstmad */

/*******************************************************************************
 *
 */
int main(int argc, char* argv[])
{
	try {
		Cgstmad::Main main;
		main.run(argc, argv);

		return EXIT_SUCCESS;

	} catch (const Cgstmad::ProgramArgsException& e) {
		std::cerr << e.what() << std::endl;
	} catch (const std::exception& e) {
		std::cerr << "Exception (" << typeid(e).name() << "): " << e.what() << std::endl;
	} catch (...) {
		std::cerr << "Unexpected exception." << std::endl;
	}

	return EXIT_FAILURE;
}
