/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#include "output_file.h"

#include <cerrno>
#include <cstring>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <sstream>

#include "exception.h"
#include "math_util.h"
#include "program_configuration.h"
#include "util.h"



namespace {

using namespace Cgstmad;

enum {
	// Output file format constants.
	AU_FILE_FORMAT = 0,
	AIFF_FILE_FORMAT = 1,
	WAVE_FILE_FORMAT = 2,

	BITS_PER_SAMPLE = 16, // size in bits per output sample
	BUFFER_SIZE = 256
};

const float OUTPUT_SCALE = 1.0f;
const float RANGE_MAX = 32767.0f; // maximum sample value



void convertToIntMsb(int value, unsigned char array[4]);
void convertToIntLsb(int value, unsigned char array[4]);
void convertToShortMsb(int value, unsigned char array[2]);
void convertToShortLsb(int value, unsigned char array[2]);

void writeIntMsb(int value, FILE *outputFile);
void writeIntLsb(int value, FILE *outputFile);
void writeShortMsb(int value, FILE *outputFile);
void writeShortLsb(int value, FILE *outputFile);

void writeAuFileHeader(int channels, size_t numberSamples, float outputRate, FILE *outputFile);
void writeAiffFileHeader(int channels, size_t numberSamples, float outputRate, FILE *outputFile);
void writeWaveFileHeader(int channels, size_t numberSamples, float outputRate, FILE *outputFile);

void writeSamplesMonoMsb(const std::vector<float>& srcBuffer, float scale, FILE *outputFile);
void writeSamplesMonoLsb(const std::vector<float>& srcBuffer, float scale, FILE *outputFile);
void writeSamplesStereoMsb(const std::vector<float>& srcBuffer, float leftScale, float rightScale, FILE *outputFile);
void writeSamplesStereoLsb(const std::vector<float>& srcBuffer, float leftScale, float rightScale, FILE *outputFile);

void convertIntToFloat80(unsigned int value, unsigned char buffer[10]);



inline
void
convertToIntMsb(int value, unsigned char array[4])
{
	array[0] = (value >> 24) & 0xff;
	array[1] = (value >> 16) & 0xff;
	array[2] = (value >> 8) & 0xff;
	array[3] = value & 0xff;
}

inline
void
convertToIntLsb(int value, unsigned char array[4])
{
	array[3] = (value >> 24) & 0xff;
	array[2] = (value >> 16) & 0xff;
	array[1] = (value >> 8) & 0xff;
	array[0] = value & 0xff;
}

inline
void
convertToShortMsb(int value, unsigned char array[2])
{
	array[0] = (value >> 8) & 0xff;
	array[1] = value & 0xff;
}

inline
void
convertToShortLsb(int value, unsigned char array[2])
{
	array[1] = (value >> 8) & 0xff;
	array[0] = value & 0xff;
}

/*******************************************************************************
 * Writes a 4-byte integer to the file, starting with the most significant
 * byte (i.e. writes the int in big-endian form). This routine will work on
 * both big-endian and little-endian architectures.
 */
void
writeIntMsb(int value, FILE* outputFile)
{
	unsigned char array[4];
	convertToIntMsb(value, array);
	if (fwrite(array, sizeof(array), 1, outputFile) != 1) {
		throw Exception("OutputFile: Error in writeIntMsb().");
	}
}

/*******************************************************************************
 * Writes a 4-byte integer to the file, starting with the least significant
 * byte (i.e. writes the int in little-endian form). This routine will work on
 * both big-endian and little-endian architectures.
 */
void
writeIntLsb(int value, FILE* outputFile)
{
	unsigned char array[4];
	convertToIntLsb(value, array);
	if (fwrite(array, sizeof(array), 1, outputFile) != 1) {
		throw Exception("OutputFile: Error in writeIntLsb().");
	}
}

/*******************************************************************************
 * Writes a 2-byte integer to the file, starting with the most significant
 * byte (i.e. writes the int in big-endian form). This routine will work on
 * both big-endian and little-endian architectures.
 */
void
writeShortMsb(int value, FILE* outputFile)
{
	unsigned char array[2];
	convertToShortMsb(value, array);
	if (fwrite(array, sizeof(array), 1, outputFile) != 1) {
		throw Exception("OutputFile: Error in writeShortMsb().");
	}
}

/*******************************************************************************
 * Writes a 2-byte integer to the file, starting with the least significant
 * byte (i.e. writes the int in little-endian form). This routine will work
 * on both big-endian and little-endian architectures.
 */
void
writeShortLsb(int value, FILE* outputFile)
{
	unsigned char array[2];
	convertToShortLsb(value, array);
	if (fwrite(array, sizeof(array), 1, outputFile) != 1) {
		throw Exception("OutputFile: Error in writeShortLsb().");
	}
}

/*******************************************************************************
 * Writes the header in AU format to the output file.
 */
void
writeAuFileHeader(int channels, size_t numberSamples, float outputRate, FILE* outputFile)
{
	/*  AU magic string: ".snd"  */
	if (fputs(".snd", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeAuFileHeader().");
	}

	/*  Header size (fixed in size at 28 bytes)  */
	writeIntMsb(28, outputFile);

	/*  Number of bytes of sound data  */
	writeIntMsb(channels * numberSamples * sizeof(short), outputFile);

	/*  Sound format:  3 is 16-bit linear  */
	writeIntMsb(3, outputFile);

	/*  Output sample rate in samples/second  */
	writeIntMsb(outputRate, outputFile);

	/*  Number of channels  */
	writeIntMsb(channels, outputFile);

	/*  Optional text description (4 bytes minimum)  */
	writeIntMsb(0, outputFile);
}

/*******************************************************************************
 * Writes the header in AIFF format to the output file.
 */
void
writeAiffFileHeader(int channels, size_t numberSamples, float outputRate, FILE* outputFile)
{
	unsigned char sampleFramesPerSecond[10];
	int soundDataSize = channels * numberSamples * sizeof(short);
	int ssndChunkSize = soundDataSize + 8;
	int formSize = ssndChunkSize + 8 + 26 + 4;

	/*  Form container identifier  */
	if (fputs("FORM", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeAiffFileHeader() (Form container identifier).");
	}

	/*  Form size  */
	writeIntMsb(formSize, outputFile);

	/*  Form container type  */
	if (fputs("AIFF", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeAiffFileHeader() (Form container type).");
	}

	/*  Common chunk identifier  */
	if (fputs("COMM", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeAiffFileHeader() (Common chunk identifier).");
	}

	/*  Chunk size (fixed at 18 bytes)  */
	writeIntMsb(18, outputFile);

	/*  Number of channels  */
	writeShortMsb(channels, outputFile);

	/*  Number of sample frames  */
	writeIntMsb(numberSamples, outputFile);

	/*  Number of bits per samples (fixed at 16)  */
	writeShortMsb(BITS_PER_SAMPLE, outputFile);

	/*  Sample frames per second (output sample rate)  */
	/*  stored as an 80-bit (10-byte) float  */
	convertIntToFloat80(static_cast<unsigned int>(outputRate), sampleFramesPerSecond);
	if (fwrite(sampleFramesPerSecond, sizeof(unsigned char), 10, outputFile) != 10) {
		throw Exception("OutputFile: Error in writeAiffFileHeader() (Sample frames per second).");
	}

	/*  Sound Data chunk identifier  */
	if (fputs("SSND", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeAiffFileHeader() (Sound Data chunk identifier).");
	}

	/*  Chunk size  */
	writeIntMsb(ssndChunkSize, outputFile);

	/*  Offset:  unused, so set to 0  */
	writeIntMsb(0, outputFile);

	/*  Block size:  unused, so set to 0  */
	writeIntMsb(0, outputFile);
}

/*******************************************************************************
 * Writes the header in WAVE format to the output file.
 */
void writeWaveFileHeader(int channels, size_t numberSamples, float outputRate, FILE* outputFile)
{
	int soundDataSize = channels * numberSamples * sizeof(short);
	int dataChunkSize = soundDataSize;
	int formSize = dataChunkSize + 8 + 24 + 4;
	int frameSize = std::ceil(channels * (BITS_PER_SAMPLE / 8.0f));
	int bytesPerSecond = std::ceil(outputRate * frameSize);

	/*  Form container identifier  */
	if (fputs("RIFF", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeWaveFileHeader() (Form container identifier).");
	}

	/*  Form size  */
	writeIntLsb(formSize, outputFile);

	/*  Form container type  */
	if (fputs("WAVE", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeWaveFileHeader() (Form container type).");
	}

	/*  Format chunk identifier (Note:  space after 't' needed)  */
	if (fputs("fmt ", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeWaveFileHeader() (Format chunk identifier).");
	}

	/*  Chunk size (fixed at 16 bytes)  */
	writeIntLsb(16, outputFile);

	/*  Compression code:  1 = PCM  */
	writeShortLsb(1, outputFile);

	/*  Number of channels  */
	writeShortLsb(channels, outputFile);

	/*  Output Sample Rate  */
	writeIntLsb(outputRate, outputFile);

	/*  Bytes per second  */
	writeIntLsb(bytesPerSecond, outputFile);

	/*  Block alignment (frame size)  */
	writeShortLsb(frameSize, outputFile);

	/*  Bits per sample  */
	writeShortLsb(BITS_PER_SAMPLE, outputFile);

	/*  Sound Data chunk identifier  */
	if (fputs("data", outputFile) == EOF) {
		throw Exception("OutputFile: Error in writeWaveFileHeader() (Sound Data chunk identifier).");
	}

	/*  Chunk size  */
	writeIntLsb(dataChunkSize, outputFile);
}

/*******************************************************************************
 * Reads the f.p. samples in the source buffer, scales them, rounds them to a
 * short (16-bit) integer, and writes them to the output file in
 * big-endian format.
 */
void
writeSamplesMonoMsb(const std::vector<float>& srcBuffer, float scale, FILE* outputFile)
{
	/*  Write the samples to file, scaling each sample  */
	unsigned char buffer[BUFFER_SIZE][2];
	size_t pos = 0;
	for (std::vector<float>::const_iterator it = srcBuffer.begin(); it != srcBuffer.end(); ++it) {
		convertToShortMsb(MathUtil::roundToInt<int>(*it * scale), buffer[pos]);
		if (++pos == BUFFER_SIZE) {
			if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
				throw Exception("OutputFile: Error in writeSamplesMonoMsb() [1].");
			}
			pos = 0;
		}
	}
	if (pos > 0) {
		if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
			throw Exception("OutputFile: Error in writeSamplesMonoMsb() [2].");
		}
	}
}

/*******************************************************************************
 * Reads the f.p. samples in the source buffer, scales them, rounds them to a
 * short (16-bit) integer, and writes them to the output file in
 * little-endian format.
 */
void
writeSamplesMonoLsb(const std::vector<float>& srcBuffer, float scale, FILE* outputFile)
{
	/*  Write the samples to file, scaling each sample  */
	unsigned char buffer[BUFFER_SIZE][2];
	size_t pos = 0;
	for (std::vector<float>::const_iterator it = srcBuffer.begin(); it != srcBuffer.end(); ++it) {
		convertToShortLsb(MathUtil::roundToInt<int>(*it * scale), buffer[pos]);
		if (++pos == BUFFER_SIZE) {
			if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
				throw Exception("OutputFile: Error in writeSamplesMonoLsb() [1].");
			}
			pos = 0;
		}
	}
	if (pos > 0) {
		if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
			throw Exception("OutputFile: Error in writeSamplesMonoLsb() [2].");
		}
	}
}

/*******************************************************************************
 * Reads the f.p. samples in the source buffer, does stereo scaling, rounds
 * them to a short (16-bit) integer, and writes them to the output file in
 * big-endian format.
 */
void
writeSamplesStereoMsb(const std::vector<float>& srcBuffer, float leftScale, float rightScale, FILE* outputFile)
{
	/*  Write the samples to file, scaling each sample  */
	unsigned char buffer[2 * BUFFER_SIZE][2];
	size_t pos = 0;
	for (std::vector<float>::const_iterator it = srcBuffer.begin(); it != srcBuffer.end(); ++it) {
		convertToShortMsb(MathUtil::roundToInt<int>(*it * leftScale), buffer[pos]);
		convertToShortMsb(MathUtil::roundToInt<int>(*it * rightScale), buffer[++pos]);
		if (++pos == 2 * BUFFER_SIZE) {
			if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
				throw Exception("OutputFile: Error in writeSamplesStereoMsb() [1].");
			}
			pos = 0;
		}
	}
	if (pos > 0) {
		if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
			throw Exception("OutputFile: Error in writeSamplesStereoMsb() [2].");
		}
	}
}

/*******************************************************************************
 * Reads the f.p. samples in the source buffer, does stereo scaling, rounds
 * them to a short (16-bit) integer, and writes them to the output file in
 * little-endian format.
 */
void
writeSamplesStereoLsb(const std::vector<float>& srcBuffer, float leftScale, float rightScale, FILE* outputFile)
{
	/*  Write the samples to file, scaling each sample  */
	unsigned char buffer[2 * BUFFER_SIZE][2];
	size_t pos = 0;
	for (std::vector<float>::const_iterator it = srcBuffer.begin(); it != srcBuffer.end(); ++it) {
		convertToShortLsb(MathUtil::roundToInt<int>(*it * leftScale), buffer[pos]);
		convertToShortLsb(MathUtil::roundToInt<int>(*it * rightScale), buffer[++pos]);
		if (++pos == 2 * BUFFER_SIZE) {
			if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
				throw Exception("OutputFile: Error in writeSamplesStereoLsb() [1].");
			}
			pos = 0;
		}
	}
	if (pos > 0) {
		if (fwrite(buffer, sizeof(unsigned char[2]), pos, outputFile) != pos) {
			throw Exception("OutputFile: Error in writeSamplesStereoLsb() [2].");
		}
	}
}

/*******************************************************************************
 * Converts an unsigned 4-byte integer to an IEEE 754 10-byte (80-bit)
 * floating point number.
 */
void
convertIntToFloat80(unsigned int value, unsigned char buffer[10])
{
	/*  Set all bytes in buffer to 0  */
	memset(buffer, 0, 10);

	unsigned short i;

	/*  Calculate the exponent  */
	unsigned int exp = value;
	for (i = 0; i < 32; i++) {
		exp >>= 1;
		if (!exp) break;
	}

	/*  Add the bias to the exponent  */
	i += 0x3fff;

	/*  Store the exponent  */
	buffer[0] = (i >> 8) & 0x7f;
	buffer[1] = i & 0xff;

	/*  Calculate the mantissa  */
	for (i = 32; i; i--) {
		if (value & 0x80000000) break;
		value <<= 1;
	}

	/*  Store the mantissa:  this should work on both big-endian and little-endian architectures  */
	buffer[2] = (value >> 24) & 0xff;
	buffer[3] = (value >> 16) & 0xff;
	buffer[4] = (value >> 8) & 0xff;
	buffer[5] = value & 0xff;
}

} /* namespace */

//==============================================================================

namespace Cgstmad {

using std::cout;
using std::endl;

/*******************************************************************************
 * Constructor.
 */
OutputFile::OutputFile(const std::vector<float>& inputBuffer,
				float maximumSampleValue,
				const ProgramConfiguration& configuration,
				const char *fileName)
		: fd_(0)
		, inputBuffer_(inputBuffer)
		, maximumSampleValue_(maximumSampleValue)
		, configuration_(configuration)
{
	/*  Open the output file  */
	fd_ = fopen(fileName, "wb");
	if (fd_ == 0) {
		throw Exception(Error("Could not open the output file: ") << fileName << " Reason: " << strerror(errno));
	}
}

/*******************************************************************************
 * Destructor.
 */
OutputFile::~OutputFile()
{
	/*  Close the output file  */
	fclose(fd_);
}

/*******************************************************************************
 * Scales the samples stored in the input buffer, and writes them to the output
 * file, with the appropriate header. Also does master volume scaling,
 * and stereo balance scaling, if 2 channels of output.
 */
void
OutputFile::write()
{
	/*  Calculate scaling constant  */
	float scale = OUTPUT_SCALE * (RANGE_MAX / maximumSampleValue_) * Util::amplitude(configuration_.volume);

	/*  Print out info  */
	if (Main::verbose()) {
		cout << std::setprecision(8);
		cout << "Number of samples:    " << inputBuffer_.size() << endl;
		cout << "Maximum sample value: " << maximumSampleValue_ << endl;
		cout << "Scale:                " << scale << endl;
	}

	float leftScale = 0.0, rightScale = 0.0;
	/*  If stereo, calculate left and right scaling constants  */
	if (configuration_.channels == 2) {
		/*  Calculate left and right channel amplitudes  */
		leftScale = -((configuration_.balance * 0.5f) - 0.5f) * scale * 2.0f;
		rightScale = ((configuration_.balance * 0.5f) + 0.5f) * scale * 2.0f;

		/*  Print out info  */
		if (Main::verbose()) {
			cout << "Left  scale: " << leftScale << endl;
			cout << "Right scale: " << rightScale << endl;
		}
	}

	/*  Scale and write out samples to the output file  */
	switch (configuration_.outputFileFormat) {
	case AU_FILE_FORMAT:
		writeAuFileHeader(configuration_.channels, inputBuffer_.size(), configuration_.outputRate, fd_);
		if (configuration_.channels == 1) {
			writeSamplesMonoMsb(inputBuffer_, scale, fd_);
		} else {
			writeSamplesStereoMsb(inputBuffer_, leftScale, rightScale, fd_);
		}
		break;
	case AIFF_FILE_FORMAT:
		writeAiffFileHeader(configuration_.channels, inputBuffer_.size(), configuration_.outputRate, fd_);
		if (configuration_.channels == 1) {
			writeSamplesMonoMsb(inputBuffer_, scale, fd_);
		} else {
			writeSamplesStereoMsb(inputBuffer_, leftScale, rightScale, fd_);
		}
		break;
	case WAVE_FILE_FORMAT:
		writeWaveFileHeader(configuration_.channels, inputBuffer_.size(), configuration_.outputRate, fd_);
		if (configuration_.channels == 1) {
			writeSamplesMonoLsb(inputBuffer_, scale, fd_);
		} else {
			writeSamplesStereoLsb(inputBuffer_, leftScale, rightScale, fd_);
		}
		break;
	default:
		throw Exception(Error("Invalid output file format: ") << configuration_.outputFileFormat);
	}
}

} /* namespace Cgstmad */
