/***************************************************************************
 *  Copyright 2007 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "simple_config_file.h"

#include <fstream>
#include <iostream>

#include "exception.h"



namespace {

const char SEPARATOR_CHAR = '=';
const char COMMENT_CHAR = '#';

using namespace Cgstmad;

void
throwException(const std::string& filePath, int lineNumber, const char* message)
{
	throw Exception(Error("Error in SimpleConfigFile: ") << filePath << " Line " << lineNumber << ": " << message);
}

template<typename T>
void
throwException(const std::string& filePath, int lineNumber, const char* message, const T& complement)
{
	throw Exception(Error("Error in SimpleConfigFile: ") << filePath << " Line " << lineNumber << ": " << message << complement);
}

}

//==============================================================================

namespace Cgstmad {

using std::cout;
using std::endl;



SimpleConfigFile::SimpleConfigFile(const std::string& filePath) : configMap_()
{
	typedef Map::iterator MI;
	typedef Map::value_type VT;

	if (Main::verbose()) cout << "Reading the configuration file: " << filePath << endl;

	std::ifstream in(filePath.c_str());
	if (!in) throw Exception(Error("Could not open the configuration file: ") << filePath);

	std::string line;
	int lineNum = 0;
	while (getline(in, line)) {
		lineNum++;
		//cout << "line:[" << line << "]" << endl;

		// Empty line.
		if (line.length() == 0) continue;

		std::string::const_iterator iter = line.begin();

		// Comment (must start at the beginning of the line).
		if (*iter == COMMENT_CHAR) {
			//if (Main::verbose()) cout << "Comment" << endl;
			continue;
		}

		// Read key.
		std::string key;
		if (isspace(*iter)) throwException(filePath, lineNum, "Space at the beginning of the line.");
		while (iter != line.end() && !isspace(*iter)) {
			key += *iter++;
		}
		if (key.length() == 0) throwException(filePath, lineNum, "Empty key.");

		// Separator.
		if (iter == line.end()) throwException(filePath, lineNum, "Missing separator.");
		++iter;
		while (iter != line.end() && isspace(*iter)) ++iter;
		if (iter == line.end() || *iter != SEPARATOR_CHAR) throwException(filePath, lineNum, "Missing separator.");
		++iter;
		if (iter == line.end() || !isspace(*iter)) throwException(filePath, lineNum, "Missing space after separator.");
		++iter;
		while (iter != line.end() && isspace(*iter)) ++iter;

		// Read value.
		std::string value;
		if (iter == line.end()) throwException(filePath, lineNum, "Missing value.");
		while (iter != line.end() && !isspace(*iter)) {
			value += *iter++;
		}
		if (isspace(*iter)) throwException(filePath, lineNum, "Space after the value.");
		if (value.length() == 0) throwException(filePath, lineNum, "Empty value.");

		if (Main::verbose()) cout << "n:" << lineNum
				<< " key:[" << key << "]"
				<< " value:[" << value << "]" << endl;

		std::pair<MI,bool> res = configMap_.insert(VT(key, value));
		if (!res.second) {
			throwException(filePath, lineNum, "Duplicate key: ", key);
		}
	}
}

void
SimpleConfigFile::readInto(bool& var, const std::string& key, bool mustExist) const
{
	Map::const_iterator iter = configMap_.find(key);
	if (iter == configMap_.end()) {
		if (mustExist) {
			throw Exception(Error("Key not found in configuration map: ") << key);
		} else {
			return;
		}
	}

	std::string value = iter->second;
	if (value == "true") {
		var = true;
	} else if (value == "false") {
		var = false;
	} else {
		throw Exception(Error("The value must be \"true\" or \"false\": ") << value);
	}

	//if (Main::verbose()) cout << "key=[" << key << "] value=[" << var << "]" << endl;
}

void
SimpleConfigFile::readInto(std::string& var, const std::string& key, bool mustExist) const
{
	Map::const_iterator iter = configMap_.find(key);
	if (iter == configMap_.end()) {
		if (mustExist) {
			throw Exception(Error("Key not found in configuration map: ") << key);
		} else {
			return;
		}
	}

	std::string value = iter->second;
	if (value.length() > 0) {
		var = value;
	} else {
		throw Exception("The value is empty.");
	}
}

} /* namespace Cgstmad */
