/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#ifndef BAND_PASS_FILTER_H_
#define BAND_PASS_FILTER_H_

#include "cgstmad_trm.h"

#include <cmath>

#include "math_util.h"



namespace Cgstmad {

/*******************************************************************************
 * Band-pass filter.
 */
template<typename F>
class BandPassFilter {
public:
	BandPassFilter(F sampleRate);

	void update(F centerFrequency, F bandwidth);
	F filter(F x);
private:
	F x1_, x2_;
	double y1_, y2_;
	double piT_, twoPiT_;
	double alpha_, beta_, gamma_;
};



/*******************************************************************************
 * Constructor.
 */
template<typename F>
BandPassFilter<F>::BandPassFilter(F sampleRate)
		: x1_(0.0)
		, x2_(0.0)
		, y1_(0.0)
		, y2_(0.0)
		, piT_(0.0)
		, twoPiT_(0.0)
		, alpha_(0.0)
		, beta_(0.0)
		, gamma_(0.0)
{
	const double period = 1.0 / sampleRate;
	piT_ = PI * period;
	twoPiT_ = TWO_PI * period;
}

/******************************************************************************
 *
 *       function:       BandPassFilter::update
 *
 *       purpose:        Sets the bandpass filter coefficients
 *                       according to the current center frequency and
 *                       bandwidth.
 *
 *       arguments:      none
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      tan, cos
 *
 ******************************************************************************/
template<typename F>
void
BandPassFilter<F>::update(F centerFrequency, F bandwidth)
{
	const double tanValue = std::tan(piT_ * bandwidth);
	beta_ = (1.0 - tanValue) / (1.0 + tanValue);
	gamma_ = (1.0 + beta_) * std::cos(twoPiT_ * centerFrequency);
	alpha_ = (1.0 - beta_) * 0.5;
}

/******************************************************************************
 *
 *       function:       BandPassFilter::filter
 *
 *       purpose:        Bandpass filter, with variable center
 *                       frequency and bandwidth.
 *
 *       arguments:      input
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      none
 *
 ******************************************************************************/
template<typename F>
F
BandPassFilter<F>::filter(F x)
{
	const double y = (alpha_ * (x - x2_)) + (gamma_ * y1_) - (beta_ * y2_);
	x2_ = x1_;
	x1_ = x;
	y2_ = y1_;
	y1_ = y;
	return y;
}

} // namespace Cgstmad

#endif /*BAND_PASS_FILTER_H_*/
