/***************************************************************************
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "ecl_sample_rate_converter.h"

#ifdef HAVE_LIBSAMPLERATE

#include <cassert>
#include <cmath>
#include <iostream>
#include <sstream>
#include <string>

#include "trm_exception.h"

#define ERROR_PREFIX "Error in libsamplerate: "



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
ECLSampleRateConverter::ECLSampleRateConverter(FloatType inputRate, FloatType outputRate)
		: srcState_(0)
		, outBufferPos_(0)
		, outBufferBlockSize_(0)
		, maximumSampleValue_(0.0)
{
	inBuffer_.reserve(INPUT_BUFFER_SIZE);

	int errorCode = 0;

	// SRC_SINC_BEST_QUALITY
	// SRC_SINC_MEDIUM_QUALITY
	// SRC_SINC_FASTEST
	// SRC_ZERO_ORDER_HOLD
	// SRC_LINEAR
	srcState_ = src_new(SRC_SINC_FASTEST, 1, &errorCode);
	if (srcState_ == 0) {
		throw TrmException(TrmError(ERROR_PREFIX) << src_strerror(errorCode));
	}

	initialize(inputRate, outputRate);
}

/*******************************************************************************
 * Destructor.
 */
ECLSampleRateConverter::~ECLSampleRateConverter()
{
	src_delete(srcState_);
}

/*******************************************************************************
 *
 */
void
ECLSampleRateConverter::initialize(FloatType inputRate, FloatType outputRate)
{
	srcData_.src_ratio = outputRate / inputRate;

	outBufferBlockSize_ = static_cast<long>(INPUT_BUFFER_SIZE * srcData_.src_ratio) + 1;
	assert(outBufferBlockSize_ > 0);
}

/*******************************************************************************
 *
 */
void
ECLSampleRateConverter::resample(bool endOfInput)
{
	const size_t outSizeTmp = outBuffer_.size();

	// Add a block to the output buffer.
	outBuffer_.resize(outSizeTmp + outBufferBlockSize_);

	srcData_.data_in = &inBuffer_[0];
	srcData_.input_frames = inBuffer_.size();
	srcData_.data_out = &outBuffer_[outSizeTmp];
	srcData_.output_frames = outBufferBlockSize_;
	srcData_.end_of_input = endOfInput;

	int retVal = src_process(srcState_, &srcData_);
	if (retVal != 0) {
		throw TrmException(TrmError(ERROR_PREFIX) << src_strerror(retVal));
	}

	if (srcData_.input_frames_used != static_cast<long>(inBuffer_.size())) {
		throw TrmException(TrmError(ERROR_PREFIX "Incomplete sample rate conversion. Input: ") << inBuffer_.size() <<
								" Used: " << srcData_.input_frames_used);
	}

	// Remove the unused positions.
	if (srcData_.output_frames_gen < outBufferBlockSize_) {
		outBuffer_.resize(outSizeTmp + srcData_.output_frames_gen);
	}

	// Update the maximum value.
	for (long i = outSizeTmp, size = outBuffer_.size(); i < size; ++i) {
		const float absVal = std::abs(outBuffer_[i]);
		if (absVal > maximumSampleValue_) {
			maximumSampleValue_ = absVal;
		}
	}

	inBuffer_.clear();
}

} /* namespace Cgstmad */

#endif /*HAVE_LIBSAMPLERATE*/
