/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef HIGH_PASS_FILTER2_H_
#define HIGH_PASS_FILTER2_H_

#include "cgstmad_trm.h"

#include <cmath>

#include "math_util.h"



namespace Cgstmad {

/*******************************************************************************
 * Second order Butterworth high-pass filter.
 */
template<typename F>
class HighPassFilter2 {
public:
	HighPassFilter2();

	void initializeCoefficients(F sampleRate, F cutoff);
	F filter(F x);
private:
	F x1_, x2_;
	double y1_, y2_;
	double a1_, a2_, b0_, b1_, b2_;
};



/*******************************************************************************
 * Constructor.
 */
template<typename F>
HighPassFilter2<F>::HighPassFilter2()
		: x1_(0.0)
		, x2_(0.0)
		, y1_(0.0)
		, y2_(0.0)
		, a1_(0.0)
		, a2_(0.0)
		, b0_(0.0)
		, b1_(0.0)
		, b2_(0.0)
{
}

template<typename F>
void
HighPassFilter2<F>::initializeCoefficients(F sampleRate, F cutoff)
{
	const double wcT2 = PI * cutoff / sampleRate;     // wc * T / 2 = pi * fc / fs
	const double c = std::cos(wcT2) / std::sin(wcT2); // cot(wc * T / 2)

	const double r = std::sqrt(2.0);
	const double c2 = c * c;
	const double rc = r * c;
	const double invD = 1.0 / (c2 + rc + 1.0);
	a1_ = (2.0 - 2.0 * c2) * invD;
	a2_ = (c2 - rc + 1.0) * invD;
	b0_ = c2 * invD;
	b1_ = -2.0 * b0_;
	b2_ = b0_;
}

template<typename F>
F
HighPassFilter2<F>::filter(F x)
{
	const double y = b0_ * x + b1_ * x1_ + b2_ * x2_ - a1_ * y1_ - a2_ * y2_;
	x2_ = x1_;
	x1_ = x;
	y2_ = y1_;
	y1_ = y;
	return y;
}

} // namespace Cgstmad

#endif /*HIGH_PASS_FILTER2_H_*/
