/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef LOW_PASS_FILTER_H_
#define LOW_PASS_FILTER_H_

#include "cgstmad_trm.h"

#include <cmath>

#include "math_util.h"



namespace Cgstmad {

/*******************************************************************************
 * First order Butterworth low-pass filter.
 */
template<typename F>
class LowPassFilter {
public:
	LowPassFilter();
	LowPassFilter(F sampleRate, F cutoff);

	void initializeCoefficients(F sampleRate, F cutoff);
	F filter(F x);
private:
	F x1_;
	double y1_;
	double a1_, b0_, b1_;
};



/*******************************************************************************
 * Constructor.
 */
template<typename F>
LowPassFilter<F>::LowPassFilter()
		: x1_(0.0)
		, y1_(0.0)
		, a1_(0.0)
		, b0_(0.0)
		, b1_(0.0)
{
}

/*******************************************************************************
 * Constructor.
 */
template<typename F>
LowPassFilter<F>::LowPassFilter(F sampleRate, F cutoff)
		: x1_(0.0)
		, y1_(0.0)
{
	initializeCoefficients(sampleRate, cutoff);
}

template<typename F>
void
LowPassFilter<F>::initializeCoefficients(F sampleRate, F cutoff)
{
	const double wcT2 = PI * cutoff / sampleRate;     // wc * T / 2 = pi * fc / fs
	const double c = std::cos(wcT2) / std::sin(wcT2); // cot(wc * T / 2)

	const double invD = 1.0 / (1.0 + c);
	a1_ = (1.0 - c) * invD;
	b0_ = invD;
	b1_ = b0_;
}

template<typename F>
F
LowPassFilter<F>::filter(F x)
{
	const double y = b0_ * x + b1_ * x1_ - a1_ * y1_;
	x1_ = x;
	y1_ = y;
	return y;
}

} // namespace Cgstmad

#endif /*LOW_PASS_FILTER_H_*/
