/***************************************************************************
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#ifndef RADIATION_IMPEDANCE_H_
#define RADIATION_IMPEDANCE_H_

#include "cgstmad_trm.h"

#include "trm_exception.h"
#include "math_util.h"



namespace Cgstmad {

/*******************************************************************************
 * The radiation impedance is calculated as in [LILJENCRANTS85], but with
 * different values of a and b.
 *
 * Zr = (d.c/A).a.(1-z^-1)/(1-b.z^-1)
 *
 * where:
 * A = pi.r^2 = area in cm^2
 * d = density of air
 * speed of sound: c = 352 m/s (temp = 35 deg C)
 * b is obtained from a table
 * a = (1+b)/2
 * Zr(w = pi) = d.c/A
 *
 * in               outT (transmitted)
 * --------->|--------->
 * <---------|
 * outR (reflected)
 *
 * in: flow
 * outR: flow
 *
 * Kz = (Zr-d.c/A)/(Zr+d.c/A)
 * pressure = ((1-Kz).in).Zr
 * outT = pressure.(pi.r/(2.d.c))
 *        (r is used here as a gain compensation; small r --> small Zr --> big Kz (high reflectivity))
 * outR = Kz.in
 *
 * Transmitted:
 * outT[n] = { (a(b+1) + 2.b).outT[n-1] + -(a.b + b^2).outT[n-2] + (a/r).in[n] + -(a(b+1)/r).in[n-1] + (a.b/r).in[n-2] } / (a+1)
 * outT[n] =              cT1.outT[n-1] +          cT2.outT[n-2] +   cT3.in[n] +         cT4.in[n-1] +     cT5.in[n-2]
 *
 * Reflected:
 * outR[n] = { (a+b).outR[n-1] + (a-1).in[n] + (b-a).in[n-1] } / (a+1)
 * outR[n] =     cR1.outR[n-1] +   cR2.in[n] +   cR3.in[n-1]
 */
template<typename F>
class RadiationImpedance {
public:
	RadiationImpedance(F sampleRate);

	void initialize(F sampleRate);
	void update(F radius /* cm */);
	void process(F in /* flow */, F& outT /* pressure.(pi.r/(2.d.c)) */, F& outR /* flow */);
private:
	F in1_, in2_;     // the previous values ([n-1] and [n-2], respectively)
	double cB0_, cB1_, cB2_, cB3_;
	double outT1_, outT2_; // the previous values ([n-1] and [n-2], respectively)
	double outR1_;         // the previous value
	double cT1_, cT2_, cT3_, cT4_, cT5_; // transmission coefficients
	double cR1_, cR2_, cR3_;             // reflection coefficients

	void initializeBCoefficients(double startSampleRate, double sampleRate, double endSampleRate,
					const double startCoefficients[4], const double endCoefficients[4]);

	static const double MIN_RADIUS_;
	static const double MAX_RADIUS_;
	static const double SAMPLE_RATE_LOWEST_;
	static const double SAMPLE_RATE_MIDDLE_;
	static const double SAMPLE_RATE_HIGHEST_;
	static const double LOWEST_SAMPLERATE_B_COEFFICIENTS_[4];
	static const double MIDDLE_SAMPLERATE_B_COEFFICIENTS_[4];
	static const double HIGHEST_SAMPLERATE_B_COEFFICIENTS_[4];
};



template<typename F> const double RadiationImpedance<F>::MIN_RADIUS_ = 0.5; // cm
template<typename F> const double RadiationImpedance<F>::MAX_RADIUS_ = 3.0; // cm
template<typename F> const double RadiationImpedance<F>::SAMPLE_RATE_LOWEST_  = 20000.0;
template<typename F> const double RadiationImpedance<F>::SAMPLE_RATE_MIDDLE_  = 32000.0;
template<typename F> const double RadiationImpedance<F>::SAMPLE_RATE_HIGHEST_ = 44100.0;
template<typename F> const double RadiationImpedance<F>::LOWEST_SAMPLERATE_B_COEFFICIENTS_[4]  = { -0.973468, 1.28928, -0.425122, 0.0523972 };
template<typename F> const double RadiationImpedance<F>::MIDDLE_SAMPLERATE_B_COEFFICIENTS_[4]  = { -0.772870, 1.45819, -0.561321, 0.0776137 };
template<typename F> const double RadiationImpedance<F>::HIGHEST_SAMPLERATE_B_COEFFICIENTS_[4] = { -0.505545, 1.27020, -0.494359, 0.0690006 };



/*******************************************************************************
 *
 */
template<typename F>
RadiationImpedance<F>::RadiationImpedance(F sampleRate)
		: in1_(0.0)
		, in2_(0.0)
		, cB0_(0.0)
		, cB1_(0.0)
		, cB2_(0.0)
		, cB3_(0.0)
		, outT1_(0.0)
		, outT2_(0.0)
		, outR1_(0.0)
		, cT1_(0.0)
		, cT2_(0.0)
		, cT3_(0.0)
		, cT4_(0.0)
		, cT5_(0.0)
		, cR1_(0.0)
		, cR2_(0.0)
		, cR3_(0.0)
{
	if (sampleRate < SAMPLE_RATE_LOWEST_) {
		throw TrmException(TrmError("Invalid internal sample rate: ") << sampleRate << " (minimum: " << SAMPLE_RATE_LOWEST_ << ")");
	} else if (sampleRate > SAMPLE_RATE_HIGHEST_) {
		throw TrmException(TrmError("Invalid internal sample rate: ") << sampleRate << " (maximum: " << SAMPLE_RATE_HIGHEST_ << ")");
	}

	if (sampleRate > SAMPLE_RATE_MIDDLE_) {
		initializeBCoefficients(SAMPLE_RATE_MIDDLE_, sampleRate, SAMPLE_RATE_HIGHEST_,
				MIDDLE_SAMPLERATE_B_COEFFICIENTS_, HIGHEST_SAMPLERATE_B_COEFFICIENTS_);
	} else {
		initializeBCoefficients(SAMPLE_RATE_LOWEST_, sampleRate, SAMPLE_RATE_MIDDLE_,
				LOWEST_SAMPLERATE_B_COEFFICIENTS_, MIDDLE_SAMPLERATE_B_COEFFICIENTS_);
	}
}

/*******************************************************************************
 * Calculate the coefficients of b for a given sample rate,
 * using linear interpolation.
 */
template<typename F>
void
RadiationImpedance<F>::initializeBCoefficients(double startSampleRate, double sampleRate, double endSampleRate,
						const double startCoefficients[4], const double endCoefficients[4])
{
	const double kStart = (endSampleRate - sampleRate) / (endSampleRate - startSampleRate);
	const double kEnd   = 1.0 - kStart;

	cB0_ = kStart * startCoefficients[0] + kEnd * endCoefficients[0];
	cB1_ = kStart * startCoefficients[1] + kEnd * endCoefficients[1];
	cB2_ = kStart * startCoefficients[2] + kEnd * endCoefficients[2];
	cB3_ = kStart * startCoefficients[3] + kEnd * endCoefficients[3];
}

/*******************************************************************************
 *
 */
template<typename F>
void
RadiationImpedance<F>::update(F radius)
{
	double r = radius;
	MathUtil::bound(r, MIN_RADIUS_, MAX_RADIUS_);

	const double r2 = r * r;
	const double b = cB0_ + cB1_ * r + cB2_ * r2 + cB3_ * r2 * r;
	const double a = (b + 1.0) * 0.5;
	const double coef = 1.0 / (a + 1.0);
	const double invRTimesCoef = (1.0 / r) * coef;
	const double ab = a * b;
	const double abPlusA = ab + a;
	const double aPlusB = a + b;

	cT1_ = (abPlusA + 2.0 * b) * coef;
	cT2_ =       -(aPlusB * b) * coef;
	cT3_ =        a * invRTimesCoef;
	cT4_ = -abPlusA * invRTimesCoef;
	cT5_ =       ab * invRTimesCoef;

	cR1_ =    aPlusB * coef;
	cR2_ = (a - 1.0) * coef;
	cR3_ =   (b - a) * coef;
}

/*******************************************************************************
 *
 */
template<typename F>
void
RadiationImpedance<F>::process(F in, F& outT, F& outR)
{
	outT = cT1_ * outT1_ + cT2_ * outT2_ + cT3_ * in + cT4_ * in1_ + cT5_ * in2_;
	outR = cR1_ * outR1_ + cR2_ * in + cR3_ * in1_;

	in2_ = in1_;
	in1_ = in;
	outT2_ = outT1_;
	outT1_ = outT;
	outR1_ = outR;
}

} // namespace Cgstmad

#endif /*RADIATION_IMPEDANCE_H_*/
