/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#include "sample_rate_converter.h"

#include <cmath>

#include "math_util.h"
#include "util.h"



#define BETA                      5.658        /*  kaiser window parameters  */



namespace Cgstmad {

/*******************************************************************************
 * Constructor.
 */
SampleRateConverter::SampleRateConverter(FloatType inputRate, FloatType outputRate)
		: timeRegisterIncrement_(0)
		, filterIncrement_(0)
		, phaseIncrement_(0)
		, timeRegister_(0)
		, padSize_(0)
		, outBufferPos_(0)
		, sampleRateRatio_(0.0)
		, maximumSampleValue_(0.0)
{
	//h_.assign(0.0); deltaH_.assign(0.0);

	outBuffer_.reserve(OUT_BUFFER_INITIAL_SIZE); // to prevent memory allocations in resample()

	initialize(inputRate, outputRate);
}

/******************************************************************************
 *
 *       function:       SampleRateConverter::initializeFilter
 *
 *       purpose:        Initializes filter impulse response and impulse delta
 *                       values.
 *
 *       arguments:      none
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      sin, cos
 *
 ******************************************************************************/
void
SampleRateConverter::initializeFilter()
{
	/*  INITIALIZE THE FILTER IMPULSE RESPONSE  */
	h_[0] = LP_CUTOFF;
	const FloatType x = static_cast<FloatType>(PI) / L_RANGE;
	for (int i = 1; i < FILTER_LENGTH; ++i) {
		const FloatType y = static_cast<FloatType>(i) * x;
		h_[i] = std::sin(y * LP_CUTOFF) / y;
	}

	/*  APPLY A KAISER WINDOW TO THE IMPULSE RESPONSE  */
	const FloatType iBeta = 1.0f / Util::iZero(BETA);
	for (int i = 0; i < FILTER_LENGTH; i++) {
		const FloatType temp = static_cast<FloatType>(i) / FILTER_LENGTH;
		h_[i] *= Util::iZero(BETA * std::sqrt(1.0f - (temp * temp))) * iBeta;
	}

	/*  INITIALIZE THE FILTER IMPULSE RESPONSE DELTA VALUES  */
	for (int i = 0; i < FILTER_LIMIT; ++i) {
		deltaH_[i] = h_[i+1] - h_[i];
	}
	deltaH_[FILTER_LIMIT] = 0.0f - h_[FILTER_LIMIT];
}

/*******************************************************************************
 *
 */
void
SampleRateConverter::initialize(FloatType inputRate, FloatType outputRate)
{
	/*  INITIALIZE FILTER IMPULSE RESPONSE  */
	initializeFilter();

	/*  CALCULATE SAMPLE RATE RATIO  */
	sampleRateRatio_ = outputRate / inputRate;

	const FloatType tmp = std::pow(static_cast<FloatType>(2.0), FRACTION_BITS);

	/*  CALCULATE TIME REGISTER INCREMENT  */
	timeRegisterIncrement_ = static_cast<unsigned int>(0.5f + tmp / sampleRateRatio_);

	/*  CALCULATE ROUNDED SAMPLE RATE RATIO  */
	const FloatType roundedSampleRateRatio = tmp / timeRegisterIncrement_;

	/*  CALCULATE PHASE OR FILTER INCREMENT  */
	if (sampleRateRatio_ >= 1.0f) {
		filterIncrement_ = L_RANGE;
	} else {
		phaseIncrement_ = static_cast<unsigned int>(0.5f + sampleRateRatio_ * FRACTION_RANGE);
	}

	/*  CALCULATE PAD SIZE  */
	padSize_ = (sampleRateRatio_ >= 1.0f) ? ZERO_CROSSINGS : static_cast<unsigned int>(ZERO_CROSSINGS / roundedSampleRateRatio) + 1;

	ringBuffer_.reset(new RingBuffer(*this));
}

/*******************************************************************************
 *
 */
void
SampleRateConverter::resample()
{
	const int bufferSize = ringBuffer_->buffer.size();

	/*  CALCULATE END POINTER  */
	int endPtr = ringBuffer_->fillPtr - ringBuffer_->padSize;

	/*  ADJUST THE END POINTER, IF LESS THAN ZERO  */
	if (endPtr < 0) endPtr += bufferSize;

	/*  ADJUST THE ENDPOINT, IF LESS THEN THE EMPTY POINTER  */
	if (endPtr < ringBuffer_->emptyPtr) endPtr += bufferSize;

	/*  UPSAMPLE LOOP (SLIGHTLY MORE EFFICIENT THAN DOWNSAMPLING)  */
	if (sampleRateRatio_ >= 1.0f) {
		//printf("Upsampling...\n");
		while (ringBuffer_->emptyPtr < endPtr) {

			/*  RESET ACCUMULATOR TO ZERO  */
			FloatType output = 0.0;

			/*  CALCULATE INTERPOLATION VALUE (STATIC WHEN UPSAMPLING)  */
			FloatType interpolation = static_cast<FloatType>(mValue(timeRegister_)) / M_RANGE;

			/*  COMPUTE THE LEFT SIDE OF THE FILTER CONVOLUTION  */
			int index = ringBuffer_->emptyPtr;
			for (unsigned int filterIndex = lValue(timeRegister_);
					filterIndex < FILTER_LENGTH;
					RingBuffer::decrementIndex(index), filterIndex += filterIncrement_) {
				output += ringBuffer_->buffer[index] * (h_[filterIndex]
						+ deltaH_[filterIndex] * interpolation);
			}

			/*  ADJUST VALUES FOR RIGHT SIDE CALCULATION  */
			timeRegister_ = ~timeRegister_;
			interpolation = static_cast<FloatType>(mValue(timeRegister_)) / M_RANGE;

			/*  COMPUTE THE RIGHT SIDE OF THE FILTER CONVOLUTION  */
			index = ringBuffer_->emptyPtr;
			RingBuffer::incrementIndex(index);
			for (unsigned int filterIndex = lValue(timeRegister_);
					filterIndex < FILTER_LENGTH;
					RingBuffer::incrementIndex(index), filterIndex += filterIncrement_) {
				output += ringBuffer_->buffer[index] * (h_[filterIndex]
						+ deltaH_[filterIndex] * interpolation);
			}

			/*  RECORD MAXIMUM SAMPLE VALUE  */
			const FloatType absoluteSampleValue = std::abs(output);
			if (absoluteSampleValue > maximumSampleValue_) {
				maximumSampleValue_ = absoluteSampleValue;
			}

			/*  CHANGE TIME REGISTER BACK TO ORIGINAL FORM  */
			timeRegister_ = ~timeRegister_;

			/*  INCREMENT THE TIME REGISTER  */
			timeRegister_ += timeRegisterIncrement_;

			/*  INCREMENT THE EMPTY POINTER, ADJUSTING IT AND END POINTER  */
			ringBuffer_->emptyPtr += nValue(timeRegister_);

			if (ringBuffer_->emptyPtr >= bufferSize) {
				ringBuffer_->emptyPtr -= bufferSize;
				endPtr -= bufferSize;
			}

			/*  CLEAR N PART OF TIME REGISTER  */
			timeRegister_ &= (~N_MASK);

			/*  PUT THE SAMPLE IN THE OUTPUT BUFFER  */
			outBuffer_.push_back(output);
		}
	} else {
		//printf("Downsampling...\n");
		/*  DOWNSAMPLING CONVERSION LOOP  */
		while (ringBuffer_->emptyPtr < endPtr) {

			/*  RESET ACCUMULATOR TO ZERO  */
			FloatType output = 0.0;

			/*  COMPUTE P PRIME  */
			unsigned int phaseIndex = static_cast<unsigned int>(0.5f + static_cast<FloatType>(fractionValue(timeRegister_)) * sampleRateRatio_);

			/*  COMPUTE THE LEFT SIDE OF THE FILTER CONVOLUTION  */
			unsigned int impulseIndex;
			int index = ringBuffer_->emptyPtr;
			while ((impulseIndex = (phaseIndex >> M_BITS)) < FILTER_LENGTH) {
				const FloatType impulse = h_[impulseIndex] +
					(deltaH_[impulseIndex] * (static_cast<FloatType>(mValue(phaseIndex)) / M_RANGE));
				output += (ringBuffer_->buffer[index] * impulse);
				RingBuffer::decrementIndex(index);
				phaseIndex += phaseIncrement_;
			}

			/*  COMPUTE P PRIME, ADJUSTED FOR RIGHT SIDE  */
			phaseIndex = static_cast<unsigned int>(0.5f + static_cast<FloatType>(fractionValue(~timeRegister_)) * sampleRateRatio_);

			/*  COMPUTE THE RIGHT SIDE OF THE FILTER CONVOLUTION  */
			index = ringBuffer_->emptyPtr;
			RingBuffer::incrementIndex(index);
			while ((impulseIndex = (phaseIndex >> M_BITS)) < FILTER_LENGTH) {
				const FloatType impulse = h_[impulseIndex] +
					(deltaH_[impulseIndex] * (static_cast<FloatType>(mValue(phaseIndex)) / M_RANGE));
				output += (ringBuffer_->buffer[index] * impulse);
				RingBuffer::incrementIndex(index);
				phaseIndex += phaseIncrement_;
			}

			/*  RECORD MAXIMUM SAMPLE VALUE  */
			const FloatType absoluteSampleValue = std::abs(output);
			if (absoluteSampleValue > maximumSampleValue_) {
				maximumSampleValue_ = absoluteSampleValue;
			}

			/*  INCREMENT THE TIME REGISTER  */
			timeRegister_ += timeRegisterIncrement_;

			/*  INCREMENT THE EMPTY POINTER, ADJUSTING IT AND END POINTER  */
			ringBuffer_->emptyPtr += nValue(timeRegister_);
			if (ringBuffer_->emptyPtr >= bufferSize) {
				ringBuffer_->emptyPtr -= bufferSize;
				endPtr -= bufferSize;
			}

			/*  CLEAR N PART OF TIME REGISTER  */
			timeRegister_ &= (~N_MASK);

			/*  PUT THE SAMPLE IN THE OUTPUT BUFFER  */
			outBuffer_.push_back(output);
		}
	}
}

} /* namespace Cgstmad */
