/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2007, 2008, 2009 Marcelo Y. Matuda                           *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/
/***************************************************************************
 *
 *     Program:       tube
 *
 *     Description:   Software implementation of the Tube
 *                    Resonance Model for speech production.
 *
 *     Author:        Leonard Manzara
 *
 *     Date:          July 5th, 1994
 *
 ***************************************************************************/

#ifndef TUBE_MODEL_H_
#define TUBE_MODEL_H_

#include "cgstmad_trm.h"

#include "tube_model_constants.h"

#include <iostream>
#include <memory>
#include <tr1/array>

#include "delay_filter.h"
#include "band_pass_filter.h"
#include "fluctuation_filter.h"
#include "glottal_noise_filter.h"
#include "low_pass_filter.h"
#include "math_util.h"
#include "noise_source.h"
#include "radiation_impedance.h"
#include "tube_parameters.h"
#include "util.h"
#include "vibrato_filter.h"



namespace Cgstmad {

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
class TubeModel {
public:
	template<typename Configuration> TubeModel(const Configuration& config, bool verbose = false);
	~TubeModel();

	void synthesize();
	void flush() { sampleRateConverter_->flush(); }

	const std::vector<float>& outputBuffer() const {
		return sampleRateConverter_->outBuffer();
	}
	FloatType maximumOutputSampleValue() const {
		return sampleRateConverter_->maximumSampleValue();
	}
	LoopController& loopController() {
		return loopController_;
	}
	FloatType sampleRate() const {
		return sampleRate_;
	}
	GlottalSource& glottalSource() {
		return *glottalSource_;
	}
	TubeParameters& parameters() {
		return parameters_;
	}
	SRC& sampleRateConverter() {
		return *sampleRateConverter_;
	}
private:
	struct SectionData {
		FloatType top[2], bottom[2];
		SectionData() { bottom[1] = bottom[0] = top[1] = top[0] = 0.0; }
	};
	struct ThreeWayJunctionData {
		FloatType alpha1, alpha2, alpha3;
		ThreeWayJunctionData() : alpha1(0.0), alpha2(0.0), alpha3(0.0) {}
	};

	TubeParameters parameters_;
	FloatType sampleRate_;
	LoopController loopController_;
	int currentPtr_;
	int prevPtr_;
	FloatType previousFricCF_;
	FloatType previousFricBW_;
	FloatType previousGlotVol_;
	std::auto_ptr<SRC> sampleRateConverter_;
	std::auto_ptr<GlottalSource> glottalSource_;
	DelayFilter<FloatType> throatDelayFilter_;
	VibratoFilter<FloatType> vibratoFilter_;
	FluctuationFilter<FloatType> fluctuationFilter_;
	BandPassFilter<FloatType> fricationBandpassFilter_;
	LowPassFilter<FloatType> throatFilter_;
	GlottalNoiseFilter<FloatType> glottalNoiseFilter_;
	RadiationImpedance<FloatType> noseRadiationImpedance_;
	RadiationImpedance<FloatType> lipsRadiationImpedance_;
	NoiseSource fricationNoiseSource_;
	NoiseSource glottalNoiseSource_;

	std::tr1::array<SectionData, TRM_TOTAL_SECTIONS> oropharynx_;
	std::tr1::array<FloatType, TRM_TOTAL_COEFFICIENTS> oropharynxCoeff_;

	std::tr1::array<SectionData, TRM_TOTAL_NASAL_SECTIONS> nasal_;
	std::tr1::array<FloatType, TRM_TOTAL_NASAL_COEFFICIENTS> nasalCoeff_;

	ThreeWayJunctionData velumJunctionData_;

	std::tr1::array<FloatType, TRM_TOTAL_FRIC_COEFFICIENTS> fricationTap_;

	TubeModel(const TubeModel&);
	TubeModel& operator=(const TubeModel&);

	void initializeNasalCavity();
	void calculateTubeCoefficients();
	void setFricationTaps();
	FloatType processVocalTract(FloatType frication, FloatType pulse, FloatType noisyPulse);
	void processJunction(FloatType coefficient, SectionData& leftSection, SectionData& rightSection);
	void processDelayJunction(SectionData& leftSection, SectionData& rightSection);
	void processThreeWayJunction(
			const ThreeWayJunctionData& junctionData,
			const FloatType section1In[2], FloatType section1Out[2],
			const FloatType section2In[2], FloatType section2Out[2],
			const FloatType section3In[2], FloatType section3Out[2]);
	void addFrication(FloatType frication, SectionData& section);

	static FloatType calculateJunctionCoefficient(FloatType leftRadius, FloatType rightRadius);
	static void calculateThreeWayJunctionCoefficients(
					FloatType radius1, FloatType radius2, FloatType radius3,
					ThreeWayJunctionData& junctionData);
};

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline void
TubeModel<LoopController, GlottalSource, SRC>::processJunction(FloatType coefficient, SectionData& leftSection, SectionData& rightSection)
{
	const FloatType cpl = coefficient * (leftSection.top[prevPtr_] + rightSection.bottom[prevPtr_]);
	leftSection.bottom[currentPtr_] = (rightSection.bottom[prevPtr_] + cpl) * parameters_.efficiencyFactor;
	rightSection.top[currentPtr_] = (leftSection.top[prevPtr_] - cpl) * parameters_.efficiencyFactor;
}

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline void
TubeModel<LoopController, GlottalSource, SRC>::processDelayJunction(SectionData& leftSection, SectionData& rightSection)
{
	leftSection.bottom[currentPtr_] = rightSection.bottom[prevPtr_] * parameters_.efficiencyFactor;
	rightSection.top[currentPtr_] = leftSection.top[prevPtr_] * parameters_.efficiencyFactor;
}

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline void
TubeModel<LoopController, GlottalSource, SRC>::processThreeWayJunction(
						const ThreeWayJunctionData& junctionData,
						const FloatType section1In[2], FloatType section1Out[2],
						const FloatType section2In[2], FloatType section2Out[2],
						const FloatType section3In[2], FloatType section3Out[2])
{
	const FloatType partialInflux = section1In[prevPtr_] + section2In[prevPtr_] + section3In[prevPtr_];
	section1Out[currentPtr_] = (section2In[prevPtr_] + section3In[prevPtr_] + junctionData.alpha1 * partialInflux) * parameters_.efficiencyFactor;
	section2Out[currentPtr_] = (section1In[prevPtr_] + section3In[prevPtr_] + junctionData.alpha2 * partialInflux) * parameters_.efficiencyFactor;
	section3Out[currentPtr_] = (section1In[prevPtr_] + section2In[prevPtr_] + junctionData.alpha3 * partialInflux) * parameters_.efficiencyFactor;
}

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline void
TubeModel<LoopController, GlottalSource, SRC>::addFrication(FloatType frication, SectionData& section)
{
	section.top[currentPtr_] += frication;
}

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline FloatType
TubeModel<LoopController, GlottalSource, SRC>::calculateJunctionCoefficient(FloatType leftRadius, FloatType rightRadius)
{
	const FloatType leftR2 = leftRadius * leftRadius;
	const FloatType rightR2 = rightRadius * rightRadius;
	return (leftR2 - rightR2) / (leftR2 + rightR2);
}

/*******************************************************************************
 *
 */
template<typename LoopController, typename GlottalSource, typename SRC>
inline void
TubeModel<LoopController, GlottalSource, SRC>::calculateThreeWayJunctionCoefficients(
							FloatType radius1, FloatType radius2, FloatType radius3,
							ThreeWayJunctionData& junctionData)
{
	const FloatType r1_2 = radius1 * radius1;
	const FloatType r2_2 = radius2 * radius2;
	const FloatType r3_2 = radius3 * radius3;
	const FloatType c = 1.0f / (r1_2 + r2_2 + r3_2);
	junctionData.alpha1  = c * (r1_2 - (r2_2 + r3_2));
	junctionData.alpha2 = c * (r2_2 - (r1_2 + r3_2));
	//junctionData.alpha3 = c * (r3_2 - (r1_2 + r2_2));
	junctionData.alpha3 = -1.0f - junctionData.alpha1 - junctionData.alpha2; // alpha1 + alpha2 + alpha3 == -1
}

/******************************************************************************
 *
 *       function:       initializeNasalCavity
 *
 *       purpose:        Calculates the scattering coefficients for the fixed
 *                       sections of the nasal cavity.
 *
 *       arguments:      none
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      none
 *
 ******************************************************************************/
template<typename LoopController, typename GlottalSource, typename SRC>
void
TubeModel<LoopController, GlottalSource, SRC>::initializeNasalCavity()
{
	// Calculate the coefficients for the internal fixed sections of the nasal cavity.
	for (int i = TRM_N2, j = TRM_NC2; i < TRM_N7; i++, j++) {
		nasalCoeff_[j] = calculateJunctionCoefficient(parameters_.nasalRadius[i],
		                                              parameters_.nasalRadius[i+1]);
	}
}

/******************************************************************************
 *
 *       function:       calculateTubeCoefficients
 *
 *       purpose:        Calculates the scattering coefficients for the vocal
 *                       tract according to the current radii.  Also calculates
 *                       the coefficients for the reflection/radiation filter
 *                       pair for the mouth and nose.
 *
 *       arguments:      none
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      none
 *
 ******************************************************************************/
template<typename LoopController, typename GlottalSource, typename SRC>
void
TubeModel<LoopController, GlottalSource, SRC>::calculateTubeCoefficients()
{
	// Calculate the coefficients for the oropharynx.
	for (int i = 0; i < (TRM_TOTAL_REGIONS - 1); ++i) {
		oropharynxCoeff_[i] = calculateJunctionCoefficient(parameters_.radius[i],
		                                                   parameters_.radius[i+1]);
	}

	// Calculate the coefficients for the velum junction.
	// Note: Since the junction is in the middle of the region 4, r1 == r2. */
	calculateThreeWayJunctionCoefficients(parameters_.radius[TRM_R4],
	                                      parameters_.radius[TRM_R4],
	                                      parameters_.velum,
	                                      velumJunctionData_);

	// Calculate the coefficient of the 1st nasal passage.
	nasalCoeff_[TRM_NC1] = calculateJunctionCoefficient(parameters_.velum,
	                                                    parameters_.nasalRadius[TRM_N2]);
}

/******************************************************************************
 *
 *       function:       setFricationTaps
 *
 *       purpose:        Sets the frication taps according to the current
 *                       position and amplitude of frication.
 *
 *       arguments:      none
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      none
 *
 ******************************************************************************/
template<typename LoopController, typename GlottalSource, typename SRC>
void
TubeModel<LoopController, GlottalSource, SRC>::setFricationTaps()
{
	// Calculate position remainder and complement.
	const int integerPart = static_cast<int>(parameters_.fricPos);
	const FloatType complement = parameters_.fricPos - integerPart;
	const FloatType remainder = 1.0 - complement;

	// Set the frication taps.
	for (int i = TRM_FC1; i < TRM_TOTAL_FRIC_COEFFICIENTS; ++i) {
		if (i == integerPart) {
			fricationTap_[i] = remainder * parameters_.fricVol;
			if ((i + 1) < TRM_TOTAL_FRIC_COEFFICIENTS) {
				fricationTap_[++i] = complement * parameters_.fricVol;
			}
		} else {
			fricationTap_[i] = 0.0;
		}
	}
}

/******************************************************************************
 *
 *       function:       processVocalTract
 *
 *       purpose:        Updates the pressure wave throughout the vocal tract,
 *                       and returns the summed output of the oral and nasal
 *                       cavities.  Also injects frication appropriately.
 *
 *       arguments:      input, frication
 *
 *       internal
 *       functions:      reflectionFilter, radiationFilter,
 *                       nasalReflectionFilter, nasalRadiationFilter
 *
 *       library
 *       functions:      none
 *
 ******************************************************************************/
template<typename LoopController, typename GlottalSource, typename SRC>
FloatType
TubeModel<LoopController, GlottalSource, SRC>::processVocalTract(FloatType pulse, FloatType noisyPulse, FloatType frication)
{
	// Move the samples in each section.
	std::swap(currentPtr_, prevPtr_);

	// Add some noise to avoid denormals.
	const FloatType denormalEpsilon = 1.0e-10;
	noisyPulse += (currentPtr_ ? denormalEpsilon : -denormalEpsilon);

	lipsRadiationImpedance_.update(parameters_.radius[TRM_R8]);

	// ===== Update the oropharynx. =====

	// Input signal.
	const FloatType glotNormalisedArea = pulse * parameters_.glottalLeakFactor;
	const FloatType glotCoeff = (glotNormalisedArea - 1.0f) / (glotNormalisedArea + 1.0f);
	oropharynx_[TRM_S1].top[currentPtr_] = (-glotCoeff * oropharynx_[TRM_S1].bottom[prevPtr_] + (1.0f - glotCoeff) * noisyPulse) * parameters_.efficiencyFactor;

	// Process junction S1-S2.
	processJunction(oropharynxCoeff_[TRM_C1], oropharynx_[TRM_S1], oropharynx_[TRM_S2]);

	// Process junction S2-S3.
	processJunction(oropharynxCoeff_[TRM_C2], oropharynx_[TRM_S2], oropharynx_[TRM_S3]);
	addFrication(fricationTap_[TRM_FC1] * frication, oropharynx_[TRM_S3]);

	// Process junction S3-S4.
	processJunction(oropharynxCoeff_[TRM_C3], oropharynx_[TRM_S3], oropharynx_[TRM_S4]);
	addFrication(fricationTap_[TRM_FC2] * frication, oropharynx_[TRM_S4]);

	// Update the 3-way junction between the middle of R4 and the nasal cavity.
	processThreeWayJunction(velumJunctionData_,
	                        oropharynx_[TRM_S4].top, oropharynx_[TRM_S4].bottom,
	                        oropharynx_[TRM_S5].bottom, oropharynx_[TRM_S5].top,
	                        nasal_[TRM_VELUM].bottom, nasal_[TRM_VELUM].top);
	addFrication(fricationTap_[TRM_FC3] * frication, oropharynx_[TRM_S5]);

	// Process junction S5-S6.
	processJunction(oropharynxCoeff_[TRM_C4], oropharynx_[TRM_S5], oropharynx_[TRM_S6]);
	addFrication(fricationTap_[TRM_FC4] * frication, oropharynx_[TRM_S6]);

	// Process junction inside R5 (S6-S7) (pure delay with damping).
	processDelayJunction(oropharynx_[TRM_S6], oropharynx_[TRM_S7]);
	addFrication(fricationTap_[TRM_FC5] * frication, oropharynx_[TRM_S7]);

	// Process junction S7-S8.
	processJunction(oropharynxCoeff_[TRM_C5], oropharynx_[TRM_S7], oropharynx_[TRM_S8]);
	addFrication(fricationTap_[TRM_FC6] * frication, oropharynx_[TRM_S8]);

	// Process junction S8-S9.
	processJunction(oropharynxCoeff_[TRM_C6], oropharynx_[TRM_S8], oropharynx_[TRM_S9]);
	addFrication(fricationTap_[TRM_FC7] * frication, oropharynx_[TRM_S9]);

	// Process junction S9-S10.
	processJunction(oropharynxCoeff_[TRM_C7], oropharynx_[TRM_S9], oropharynx_[TRM_S10]);
	addFrication(fricationTap_[TRM_FC8] * frication, oropharynx_[TRM_S10]);

	FloatType lipsOutput = 0.0;
	lipsRadiationImpedance_.process(oropharynx_[TRM_S10].top[prevPtr_], lipsOutput, oropharynx_[TRM_S10].bottom[currentPtr_]);

	// ===== Update the nasal cavity. =====

	processJunction(nasalCoeff_[TRM_NC1], nasal_[TRM_VELUM], nasal_[TRM_N2]);
	processJunction(nasalCoeff_[TRM_NC2], nasal_[TRM_N2]   , nasal_[TRM_N3]);
	processJunction(nasalCoeff_[TRM_NC3], nasal_[TRM_N3]   , nasal_[TRM_N4]);
	processJunction(nasalCoeff_[TRM_NC4], nasal_[TRM_N4]   , nasal_[TRM_N5]);
	processJunction(nasalCoeff_[TRM_NC5], nasal_[TRM_N5]   , nasal_[TRM_N6]);
	processJunction(nasalCoeff_[TRM_NC6], nasal_[TRM_N6]   , nasal_[TRM_N7]);

	FloatType noseOutput = 0.0;
	noseRadiationImpedance_.process(nasal_[TRM_N7].top[prevPtr_], noseOutput, nasal_[TRM_N7].bottom[currentPtr_]);

	// Return the summed output from mouth and nose.
	return lipsOutput + noseOutput;
}

/*******************************************************************************
 * Constructor.
 */
template<typename LoopController, typename GlottalSource, typename SRC>
template<typename Configuration>
TubeModel<LoopController, GlottalSource, SRC>::TubeModel(const Configuration& config, bool verbose)
		: parameters_(config)
		, sampleRate_((Util::speedOfSound(TUBE_MODEL_AIR_TEMPERATURE) * TRM_TOTAL_SECTIONS * 100.0) / config.length)
		, loopController_(sampleRate_, parameters_)
		, currentPtr_(1)
		, prevPtr_(0)
		, previousFricCF_(-1.0)  // initialize with an invalid center frequency
		, previousFricBW_(-1.0)  // initialize with an invalid bandwidth
		, previousGlotVol_(-1.0) // initialize with an invalid amplitude
		, sampleRateConverter_(new SRC(sampleRate_, config.outputRate))
		, glottalSource_(new GlottalSource(sampleRate_, config))
		, throatDelayFilter_(sampleRate_, config.throatOffset / (100.0 * Util::speedOfSound(TUBE_MODEL_AIR_TEMPERATURE)))
		, vibratoFilter_(sampleRate_, config.glottalVibratoFrequency)
		, fluctuationFilter_(sampleRate_, config.glottalFluctuationCutoff)
		, fricationBandpassFilter_(sampleRate_)
		, throatFilter_(sampleRate_, config.throatCutoff)
		, noseRadiationImpedance_(sampleRate_)
		, lipsRadiationImpedance_(sampleRate_)
		, fricationNoiseSource_(1L)
		, glottalNoiseSource_(303985330L)
{
	oropharynxCoeff_.assign(0.0);
	nasalCoeff_.assign(0.0);
	fricationTap_.assign(0.0);

	// Initialize nasal cavity fixed scattering coefficients.
	initializeNasalCavity();

	// Prepare the nose radiation impedance.
	noseRadiationImpedance_.update(config.nasalRadius[TRM_TOTAL_NASAL_SECTIONS - 1]);

	if (verbose) std::cout << "Internal sample rate: " << sampleRate_ << std::endl;
}

/*******************************************************************************
 * Destructor.
 */
template<typename LoopController, typename GlottalSource, typename SRC>
TubeModel<LoopController, GlottalSource, SRC>::~TubeModel()
{
}

/******************************************************************************
 *
 *       function:       synthesize
 *
 *       purpose:        Performs the actual synthesis of sound samples.
 *
 ******************************************************************************/
template<typename LoopController, typename GlottalSource, typename SRC>
void
TubeModel<LoopController, GlottalSource, SRC>::synthesize()
{
	while (!loopController_.isFinished()) {

		// Update the shape of the glottal pulse, if necessary.
		if (parameters_.glotVol != previousGlotVol_) {
			glottalSource_->requestUpdate(parameters_.glotVol);
			previousGlotVol_ = parameters_.glotVol;
		}

		// Get the glottal pulse and the glottal noise envelope samples.
		FloatType pulse, pulseNoiseEnvelope;
		glottalSource_->get(
				fluctuationFilter_.filter(
						vibratoFilter_.filter(parameters_.glotFreq, parameters_.glottalVibratoFactor),
						parameters_.glottalFluctuationFactor),
				pulse,
				pulseNoiseEnvelope);

		// Get the glottal low-pass filtered noise sample.
		const FloatType lpGlottalNoise = glottalNoiseFilter_.filter(glottalNoiseSource_.get());

		// Get the noisy glottal pulse sample.
		const FloatType noisyPulse =
			parameters_.glotVol * (
				pulse * parameters_.glottalPulseFactor
				+ lpGlottalNoise * (pulseNoiseEnvelope * (parameters_.glottalNoisePeakLevel - parameters_.glottalNoiseMinimumLevel)
				                    + parameters_.glottalNoiseMinimumLevel)
			);

		if (parameters_.bypassTube) {
			// Output the sample.
			sampleRateConverter_->put(noisyPulse);
		} else {
			// Get the frication noise sample.
			FloatType fricationNoise = fricationNoiseSource_.get();

			const FloatType modulatedFricationNoise = pulse * fricationNoise;

			if (parameters_.modulation) {
				// Cross-mix pure noise with modulated noise.
				FloatType crossmix = parameters_.glotVol * parameters_.crossmixFactor;
				if (crossmix > 1.0) crossmix = 1.0;
				fricationNoise = (modulatedFricationNoise * crossmix) + (fricationNoise * (1.0 - crossmix));
			}

			// Update the frication band-pass filter, if necessary.
			if (parameters_.fricCF != previousFricCF_ || parameters_.fricBW != previousFricBW_) {
				fricationBandpassFilter_.update(parameters_.fricCF, parameters_.fricBW);
				previousFricCF_ = parameters_.fricCF;
				previousFricBW_ = parameters_.fricBW;
			}

			// Prepare the tube model.
			calculateTubeCoefficients();
			setFricationTaps();

			// Get the output from the vocal tract.
			FloatType output = processVocalTract(pulse,
			                                     noisyPulse + (parameters_.aspVol * lpGlottalNoise),
			                                     fricationBandpassFilter_.filter(fricationNoise));

			// Get the output from the throat.
			output += throatDelayFilter_.filter(parameters_.throatFactor * throatFilter_.filter(noisyPulse));

			// Output the sample.
			sampleRateConverter_->put(output);
		}

		loopController_.next();
	}
}

} /* namespace Cgstmad */

#endif /*TUBE_MODEL_H_*/
