/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#ifndef TUBE_PARAMETERS_H_
#define TUBE_PARAMETERS_H_

#include "cgstmad_trm.h"

#include <cmath>

#include "tube_model_constants.h"
#include "util.h"



namespace Cgstmad {

struct TubeParameters {
	FloatType glotFreq;                 // glottal source fundamental frequency (50 - 600 Hz)
	FloatType glotVol;                  // glottal source volume (0.0 - 1.0)
	FloatType aspVol;                   // aspiration noise volume (0.0 - 1.0)
	FloatType fricVol;                  // frication volume (0.0 - 1.0)
	FloatType fricPos;                  // frication position (0.0 - 7.0)
	FloatType fricCF;                   // frication center frequency (100 - 5500 Hz)
	FloatType fricBW;                   // frication bandwidth (250 - 4500 Hz)
	FloatType radius[TRM_TOTAL_REGIONS];// oropharynx radiuses (0.1 - 3.0 cm)
	FloatType velum;                    // velum radius (0.1 - 3.0 cm)
	FloatType test1;                    // test parameter 1
	FloatType test2;                    // test parameter 2

	FloatType efficiencyFactor;         // 1.0 - loss_factor / 100.0 (0.0 - 1.0)
	FloatType crossmixFactor;           // 1.0 / amplitude(mix_offset) (0.0 - 30.0)
	FloatType throatFactor;             // amplitude(throatVol) (0.0 - 1.0)
	FloatType glottalNoisePeakLevel;    // glottal noise peak level (0.0 - 1.0)
	FloatType glottalNoiseMinimumLevel; // glottal noise minimum level (0.0 - 1.0)
	FloatType glottalPulseFactor;       // glottal pulse level (0.0 - 1.0)
	FloatType glottalLeakFactor;        // factor that will determine the leak at the glottal end (0.0 - 0.5)
	FloatType glottalFluctuationFactor; // glottal source fluctuation factor
	FloatType glottalVibratoFactor;     // glottal source vibrato factor
	FloatType nasalRadius[TRM_TOTAL_NASAL_SECTIONS]; // fixed nasal radii (0.1 - 3.0 cm)
	bool modulation;                    // pulse modulation of noise (false = OFF, true = ON)
	bool bypassTube;                    // if true, output only the glottal wave

	TubeParameters()
			: glotFreq(0.0)
			, glotVol(0.0)
			, aspVol(0.0)
			, fricVol(0.0)
			, fricPos(0.0)
			, fricCF(0.0)
			, fricBW(0.0)
			, velum(0.0)
			, test1(0.0)
			, test2(0.0)
			, efficiencyFactor(0.0)
			, crossmixFactor(0.0)
			, throatFactor(0.0)
			, glottalNoisePeakLevel(0.0)
			, glottalNoiseMinimumLevel(0.0)
			, glottalPulseFactor(0.0)
			, glottalLeakFactor(0.0)
			, glottalFluctuationFactor(0.0)
			, glottalVibratoFactor(0.0)
			, modulation(false)
			, bypassTube(false)
	{
		for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) radius[i] = 0.0;
		for (int i = 0; i < TRM_TOTAL_NASAL_SECTIONS; ++i) nasalRadius[i] = 0.0;
	}

	template<typename Configuration>
	void load(const Configuration& config) {
		glotFreq = 0.0;
		glotVol  = 0.0;
		aspVol   = 0.0;
		fricVol  = 0.0;
		fricPos  = 0.0;
		fricCF   = 0.0;
		fricBW   = 0.0;
		for (int i = 0; i < TRM_TOTAL_REGIONS; ++i) {
			radius[i] = 0.0;
		}
		velum    = 0.0;
		test1    = 0.0;
		test2    = 0.0;
		efficiencyFactor         = 1.0f - (config.lossFactor / 100.0f);
		crossmixFactor           = 1.0f / Util::amplitude(config.mixOffset);
		throatFactor             = Util::amplitude(config.throatVol);
		glottalNoisePeakLevel    = config.glottalNoisePeakLevel;
		glottalNoiseMinimumLevel = config.glottalNoiseMinimumLevel;
		glottalPulseFactor       = config.glottalPulseFactor;
		glottalLeakFactor        = config.glottalLeakFactor;
		glottalFluctuationFactor = config.glottalFluctuationFactor;
		glottalVibratoFactor     = config.glottalVibratoFactor;
		nasalRadius[0] = 0.0;
		for (int i = 1; i < TRM_TOTAL_NASAL_SECTIONS; ++i) {
			nasalRadius[i] = config.nasalRadius[i];
		}
		modulation               = config.modulation;
		bypassTube               = config.bypassTube;

		if (glottalNoisePeakLevel < glottalNoiseMinimumLevel) {
			glottalNoisePeakLevel = glottalNoiseMinimumLevel;
		}
	}

	template<typename Configuration>
	TubeParameters(const Configuration& config) {
		load(config);
	}
};

} // namespace Cgstmad

#endif /* TUBE_PARAMETERS_H_ */
