/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#ifndef UTIL_H_
#define UTIL_H_

#include "cgstmad_trm.h"

#include <cmath>



namespace Cgstmad {
namespace Util {

double iZero(double x);
template<typename F> F speedOfSound(F temperature);
template<typename F> F amplitude(F decibelLevel);
template<typename F> F decibel(F amplitude);
template<typename F> F frequency(F pitch);
template<typename F> F pitch(F frequency);

template<typename F>
struct UtilConstants {
	// Volume constants.
	static const F VOL_MAX;
	static const F VOL_MIN;
	static const F MIN_AMPLITUDE;

	// Pitch constants.
	static const F PITCH_BASE;
	static const F PITCH_OFFSET;

	static const F INV_LOG2;
};

template<typename F> const F UtilConstants<F>::VOL_MAX = 0.0;
template<typename F> const F UtilConstants<F>::VOL_MIN = -100.0;
template<typename F> const F UtilConstants<F>::MIN_AMPLITUDE = std::pow(static_cast<F>(10.0), VOL_MIN * 0.05f);
template<typename F> const F UtilConstants<F>::PITCH_BASE = 220.0;
template<typename F> const F UtilConstants<F>::PITCH_OFFSET = 3.0; // middle C = 0.0
template<typename F> const F UtilConstants<F>::INV_LOG2 = 1.0f / std::log(static_cast<F>(2.0));



/******************************************************************************
 *
 *	function:	speedOfSound
 *
 *	purpose:	Returns the speed of sound according to the value of
 *                       the temperature (in Celsius degrees).
 *
 *       arguments:      temperature
 *
 *	internal
 *	functions:	none
 *
 *	library
 *	functions:	none
 *
 ******************************************************************************/
template<typename F>
F
speedOfSound(F temperature)
{
	return 331.4f + (0.6f * temperature);
}

/******************************************************************************
 *
 *       function:       amplitude
 *
 *       purpose:        Converts dB value to amplitude value.
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      pow
 *
 ******************************************************************************/
template<typename F>
F
amplitude(F decibelLevel)
{
	if (decibelLevel <= UtilConstants<F>::VOL_MIN) return 0.0f;
	if (decibelLevel >= UtilConstants<F>::VOL_MAX) return 1.0f;

	/*  ELSE RETURN INVERSE LOG VALUE  */
	return std::pow(static_cast<F>(10.0), decibelLevel * 0.05f);
}

/*******************************************************************************
 * Converts amplitude to decibel.
 */
template<typename F>
F
decibel(F amplitude)
{
	return (amplitude > UtilConstants<F>::MIN_AMPLITUDE) ? std::log10(amplitude) * 20.0f : UtilConstants<F>::VOL_MIN;
}

/******************************************************************************
 *
 *       function:       frequency
 *
 *       purpose:        Converts a given pitch (0 = middle C) to the
 *                       corresponding frequency.
 *
 *       internal
 *       functions:      none
 *
 *       library
 *       functions:      pow
 *
 ******************************************************************************/
template<typename F>
F
frequency(F pitch)
{
	return UtilConstants<F>::PITCH_BASE * std::pow(static_cast<F>(2.0), (pitch + UtilConstants<F>::PITCH_OFFSET) / 12.0f);
}

/*******************************************************************************
 * Converts a given frequency to the corresponding pitch (0 = middle C).
 */
template<typename F>
F
pitch(F frequency)
{
	return std::log(frequency / UtilConstants<F>::PITCH_BASE) * UtilConstants<F>::INV_LOG2 * 12.0f - UtilConstants<F>::PITCH_OFFSET;
}

} /* namespace Util */
} /* namespace Cgstmad */

#endif /*UTIL_H_*/
