/***************************************************************************
 *  Copyright 1991, 1992, 1993, 1994, 1995, 1996, 2001, 2002, 2004, 2007   *
 *    David R. Hill, Leonard Manzara, Craig Schock,                        *
 *    Steve Nygard                                                         *
 *  Copyright 2008 Marcelo Y. Matuda                                       *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
/***************************************************************************
 * 2007-12-21  mym                                                         *
 *   This code is a translated (from C to C++) and modified version of a   *
 *   GnuSpeech's source code.                                              *
 *                                                                         *
 ***************************************************************************/

#ifndef WAVETABLE_H_
#define WAVETABLE_H_

#include "cgstmad_trm.h"

#include <tr1/array>

#include "glottal_noise_envelope.h"
#include "math_util.h"
#include "tube_model_constants.h"



namespace Cgstmad {

/*******************************************************************************
 * Glottal wavetable.
 */
class Wavetable {
public:
	enum {
		TABLE_LENGTH = 512
	};
	typedef std::tr1::array<float, TABLE_LENGTH> WavetableArray;

	template<typename Configuration>
	Wavetable(float sampleRate, const Configuration& config);

	void requestUpdate(float amplitude);
	void get(float frequency, float& pulse, float& noiseEnvelope);

	template<typename T>
	static void fillWavetable(int tableDiv1, int tableDiv2, int tableDiv2Max, int tableDiv3Max,
					bool smoothClosure, float returnPhaseStartLevel, float returnPhaseDuration,
					T& wavetable, bool update);
private:
	const bool smoothClosure_;
	const int tableDiv1_;
	const int tableDiv2Min_;
	const int tableDiv2Max_;
	const int tableDiv3Max_;
	const float tnDelta_;
	const float returnPhaseStartLevel_;
	const float returnPhaseDuration_;
	const float basicIncrement_;
	float currentPosition_;
	float newAmplitude_;
	WavetableArray wavetable_;
	GlottalNoiseEnvelope glottalNoiseEnvelope_;

	void incrementPosition(float frequency);
	void update(float amplitude);

	static float bound(float value);
};



/*******************************************************************************
 * Initialize the wavetable with a glottal pulse.
 */
template<typename T>
void
Wavetable::fillWavetable(int tableDiv1, int tableDiv2, int tableDiv2Max, int tableDiv3Max,
				bool smoothClosure, float returnPhaseStartLevel, float returnPhaseDuration,
				T& wavetable, bool update)
{
	if (!update) {
		// Calculate rise portion of wave table.
		const float dx = 1.0f / tableDiv1;
		float x = 0.0f;
		for (int i = 0; i < tableDiv1; ++i, x += dx) {
			const float x2 = x * x;
			const float x3 = x2 * x;
			wavetable[i] = (3.0f * x2) - (2.0f * x3);
		}
	}

	// Calculate fall portion of wave table.
	float slope0 = 0.0f;
	float y = 0.0f;
	int i = tableDiv1;
	{
		const int tnLength = tableDiv2 - tableDiv1;
		const float dx = 1.0f / tnLength;
		float x = 0.0f;
		for ( ; i < tableDiv2; ++i, x += dx) {
			y = 1.0f - (x * x);
			if (smoothClosure && y < returnPhaseStartLevel) {
				slope0 = (-2.0f * x) / tnLength;
				break;
			}
			wavetable[i] = y;
		}
	}

	// Return phase.
	if (smoothClosure && slope0 < 0.0f) {
		int tableDiv3 = tableDiv2 + MathUtil::roundToInt<int>(TABLE_LENGTH * (returnPhaseDuration / 100.0f));
		if (tableDiv3 > TABLE_LENGTH) {
			tableDiv3 = TABLE_LENGTH;
		}
		const int length = tableDiv3 - i;
		if (length > 0) {
			const float adjustedSlope = slope0 * length;
			const float a1 = adjustedSlope + 2.0f * y;
			const float a2 = adjustedSlope + 3.0f * y;
			const float dx = 1.0f / length;
			float x = -1.0f;
			for ( ; i < tableDiv3; ++i, x+= dx) {
				const float x2 = x * x;
				const float x3 = x2 * x;
				wavetable[i] = a1 * x3 + a2 * x2;
			}
		}
	}

	// Set closed portion of wave table.
	if (update) {
		if (smoothClosure) {
			for ( ; i < tableDiv3Max; ++i) {
				wavetable[i] = 0.0f;
			}
		} else {
			for ( ; i < tableDiv2Max; ++i) {
				wavetable[i] = 0.0f;
			}
		}
	} else {
		for ( ; i < TABLE_LENGTH; ++i) {
			wavetable[i] = 0.0f;
		}
	}
}

/*******************************************************************************
 * Constructor.
 */
template<typename Configuration>
Wavetable::Wavetable(float sampleRate, const Configuration& config)
		: smoothClosure_(config.smoothClosure)
		// Calculate pulse wave table parameters.
		, tableDiv1_(MathUtil::roundToInt<int>(TABLE_LENGTH * (config.tp / 100.0f)))
		, tableDiv2Min_(MathUtil::roundToInt<int>(TABLE_LENGTH * ((config.tp + config.tnMin) / 100.0f)))
		, tableDiv2Max_(MathUtil::roundToInt<int>(TABLE_LENGTH * ((config.tp + config.tnMax) / 100.0f)))
		, tableDiv3Max_(std::min(
					static_cast<int>(TABLE_LENGTH),
					MathUtil::roundToInt<int>(TABLE_LENGTH * ((config.tp + config.tnMax + config.returnPhaseDuration) / 100.0f))))
		, tnDelta_(tableDiv2Max_ - tableDiv2Min_)
		//
		, returnPhaseStartLevel_(config.returnPhaseStartLevel)
		, returnPhaseDuration_(config.returnPhaseDuration)
		, basicIncrement_(TABLE_LENGTH / sampleRate)
		, currentPosition_(TABLE_LENGTH)
		, newAmplitude_(-1.0f)
		, glottalNoiseEnvelope_(config.glottalNoiseEndLevelA, config.glottalNoiseStartLevelB)
{
	// Calculate the initial glottal pulse and store it in the wavetable, for use in the oscillator.
	fillWavetable(tableDiv1_, tableDiv2Max_, tableDiv2Max_, tableDiv3Max_,
			smoothClosure_, returnPhaseStartLevel_, returnPhaseDuration_,
			wavetable_, false);
}

} /* namespace Cgstmad */

#endif /*WAVETABLE_H_*/
