/***************************************************************************
 *  Copyright 2007, 2008 Marcelo Y. Matuda                                 *
 *                                                                         *
 *  This program is free software: you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation, either version 3 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "xml_util.h"

#ifdef USE_XERCES

#include <iostream>
#include <xercesc/util/PlatformUtils.hpp>

#include "xml_exception.h"
#include "text.h"

#define TRANSCODER_BLOCK_SIZE (4 * 1024)



namespace Cgstmad {

using std::cout;
using std::endl;

/*******************************************************************************
 * Constructor.
 */
LocalXmlErrorHandler::LocalXmlErrorHandler() : util_() {}

/*******************************************************************************
 * Destructor.
 */
LocalXmlErrorHandler::~LocalXmlErrorHandler() {}

void
LocalXmlErrorHandler::warning(const SAXParseException& origExc)
{
	try {
		std::string msg;
		util_.convertFromXmlString(origExc.getMessage(), msg);
		msg.insert(0, "Warning in XML parsing: ");
		throw XmlException(msg.c_str());
	} catch (...) {
		throw XmlException("Warning in XML parsing.");
	}
}

void
LocalXmlErrorHandler::error(const SAXParseException& origExc)
{
	try {
		std::string msg;
		util_.convertFromXmlString(origExc.getMessage(), msg);
		msg.insert(0, "Error in XML parsing: ");
		throw XmlException(msg.c_str());
	} catch (...) {
		throw XmlException("Error in XML parsing.");
	}
}

void
LocalXmlErrorHandler::fatalError(const SAXParseException& origExc)
{
	try {
		std::string msg;
		util_.convertFromXmlString(origExc.getMessage(), msg);
		msg.insert(0, "Fatal error in XML parsing: ");
		throw XmlException(msg.c_str());
	} catch (...) {
		throw XmlException("Fatal error in XML parsing.");
	}
}

void
LocalXmlErrorHandler::resetErrors() {}



/*******************************************************************************
 * Constructor.
 */
XmlUtil::XmlUtil()
		: transcoder_(0)
{
	XMLTransService::Codes resCode;
	transcoder_ = XMLPlatformUtils::fgTransService->makeNewTranscoderFor(
				INTERNAL_ENCODING, resCode, TRANSCODER_BLOCK_SIZE);
	switch (resCode) {
	case XMLTransService::UnsupportedEncoding:
		delete transcoder_;
		throw XmlException("XmlUtil(): Error in XMLPlatformUtils::fgTransService->makeNewTranscoderFor(): UnsupportedEncoding: " INTERNAL_ENCODING);
	case XMLTransService::InternalFailure:
		delete transcoder_;
		throw XmlException("XmlUtil(): Error in XMLPlatformUtils::fgTransService->makeNewTranscoderFor(): InternalFailure");
	case XMLTransService::SupportFilesNotFound:
		delete transcoder_;
		throw XmlException("XmlUtil(): Error in XMLPlatformUtils::fgTransService->makeNewTranscoderFor(): SupportFilesNotFound");
	default: // XMLTransService::Ok
		break;
	}
}

/*******************************************************************************
 * Destructor.
 */
XmlUtil::~XmlUtil()
{
	delete transcoder_;
}

/*******************************************************************************
 * Notes:
 * - *node must be a DOMElement.
 * - If the attribute does not exist, returns an empty string.
 */
void
XmlUtil::getAttributeValue(const DOMNode* node, const char* name, std::string& outValue)
{
	const DOMElement* elem = dynamic_cast<const DOMElement*>(node);
	if (elem == 0) {
		//std::string nodeName;
		//convertFromXmlString(node->getNodeName(), nodeName);
		//throw XmlException(XmlError("Node is not an Element: ") << nodeName);
	}

	XmlString attrName;
	convertToXmlString(name, attrName);
	convertFromXmlString(elem->getAttribute(attrName.c_str()), outValue);
}

void
XmlUtil::getText(const DOMNode* node, std::string& outText)
{
	DOMNode* textNode = node->getFirstChild();
	if (textNode == 0 || !isText(textNode)) {
		//throw XmlException("XmlUtil::getText: Text node not found.");
	}
	convertFromXmlString(textNode->getNodeValue(), outText);
}

bool
XmlUtil::isElement(const DOMNode* node)
{
	return node->getNodeType() == DOMNode::ELEMENT_NODE;
}

bool
XmlUtil::isText(const DOMNode* node)
{
	return node->getNodeType() == DOMNode::TEXT_NODE;
}

bool
XmlUtil::isNamed(const DOMNode* node, const XMLCh* name)
{
	return XMLString::equals(name, node->getNodeName());
}

bool
XmlUtil::isNamedElement(const DOMNode* node, const XMLCh* name)
{
	if (!isElement(node)) return false;
	return XMLString::equals(name, node->getNodeName());
}

void
XmlUtil::convertToXmlString(const char* from, XmlString& to)
{
	size_t len = strlen(from);
	to.resize(len);
	if (len > charSizes_.size()) {
		charSizes_.resize(len);
	}
	unsigned int bytesEaten;
	transcoder_->transcodeFrom(reinterpret_cast<const XMLByte*>(from),
					len,
					&to[0],
					BUF_SIZE - 1,
					bytesEaten,
					&charSizes_[0]);
	if (bytesEaten != len) {
		throw XmlException("XmlUtil::convertToXmlString: Error in XMLTranscoder::transcodeFrom: partial transcoding.");
	}
}

void
XmlUtil::convertFromXmlString(const XMLCh* from, std::string& to)
{
	to.clear();
	unsigned int len = XMLString::stringLen(from);
	unsigned int charsEaten;
	try {
		while (len > 0) {
			transcoder_->transcodeTo(from,
						len,
						xmlByteBuf_,
						BUF_SIZE - 1,
						charsEaten,
						XMLTranscoder::UnRep_Throw); // UnRep_RepChar / UnRep_Throw
			xmlByteBuf_[charsEaten] = 0;
			to += reinterpret_cast<char*>(xmlByteBuf_);
			from += charsEaten;
			len -= charsEaten;
		}
	} catch (const XMLException& e) {
		throw XmlException("XmlUtil::convertFromXmlString: Error in XMLTranscoder::transcodeTo: invalid character (probably).");
	}
}

} /* namespace Cgstmad */

#endif /* USE_XERCES */
